#include "DfpFsmDev.h"

#include <QDebug>
#include <QLoggingCategory>

#include "DfpFsmStates.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.device")
}

DfpFsmDev::DfpFsmDev(const DeviceIndex &i, QObject *p)
    : BaseFsm(p)
    , index(i)
{
    Q_ASSERT(!index.isNull());
    setObjectName(__FUNCTION__);
    recreateInput(new DevFlashFsmInput());
}

void DfpFsmDev::setFileStatus(DFP_FILE fileValid)
{
    auto i = getInput();
    if(i->fileStatus == fileValid)
        return;
    i->fileStatus = fileValid;
    evaluate();
}

void DfpFsmDev::setDeviceOnline(bool online)
{
    auto i = getInput();
    if(i->online == online)
        return;
    i->online = online;
    evaluate();
}

void DfpFsmDev::setAllowTransit(bool v)
{
    auto i = getInput();
    if(i->allowTransit == v)
        return;
    i->allowTransit = v;
    evaluate();
}

bool DfpFsmDev::evaluate()
{
    if(BaseFsm::evaluate())
        return true;

    const auto &i = getInput();
    switch (getState()) {
    case DfpFsmStateError:
    {
        if(!i->allowTransit)
            return false;
        auto st = getPrefSt();
        if(st!=getState()) {
            setState(st);
            return true;
        }
        break;
    }
    case DfpFsmStateAlien:
    case DfpFsmStateMine:
    {
        auto st = getPrefSt();
        if(st!=getState()) {
            setState(st);
            return true;
        }
        if(hasRequestState()) {
            acceptRequestedState();
            return true;
        }

        break;
    }
    case DfpFsmStateErasing:
    case DfpFsmStateReading:
    case DfpFsmStateWriting:
        if(!i->online){
            setState(DfpFsmStateError);
            return true;
        }

        if(i->allowTransit)
            i->allowTransit = false;
        if(hasRequestState()) {
            auto st = getRequestState();
            if(st==DfpFsmStateDone) {
                acceptRequestedState();
                return true;
            }
        }
        break;
    case DfpFsmStateDone:
    {
        if(i->allowTransit) {
            setState(getPrefSt());
            return true;
        }
        break;
    }
    default:;
    }
    return false;
}

void DfpFsmDev::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 DfpFsmDev::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case DfpFsmStateErasing: return 5*60*1000;
    case DfpFsmStateWriting: return 10*60*1000;
    case DfpFsmStateReading: return 3*60*1000;
    default: return BaseFsm::stateTimeoutMs(state);
    }
}

bool DfpFsmDev::allOk() const
{
    return getPrefSt() != DfpFsmStateError;
}

FsmState DfpFsmDev::getPrefSt() const
{
    const auto &i = getInput();
    if(!i->online)
        return DfpFsmStateError;
    switch(i->fileStatus) {
    case DFP_FILE_MINE: return DfpFsmStateMine;
    case DFP_FILE_ALIEN: return DfpFsmStateAlien;
    default: return DfpFsmStateError;
    }
}

QDebug operator<<(QDebug dbg, const DevFlashFsmInput &in)
{
    dbg << BaseFsmInput(in);

    dbg << "task:"
        << (in.task.has_value() ? dfpTaskToString(in.task.value()) : "-")
        << "; ";
    dbg << "fileStatus:" << in.fileStatus << "; ";
    dbg << "online:" << in.online << "; ";
    dbg << "allowTransit:" << in.allowTransit << "; ";
    return dbg;
}
