#include "MongoViewer.h"
#include "MongoViewerConfig.h"
#include "mongo/BaseConfig.h"
#include "mongo/MongoDB.h"
#include "mongo/MongoConsts.h"
#include "mongo/MongoConsts.h"
#include "mongo/RootConfigUtil.h"
#include "mongoqt/BsonQtConverter.h"
#include "mongoqt/MongoCredentialsFile.h"
#include "base/RcRunId.h"
#include "base/DaqModuleConverter.h"
#include "base/ConfigSelectorConverter.h"
#include "base/ConfigMetaConverter.h"
#include "base/ConfigDocumentConverter.h"
#include "base/ConfigDocument.h"

#include <QDebug>
#include <QtMath>
#include <QJsonArray>
#include <QJsonDocument>
#include <QCoreApplication>
#include <QElapsedTimer>
#include <QMetaType>

#ifndef Q_MOC_RUN
#include <bsoncxx/builder/stream/document.hpp>
#include <bsoncxx/builder/basic/document.hpp>
#include <bsoncxx/builder/basic/kvp.hpp>
#include <bsoncxx/json.hpp>
#include <bsoncxx/stdx/make_unique.hpp>
#include <bsoncxx/builder/basic/array.hpp>
#include <bsoncxx/types.hpp>

#include <mongocxx/client.hpp>
#include <mongocxx/exception/bulk_write_exception.hpp>
#include <mongocxx/exception/error_code.hpp>
#include <mongocxx/exception/logic_error.hpp>
#include <mongocxx/exception/operation_exception.hpp>
#include <mongocxx/exception/server_error_code.hpp>
#include <mongocxx/instance.hpp>
#include <mongocxx/logger.hpp>
#include <mongocxx/options/client.hpp>
#include <mongocxx/uri.hpp>
#endif

using namespace std;
using namespace bsoncxx;
using bsoncxx::document::view_or_value;
using bsoncxx::builder::basic::kvp;
using bsoncxx::builder::basic::make_document;
using bsoncxx::builder::basic::make_array;
using bsoncxx::builder::stream::document;
using bsoncxx::builder::stream::finalize;
using bsoncxx::type;

namespace {
    struct record_history{
        QString program_type;
        QString run_index;
        int run_number;
    };

    struct record_saved{
        QString program_type;
        QString program_index;
        QString config_name;
    };

    bool datetime_less_than(const QPair<record_history, QDateTime> &v1, const QPair<record_history, QDateTime> &v2)
    {
        return v1.second < v2.second;
    }

    bool datetime_less_than(const QPair<record_saved, QDateTime> &v1, const QPair<record_saved, QDateTime> &v2)
    {
        return v1.second < v2.second;
    }

    template<typename T> bsoncxx::document::view_or_value mk_doc(std::string key, T val)
    {
            return make_document(kvp(key, val));
    }
}

MongoViewer::MongoViewer()
{}

QList<MongoHost> MongoViewer::get_all_hosts(const MongoViewerConfig &config)
{
    QList<MongoHost> hosts = MongoCredentialsFile::get_all_hosts();
    for(int i=0; i<hosts.size(); ++i){
        hosts[i].is_set = set_uri(hosts[i].get_uri_string());
        if(hosts[i].is_set){
            hosts[i].is_connect = try_connection();
            if(hosts[i].is_connect){
                auto client = get_connection();
                mongocxx::database admin = (*client)[QString("admin").toStdString()];
                bsoncxx::document::value res = admin.run_command(mk_doc("listDatabases", 1));
                QJsonObject obj = BsonQtConverter::to_json(res);
                QJsonArray array = obj["databases"].toArray();
                for(int k=0; k<array.size(); ++k){
                    QString db_name = array[k].toObject()["name"].toString();
                    if(!config.show_all_collections &&
                      (db_name != QString::fromStdString(DB)))
                        continue;
                    MongoDatabase db;
                    db.db_name = db_name;
                    db.size = array[k].toObject()["sizeOnDisk"].toInt();
                    mongocxx::database db_cxx = (*client)[db.db_name.toStdString()];
                    bsoncxx::document::value res_col = db_cxx.run_command(
                        make_document(kvp("listCollections", 1.0),
                            kvp("nameOnly", "true")));
                    QJsonObject obj_col = BsonQtConverter::to_json(res_col);
                    QJsonArray array_col = obj_col["cursor"].toObject()["firstBatch"].toArray();
                    for(int k=0; k<array_col.size(); ++k){
                        QString col_name = array_col[k].toObject()["name"].toString();
                        if(!config.show_all_collections &&
                          (col_name != COLLECTION_HISTORY) &&
                          (col_name != COLLECTION_SAVED))
                            continue;
                        MongoCollection collection;
                        collection.name = col_name;
                        db.collections.append(collection);
                    }
                    hosts[i].databeses.append(db);
                }
            }
        }
    }
    return hosts;
}

QList<ConfigDocument> MongoViewer::aggregate(const MongoViewerConfig &config)
{
    qInfo() << "=== MongoViewer::aggregate ";
    QList<ConfigDocument> to_return;
    set_uri(config.selected_host.get_uri_string());
    set_db_collection(config.selected_db.toString(), config.selected_col.toString());
    mongocxx::collection col = get_collection();
    mongocxx::pipeline p{};
//    p.skip((config.page_number-1) * config.rows_on_page);
//    p.limit(config.rows_on_page);
    auto cursor = col.aggregate(p, mongocxx::options::aggregate{});
    QList<QJsonObject>::iterator i;
    for(const auto &i : cursor){
        QJsonObject obj = BsonQtConverter::to_json(i);
        ConfigDocument doc = QVariant::fromValue(obj).value<ConfigDocument>();
        to_return.append(doc);
    }
    return to_return;
}


QList<QJsonObject> MongoViewer::agregate_json(const MongoViewerConfig &config)
{
    qInfo() << __PRETTY_FUNCTION__ << "-----";
    QList<QJsonObject> objects;
    if(config.selected_host.isEmpty())
        return objects;

    try {
        set_uri(config.selected_host.get_uri_string());
        set_db_collection(config.selected_db.toString(), config.selected_col.toString());
        mongocxx::collection col = get_collection();
        mongocxx::pipeline p1{};

        //--- filters
        if(!config.program_type_filter.isEmpty()){
            p1.match(mk_doc(program_type, config.program_type_filter.toStdString()));
        }

        if(config.selected_col.isHistory()){
            if(!config.run_index_filter.isEmpty()){
                p1.match(mk_doc(run_index, config.run_index_filter.toStdString()));
            }

            if(config.show_run_number){
                p1.match(mk_doc(run_number, mk_doc("$gte", config.run_number_from)));
                p1.match(mk_doc(run_number, mk_doc("$lte", config.run_number_to)));
            }

            if(config.show_period){
                bsoncxx::types::b_date date_from_bson =
                        bsoncxx::types::b_date(
                            std::chrono::system_clock::from_time_t(
                                config.period_from.toSecsSinceEpoch()));

                bsoncxx::types::b_date date_to_bson =
                        bsoncxx::types::b_date(
                            std::chrono::system_clock::from_time_t(
                                config.period_to.toSecsSinceEpoch()));

                p1.match(mk_doc(k_date_of_modification, mk_doc("$gte", date_from_bson)));
                p1.match(mk_doc(k_date_of_modification, mk_doc("$lte", date_to_bson)));
            }
        }
        if(config.type == MongoViewerConfig::SAVED){
            if(!config.program_index_filter.isEmpty()){
                p1.match(mk_doc(program_index, config.program_index_filter.toStdString()));
            }
        }

        p1.skip((config.page_number-1) * config.rows_on_page);
        p1.limit(config.rows_on_page);
        auto cursor = col.aggregate(p1, mongocxx::options::aggregate{});
        for (const bsoncxx::document::view& doc : cursor) {
            QJsonObject res = BsonQtConverter::to_json(doc);
            objects.append(res);
        }
        return objects;
    } catch (const std::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo agregate mongo collection warning: %1").arg(str);
        return objects;
    }
}

QMap<ConfigSelector, QJsonObject> MongoViewer::agregate_cs_json(const MongoViewerConfig &config)
{
    return QMap<ConfigSelector, QJsonObject>();
}

QMap<RcRunId, QJsonObject> MongoViewer::agregate_ri_json(const MongoViewerConfig &config)
{
    return QMap<RcRunId, QJsonObject>();
}

QList<ConfigSelector> MongoViewer::get_config_selectors(const MongoViewerConfig &config)
{
    QList<ConfigSelector> to_return;
    if(config.type != MongoViewerConfig::SAVED)
        return to_return;
    set_uri(config.selected_host.get_uri_string());
    set_db_collection(config.selected_db.toString(), config.selected_col.toString());
    mongocxx::collection collection = get_collection();
    QJsonObject projection;

    projection[_id] = 1;
    projection[k_cs_program_type] = 1;
    projection[k_cs_program_index] = 1;
    projection[k_cs_configuration_name] = 1;
    projection[k_current_host] = 1;
    projection[k_date_of_modification] = 1;
    projection[k_config_selector] = 1;
    projection[k_meta] = 1;
    QList<QJsonObject> result = find(QJsonObject(), projection);
    QList<QJsonObject>::iterator i;
    for(i=result.begin(); i!=result.end(); ++i){
//        qInfo().noquote().nospace() << "----" << QJsonDocument(*i).toJson();
        ConfigSelector cs = QVariant::fromValue(*i).value<ConfigSelector>();
        ConfigMeta cm = QVariant::fromValue(*i).value<ConfigMeta>();
        // new format
        if(cs.getType().toString().isEmpty()){
            cs = QVariant::fromValue((*i).value(k_config_selector).toObject()).value<ConfigSelector>();
            cm = QVariant::fromValue((*i).value(k_meta).toObject()).value<ConfigMeta>();
        }
        qInfo() << "--- " << cs.toString() << "=== " << cm.toString();
        to_return.append(cs);
    }
    return to_return;
}

QStringList MongoViewer::get_program_types(const MongoViewerConfig &config)
{
    update_filters(config);
    return all_program_types;
}

QStringList MongoViewer::get_program_indexes(const MongoViewerConfig &config)
{
    update_filters(config);
    return all_program_indexes;
}

void MongoViewer::update_filters(const MongoViewerConfig &config)
{
    set_uri(config.selected_host.get_uri_string());
    set_db_collection(config.selected_db.toString(), config.selected_col.toString());
    mongocxx::collection collection = get_collection();
    QJsonObject projection;
    projection[program_type] = 1;
    projection[program_index] = 1;
    projection[run_index] = 1;
    projection[run_number] = 1;
    projection[k_date_of_modification] = 1;
    projection[k_config_selector] = 1;
    QList<QJsonObject> result = find(QJsonObject(), projection);
    QList<QJsonObject>::iterator i;
    for(i=result.begin(); i!=result.end(); ++i){
        QString program_type_str = (*i).value(program_type).toString();
        QString program_index_str = (*i).value(program_index).toString();
        QString run_index_str = (*i).value(program_index).toString();
        int run_number_str = (*i).value(run_number).toInt();
        QDateTime datetime_str = QDateTime::fromString((*i).value(run_number).toString(), Qt::ISODate);

        if(program_type_str.isEmpty())
            program_type_str = (*i).value(k_config_selector).toObject().value(program_type).toString();
        if(program_index_str.isEmpty())
            program_index_str = (*i).value(k_config_selector).toObject().value(program_index).toString();

        all_program_types.append(program_type_str);
        all_datetimes.append(datetime_str);
        if(config.type == MongoViewerConfig::SAVED){
            all_program_indexes.append(program_index_str);
        }
        if(config.type == MongoViewerConfig::HISTORY){
            all_run_indexes.append(run_index_str);
            all_run_numbers.append(run_number_str);
        }
    }
}

//        qInfo() << "--- " << program_type_str;
//        if(program_type_str.isEmpty())
//            qInfo().noquote().nospace() << QJsonDocument(*i).toJson();

QList<RcRunId> MongoViewer::get_rc_run_ids(const MongoViewerConfig &config)
{
    QList<RcRunId> to_return;
    if(config.type != MongoViewerConfig::HISTORY)
        return to_return;
    set_uri(config.selected_host.get_uri_string());
    set_db_collection(config.selected_db.toString(), config.selected_col.toString());
    mongocxx::collection collection = get_collection();
    QJsonObject projection;

    projection[_id] = 0;
    projection[run_index] = 1;
    projection[run_number] = 1;
//    projection[uuid] = 1;

    QList<QJsonObject> result = find(QJsonObject(), projection);
    QList<QJsonObject>::iterator i;
    for(i=result.begin(); i!=result.end(); ++i){
        RcRunId rc = QVariant::fromValue(*i).value<RcRunId>();
        to_return.append(rc);
    }
    return to_return;
}

QStringList MongoViewer::get_rc_program_type(const MongoViewerConfig &config)
{
    QStringList to_return;
    if(config.type != MongoViewerConfig::HISTORY)
        return to_return;
    set_uri(config.selected_host.get_uri_string());
    set_db_collection(config.selected_db.toString(), config.selected_col.toString());
    mongocxx::collection collection = get_collection();
    QJsonObject projection;

    projection[_id] = 0;
    projection[program_type] = 1;

    QList<QJsonObject> result = find(QJsonObject(), projection);
    QList<QJsonObject>::iterator i;
    for(i=result.begin(); i!=result.end(); ++i){
        to_return.append((*i)[program_type].toString());
    }
    return to_return;
}


//QList<QPair<QString, QString>> MongoViewer::get_saved_filters()
//{
//    QList<QPair<QString, QString>> to_return;
//    if(!config.type.is_saved())
//        return to_return;
//    set_uri(config.selected_host.get_uri_string());
//    set_db_collection(config.selected_db.toString(), config.selected_col.toString());
//    mongocxx::collection collection = get_collection();

//    QJsonObject projection;

//    projection[program_type] = 1;
//    projection[_id] = 0;
//    projection[program_index] = 1;

//    QList<QJsonObject> result = find(QJsonObject(), projection);

//    for(int i=0; i<result.size(); ++i){
//        QString program_type_ = result[i][program_type].toString();
//        QString program_index_ = result[i][program_index].toString();
//        to_return.append(QPair<QString, QString>(program_type_, program_index_));
//    }
//    return to_return;
//}

//QList<QPair<QString, QString>> MongoViewer::get_history_filters()
//{
//    QList<QPair<QString, QString>> to_return;
//    if(!config.type.is_history() )
//        return to_return;

//    QJsonObject projection;
//    projection[program_type] = 1;
//    projection[_id] = 0;
//    projection[run_index] = 1;

//    QList<QJsonObject> result = find(QJsonObject(), projection);
//    for(int i=0; i<result.size(); ++i){
//        QString program_type_ = result[i][program_type].toString();
//        QString program_index_ = result[i][run_index].toString();
//        to_return.append(QPair<QString, QString>(program_type_, program_index_));
//    }
//    return to_return;
//}

