#include "EvbSignalRouter.h"

#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PnpDiscovery.h"
#include "TelnetRemoteControl/TelnetRemoteControl.h"
#include "remote-control-server/RemoteControlServer.h"

#include "AdjustBufferModel.h"
#include "EvBConfigDb.h"
#include "EvBConfigDispatcher.h"
#include "EventBuilderFsm.h"
#include "EvbMainWindow.h"
#include "FileWriter.h"
#include "RunControl.h"
#include "Client.h"
#include "ClientManagerFsm.h"
#include "Combiner.h"
#include "CombinerFsm.h"
#include "StatisticDialog.h"
#include "TcpClient.h"

namespace {
template <typename Func1, typename Func2>
inline void connect(const typename QtPrivate::FunctionPointer<Func1>::Object *sender, Func1 signal,
                    const typename QtPrivate::FunctionPointer<Func2>::Object *receiver, Func2 slot,
                    Qt::ConnectionType atype=Qt::AutoConnection)
{
    QObject::connect(sender, signal, receiver, slot, atype);
}
}

void EvbSignalRouter::connectAll()
{
    connect_PnpServer();
    connect_RemoteControlServer();
    connect_Combiner();
    connect_Fsm();
    connect_WidgetRunControl();
    connect_TelnetRemoteControl();
    connect_StatisticDialog();
    if(configDispatcher) {
        configDispatcher->subscribeWriterConfig(fileWriter);
        configDispatcher->subscribeCS(fileWriter);
    }
}

void EvbSignalRouter::connect_Client(const Client *cl)
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(client_manager_fsm);

    connect(mainWindow, &EvbMainWindow::requestState, cl, &Client::setClientRequestState);
    connect(cl, &Client::clientStateUpdated, client_manager_fsm, &ClientManagerFsm::setClientState);
    connect(mainWindow, &EvbMainWindow::destroyClient, cl, &Client::destroyClient);
    connect(mainWindow, &EvbMainWindow::destroyAll, cl, &Client::deleteLater);
    connect(mainWindow, &EvbMainWindow::updateClientReadout, cl, &Client::updateClientReadout);

    connect(cl, &Client::statUpdated, statisticDialog, &StatisticDialog::analyzerStatUpdated);

    connect(cl, &Client::evSizePeriodUpdated, adjustBufferModel, &AdjustBufferModel::clientEvSizePeriodUpdated);
    connect(mainWindow, &EvbMainWindow::printFsmInput, cl, &Client::reqPrintFsmInput);
}

void EvbSignalRouter::connect_TcpClient(const TcpClient *tcp)
{
    if(!tcp)
        return;

    if(configDispatcher)
        configDispatcher->subscribeClientConfig(tcp);

    PnpDiscovery::subscribeCacheUpdaed(tcp);
    if(mainWindow) {
        connect(tcp, &TcpClient::sendStatus, mainWindow, &EvbMainWindow::updateClientStatus);
    }
    if(adjustBufferModel)
        connect(tcp, &TcpClient::inputBufferSizeUpdated,
                adjustBufferModel, &AdjustBufferModel::clientBuffSizeUpdated);
    if(statisticDialog)
        connect(tcp, &TcpClient::sendStatus,
                statisticDialog, &StatisticDialog::updateTcpStatus);
}

void EvbSignalRouter::connect_PnpServer()
{
    if(!mainWindow)
        return;

    PnpDiscovery::subscribeCacheUpdaed(mainWindow);
    PnpDiscovery::subscribeMyProgram(mainWindow);
}

void EvbSignalRouter::connect_RemoteControlServer()
{
    if(!remoteControlServer || !mainWindow)
        return;

    connect(remoteControlServer, &RemoteControlServer::startRequest,
            mainWindow, &EvbMainWindow::remoteStartRequest);
    connect(remoteControlServer, &RemoteControlServer::stopRequest,
            mainWindow, &EvbMainWindow::remoteStopRequest);
    connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
            mainWindow, &EvbMainWindow::onRemoteStatusUpdated);
    connect(remoteControlServer, &RemoteControlServer::gotRunDocument,
            combiner, &Combiner::gotRunDocument);

    connect(mainWindow, &EvbMainWindow::destroyAll,
            remoteControlServer, &RemoteControlServer::deleteLater);
    connect(mainWindow, &EvbMainWindow::sendRemoteState,
            remoteControlServer, &RemoteControlServer::sendState);
    connect(mainWindow, &EvbMainWindow::sendRunError,
            remoteControlServer, &RemoteControlServer::onRunError);

    connect(mainWindow, &EvbMainWindow::sendRunWarning,
            remoteControlServer, &RemoteControlServer::onRunWarning);
    connect(mainWindow, &EvbMainWindow::sendBackPreasure,
            remoteControlServer, &RemoteControlServer::onBackPressure);

    connect(mainWindow, &EvbMainWindow::allowRemoteControl,
            remoteControlServer, &RemoteControlServer::startListen);
    connect(mainWindow, &EvbMainWindow::sendRunDocumentToRc,
            remoteControlServer, &RemoteControlServer::config_transfer);
}

void EvbSignalRouter::connect_Combiner()
{
    if(!combiner)
        return;

    connect(mainWindow, &EvbMainWindow::destroyAll,
            combiner, &QObject::deleteLater);

    if(adjustBufferModel) {
        connect(combiner, &Combiner::evSizePeriodUpdated,
                adjustBufferModel, &AdjustBufferModel::combinerEvSizePeriodUpdated);
        connect(combiner, &Combiner::outputBufferSizeUpdated,
                adjustBufferModel, &AdjustBufferModel::combinerBuffSizeUpdated);
    }

    configDispatcher->subscribeCombinerConfig(combiner);
    configDispatcher->subscribeEvNumberConfig(combiner);

    connect(mainWindow, &EvbMainWindow::requestState, combiner, &Combiner::setCombRequestState);
    connect(mainWindow, &EvbMainWindow::destroyClient, combiner, &Combiner::destroyClient);
    connect(mainWindow, &EvbMainWindow::analizerListUpdated, combiner, &Combiner::updateAnalizerList);

    //    connect(combiner, &Combiner::clientStateUpdated, fsm, &EventBuilderFsm::setClientState);
    connect(combiner, &Combiner::error, mainWindow, &EvbMainWindow::combinerError);

    connect(combiner, &Combiner::statusRaw, mainWindow, &EvbMainWindow::updateCombinerStats);
    connect(combiner, &Combiner::fileWriterStatsUpdated, mainWindow, &EvbMainWindow::fileWriterStatsChanged);
    connect(mainWindow, &EvbMainWindow::setMonitorPortNum, combiner, &Combiner::setMonitorPortNum);
}

void EvbSignalRouter::connect_Fsm()
{
    if(!evbFsm || !client_manager_fsm)
        return;

    if(combinerFsm) {
        if(evbFsm)
            connect(combinerFsm, &BaseFsm::stateChanged,
                    evbFsm, &EventBuilderFsm::setCombinerState);
        if(mainWindow)
            connect(combinerFsm, &BaseFsm::stateChanged,
                    mainWindow, &EvbMainWindow::updateCombinerFsmLabel);
    }

    if(mainWindow) {
        connect(evbFsm, &EventBuilderFsm::stateChanged,
                mainWindow, &EvbMainWindow::onEvbFsmStateChanged);
        connect(client_manager_fsm, &ClientManagerFsm::stateChanged,
                mainWindow, &EvbMainWindow::updateClientFsmLabel);
        connect(mainWindow, &EvbMainWindow::printFsmInput,
                evbFsm, &EventBuilderFsm::printFsmInput);
        connect(mainWindow, &EvbMainWindow::printFsmInput,
                client_manager_fsm, &BaseFsm::printFsmInput);
        connect(mainWindow, &EvbMainWindow::printFsmInput,
                combinerFsm, &BaseFsm::printFsmInput);
        connect(client_manager_fsm, &ClientManagerFsm::stateChanged,
                evbFsm, &EventBuilderFsm::setClientManagerFsmState);

        connect(mainWindow, &EvbMainWindow::requestState,
                client_manager_fsm, &ClientManagerFsm::setRequestState);

        connect(evbFsm, &EventBuilderFsm::doReqStateAction,
                mainWindow, &EvbMainWindow::doReqStateAction, Qt::DirectConnection);
    }
    if(widgetRunControl)
        connect(evbFsm, &EventBuilderFsm::stateChanged,
                widgetRunControl, &RunControl::onStateChanged);
    if(remoteControlServer)
        connect(evbFsm, &EventBuilderFsm::stateChanged,
                remoteControlServer, &RemoteControlServer::onFsmStateChange);
}

void EvbSignalRouter::connect_WidgetRunControl()
{
    if(!widgetRunControl)
        return;

    if(mainWindow) {
        connect(widgetRunControl, &RunControl::startLocalRun,
                mainWindow, &EvbMainWindow::startLocalRun);
        connect(widgetRunControl, &RunControl::stopLocalRun,
                mainWindow, &EvbMainWindow::stopLocalRun);
    }
}

void EvbSignalRouter::connect_TelnetRemoteControl() const
{
    if(!telnetRc)
        return;

    if(evbFsm)
        connect(evbFsm, &BaseFsm::stateChanged,
            telnetRc, &TelnetRemoteControl::onFsmStateChange);

    if(telnetRc) {
        connect(telnetRc, &TelnetRemoteControl::startRun, mainWindow, &EvbMainWindow::startLocalRun);
        connect(telnetRc, &TelnetRemoteControl::stopRun, mainWindow, &EvbMainWindow::stopLocalRun);
    }
}

void EvbSignalRouter::connect_StatisticDialog()
{
    if(!statisticDialog)
        return;
    configDispatcher->subscribeCombinerConfig(statisticDialog);
    if(combiner)
        connect(combiner, &Combiner::statusRaw,
                statisticDialog, &StatisticDialog::updateCombinerStats);
    if(mainWindow)
        connect(mainWindow, &EvbMainWindow::analizerListUpdated,
                statisticDialog, &StatisticDialog::analizerListUpdated);
    if(fileWriter)
        connect(fileWriter, &FileWriter::statusUpdated,
                statisticDialog, &StatisticDialog::statusUpdated);
}

