//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "TrigMainForm.h"
#include "ui_TrigMainForm.h"

#include "RcConfig.h"
#include "RcConfigDispatcher.h"
#include "RunControlRegs.h"
#include "TrigChannelConfig.h"
#include "TrigChannelState.h"
#include "TrigDevice.h"
#include "base/gui_utils.h"
#include "metric-sender/MetricSender.h"
#include "rc_gui_utils.h"

#include <QComboBox>
#include <QHostAddress>
#include <QMessageBox>
#include <QThread>

namespace {
enum TRIG_STAT_COLS {
//    TRIG_STAT_COL_CH_NUM,
//    TRIG_STAT_COL_TRIG_IN,
    TRIG_STAT_COL_TIME_STBY,
    TRIG_STAT_COL_TIME_RUN,
    TRIG_STAT_COL_TIME_READOUT,
    TRIG_STAT_COL_TIME_XOFF,
//    TRIG_STAT_COL_STATE,
    TRIG_STAT_COL_TRIG,
    TRIG_STAT_COL_TRIG_RUN,
    TRIG_STAT_COL_TRIG_SPILL,
    TRIG_STAT_COL_TRIG_RUN_SPILL,
    TRIG_STAT_COL_TRIG_ACCEPTED,
    TRIG_STAT_COL_TOTAL
};

inline const double scaleDeadTime = 1e6;
} // namespace

TrigMainForm::TrigMainForm(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::TrigMainForm),
    config(new RcConfig),
    prevTrigStatus(new TrigStatus)
{
    ui->setupUi(this);
    MultiSignalBlocker msb1({this});
    MultiSignalBlocker msb2(children());

    trigChannelConfig = new TrigChannelConfig(ui->tableWidgetChCfg);
    qRegisterMetaType<TrigChannelState>();
    qRegisterMetaType<SpillPlotData>();
    qRegisterMetaType<TriggerHistogramChannelData>();
    qRegisterMetaType<TriggerHistogramData>();

//    connect(ui->trigConfigForm, &TrigConfigForm::configUpdated,
//            this, &TrigMainForm::configUpdated);
    connect(trigChannelConfig, &TrigChannelConfig::configUpdated,
            this, &TrigMainForm::configUpdated);
//    connect(this, &TrigMainForm::updateLeds,
//            ui->trigLedForm, &TrigLedForm::updateLeds);
    connect(this, &TrigMainForm::updateChState,
            trigChannelConfig, &TrigChannelConfig::updateChState);
    connect(this, &TrigMainForm::rcFsmStateChanged,
            trigChannelConfig, &TrigChannelConfig::onRcFsmStateChange);
    connect(this, &TrigMainForm::rcFsmStateChanged,
            this, &TrigMainForm::onRcFsmStateChange);

    if (ui->tableWidgetStats->rowCount() != TRIG_STAT_COL_TOTAL)
        ui->tableWidgetStats->setRowCount(TRIG_STAT_COL_TOTAL);
    for(int row=0;row<TRIG_STAT_COL_TOTAL; ++row) {
        auto item =  new QTableWidgetItem();
        item->setFont({"Monospace"});
        ui->tableWidgetStats->setItem(row, 0, item);
    }
#if QT_VERSION >= 0x050000
    ui->tableWidgetStats->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui->tableWidgetStats->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    ui->tableWidgetStats->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
    ui->tableWidgetStats->verticalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    connect(this, &TrigMainForm::spillPlotUpdated,
            ui->trigHistView, &TrigHistView::spillPlotUpdated);
    connect(this, &TrigMainForm::triggerHistogramUpdated,
            ui->trigHistView, &TrigHistView::triggerHistogramUpdated);
}

TrigMainForm::~TrigMainForm()
{
    delete prevTrigStatus;
    delete ui;
    delete config;
}

void TrigMainForm::setConfigDispatcher(RcConfigDispatcher *p)
{
    ui->trigHistView->setConfigDispatcher(p);
    trigChannelConfig->setConfigDispatcher(p);

    if (configDispatcher) {
        disconnect(configDispatcher);
    }
    configDispatcher = p;
    if (! configDispatcher)
        return;
    configDispatcher->subscribe(this);
    configDispatcher->subscribeSpillConfig(ui->widgetSpill);
}

void TrigMainForm::setConfig(const RcConfig &newConfig)
{
    static RcConfig oldConfig;
    if (newConfig == oldConfig)
        return;

    *config = newConfig;
    bool trigModified = !(newConfig.rcTrigConfig == oldConfig.rcTrigConfig);
    oldConfig = newConfig;
    if (trigModified)
        restoreGuiFromConfig();
}

void TrigMainForm::setFirmwareBuildParams(const FirmwareBuildParams &fbp)
{
    ui->trigHistView->setFirmwareBuildParams(fbp);
    trigChannelConfig->setNumberOfChannels(fbp.nch);
}

void TrigMainForm::updateExtTrigGroup()
{
    int count = config->extTrigChannelCount();
    ui->checkBoxTrigIn->setText(QString("External TTL")
                                    .append((count>1) ? " 1" : ""));
    ui->checkBoxTrigIn2->setVisible(count >= 2);
    ui->checkBoxTrigIn3->setVisible(count >= 3);
    ui->checkBoxTrigIn4->setVisible(count >= 4);
}

double TrigMainForm::timeFromClocks(quint64 t)
{
    return (double)t/TrigDevice::getClockFreqHz();
}

void TrigMainForm::updateRemoteBlocking(bool remoteConnected)
{
//    restoreGuiFromConfig();
    ui->groupBoxTrigSrc->setEnabled(!remoteConnected);
    ui->groupBoxTrigState->setEnabled(!remoteConnected);
    ui->widgetSpill->updateRemoteBlocking(remoteConnected);
}

QString TrigMainForm::timeStr(quint64 t)
{
    quint64 msec = 1000.0 * timeFromClocks(t);
    quint64 sec = msec/1000;
    quint64 min = sec/60;
    quint64 h = min/60;
    QString str = QString("%1:%2:%3.%4")
            .arg(h)
            .arg(min % 60, 2, 10, QChar('0'))
            .arg(sec % 60, 2, 10, QChar('0'))
            .arg(msec % 1000, 3, 10, QChar('0'));
    return str;
}

void TrigMainForm::sendTrigStatusMetrics(const TrigStatus &prevTrigStatus, const TrigStatus &status)
{
    const QMap<QString, QString> tags({{"serial", status.getDeviceIndex().getSerialStr()}});
    // Gauges - Summary
    MetricSender::send("rc.eventNumber", status.eventNumber(), tags);

    if (status.spillStatus) {
        MetricSender::send("rc.spillNumber", status.spillStatus->count, tags);
        { // Gauges - Spill Timers
            MetricFields fields({{"current", status.spillStatus->curSpillTimeMs},
                                 {"TTL", status.spillStatus->ttlLastPeriodMs},
                                 /*{"WR", status.spill.WrLastPeriodMs}*/});
            MetricSender::send("rc.spill.period", fields, tags);
        }
    }

    { // WR
        MetricFields fields({{"other", status.wr.errors.OtherErrorCount - prevTrigStatus.wr.errors.OtherErrorCount},
                             {"code", status.wr.errors.CodeErrorCount  - prevTrigStatus.wr.errors.CodeErrorCount},
                             {"linkDown", status.wr.errors.LinkLostCount   - prevTrigStatus.wr.errors.LinkLostCount},
                             {"syncLoss", status.wr.errors.SyncLostCount   - prevTrigStatus.wr.errors.SyncLostCount}});
        MetricSender::send("rc.errors.wr", fields, tags);
    }
    // other
//    MetricSender::send("rc.temp", status.deviceAttr.temp, tags);
    {
        const auto last = status.statCounters.rc_counters;
        const auto prev = prevTrigStatus.statCounters.rc_counters;
        MetricFields fields({{"All", last.trigAll()      - prev.trigAll()},
                             {"NotRunAndNotSpill", last.trigCountNotRunAndNotSpill() - prev.trigCountNotRunAndNotSpill()},
                             {"NotRunAndSpill", last.trigCountNotRunAndSpill() - prev.trigCountNotRunAndSpill()},
                             {"RunAndNotSpill", last.trigCountRunAndNotSpill() - prev.trigCountRunAndNotSpill()},
                             {"Accepted", last.trigAccepted() - prev.trigAccepted()},
                             {"DeadtimeMissed", last.trigDeadtimeMissed() - prev.trigDeadtimeMissed()},
                             {"ReadoutMissed", last.trigReadoutMissed()   - prev.trigReadoutMissed()},
                             {"Missed", last.trigMissed()   - prev.trigMissed()}
        });
        MetricSender::send("rc.triggers", fields, tags);
    }
    {
        const auto last = status.statCounters.rc_counters;
        const auto prev = prevTrigStatus.statCounters.rc_counters;
        MetricFields fields(
                    {{"Clock", timeFromClocks(last.timeClock() - prev.timeClock())},
                     {"Run", timeFromClocks(last.timeRun() - prev.timeRun())},
                     {"Spill", timeFromClocks(last.timeSpill() - prev.timeSpill())},
                     {"Readout", timeFromClocks(last.timeReadout() - prev.timeReadout())},
                     {"Standby", timeFromClocks(last.timeStandby() - prev.timeStandby())},
                     {"Xoff", timeFromClocks(last.timeXoff() - prev.timeXoff())}});
        MetricSender::send("rc.time", fields, tags);
    }

    // per-channel counters
    for(int ch=0; ch<status.nch; ch++) {
        bool xOff = config->rcTrigConfig.trig.xOffMask&(1<<ch);
        if (!xOff)
            continue;
        // qint64 events = status.statCounters.xOffEventCnt[ch] - prevTrigStatus.statCounters.xOffEventCnt[ch];
        qint64 events = status.statCounters.rc_counters.trigAccepted() - prevTrigStatus.statCounters.rc_counters.trigAccepted();
        qint64 clocks = status.statCounters.rc_counters.cnt[RcCounters::TIME_RUN_SPILL][ch]
            - prevTrigStatus.statCounters.rc_counters.cnt[RcCounters::TIME_RUN_SPILL][ch];
        double time = timeFromClocks(clocks);
        if (events == 0 && clocks == 0) continue; // unused channel
        double eventTime = (events > 0) ? time / events : 0.;
        QString descr = config->rcTrigConfig.channelDescr[ch];
        if (descr.isEmpty()) descr = QString("%1").arg(ch);
        auto tags2 = tags;
        tags2["channel"] = QString::number(ch);
        tags2["channel_descr"] = descr;
        MetricSender::send("rc.channel.time_xoff", eventTime, tags2);
    }
}

void TrigMainForm::trigStatusUpdated(const TrigStatus &status)
{
    const auto &coreStatus = status.coreStatus;
    const auto &statCounters = status.statCounters;
    const auto &rc_counters = statCounters.rc_counters;

    QTableWidget *tableStats = ui->tableWidgetStats;
//    if (status.spill.count < prevTrigStatus->spill.count) {
//        *prevTrigStatus = TrigStatus(prevTrigStatus->getDeviceIndex());
//    }
    if (status.spillStatus) {
        if (prevTrigStatus->spillStatus) {
            auto &ss = status.spillStatus;
            auto &pss = prevTrigStatus->spillStatus;
            bool doSend = false;
            if(ss->count == pss->count) {
                if(pss->status.reg.bit.phase < ss->status.reg.bit.phase && ss->isFlushPhase())
                    doSend = true;
            } else if(ss->count == pss->count+1) {
                if(ss->isFlushPhase())
                    doSend = true;
            } else {
                doSend = true;
            }
            if(doSend) {
                sendTrigStatusMetrics(*prevTrigStatus, status);
                *prevTrigStatus = status;
            }
        } else {
            *prevTrigStatus = status;
        }
    }
    QString digitStyle = "font-weight: bold; font-family: courier; font-size: 20pt";
    {
        QString text = QString("Run    %1").arg(rcRunId.toString());
        QString curText = ui->labelRunNumber->text();
        if (text != curText) {
            ui->labelRunNumber->setStyleSheet(digitStyle);
            ui->labelRunNumber->setText(text);
        }
    }

    QString spillText;
    if (status.spillStatus)
        spillText = QString("Spill  %1").arg(status.spillStatus->count);
    ui->labelSpillNumber->setText(spillText);
    ui->labelSpillNumber->setStyleSheet(digitStyle);

    ui->labelEventNumber->setText(QString("Event  %L1").arg(status.eventNumber()));
    ui->labelEventNumber->setStyleSheet(digitStyle);

    auto deadtimeMissedText = QString("Deadtime Missed %1 %").arg(rc_counters.deadtimeMissRatio()*100.0, 0, 'f', 2);
    auto readoutMissedText = QString("Readout Missed %1 %").arg(rc_counters.readoutMissRatio()*100.0, 0, 'f', 2);
    ui->labelEfficiency->setText(deadtimeMissedText + "\n" + readoutMissedText);
    ui->labelEfficiency->setStyleSheet(digitStyle);


    TrigChannelState chState;
    chState.globalEventCount = status.eventNumber();
    chState.xOff = coreStatus.curXOffState;
    chState.timeout = coreStatus.curXOffTimeoutState;
    chState.timeoutXoff = statCounters.xOffHiTimeoutCnt;
    chState.timeoutRead = statCounters.xOffLowTimeoutCnt;
    chState.globalXoffTime = 1.0 * rc_counters.timeXoff() / TrigDevice::getClockFreqHz();
    for(int cntType : rc_counters.cnt.keys()) {
//        if (cntType % 2) continue; // not a time counter
        if (cntType != RcCounters::TIME_RUN_SPILL) continue; // make only run spill time
        for(int ch : rc_counters.cnt[cntType].keys()) {
            chState.eventCount[ch] = rc_counters.trigAccepted();
            chState.xOffTime[ch] = 1.0 * rc_counters.cnt[cntType][ch] / TrigDevice::getClockFreqHz();
            chState.xOffTimeStr[ch] = timeStr(chState.xOffTime[ch]);
        }
    }
    emit updateChState(chState);
//    emit updateChState(status.curXOffState, status.curXOffTimeoutState,
//                       xOffTimeMap, xOffTimeStrMap,
//                       status.xOffHiTimeoutCnt, status.xOffLowTimeoutCnt);

    //    tableStats->item(TRIG_STAT_COL_TRIG_IN, 0)->setText(QString("%1").arg(status.trigSrc,16,2,QChar('0')));

    const int fieldWidth = 14;
    qint64 timeXoff = rc_counters.timeXoff();
    qint64 timeReadout = rc_counters.timeReadout();
    qint64 timeStandby = rc_counters.timeStandby();
    qint64 timeRun = rc_counters.timeRun();
    tableStats->item(TRIG_STAT_COL_TIME_XOFF,    0)->setText(timeStr(timeXoff));
    tableStats->item(TRIG_STAT_COL_TIME_READOUT, 0)->setText(timeStr(timeReadout));
    tableStats->item(TRIG_STAT_COL_TIME_STBY,    0)->setText(timeStr(timeStandby));
    tableStats->item(TRIG_STAT_COL_TIME_RUN,     0)->setText(timeStr(timeRun));
//    tableStats->item(TRIG_STAT_COL_STATE, 0)->setText(stateStr);
    tableStats->item(TRIG_STAT_COL_TRIG,            0)->setText(QString("%L1").arg(rc_counters.trigAll(), fieldWidth));
    tableStats->item(TRIG_STAT_COL_TRIG_RUN,        0)->setText(QString("%L1").arg(rc_counters.trigOnRun(), fieldWidth));
    tableStats->item(TRIG_STAT_COL_TRIG_SPILL,      0)->setText(QString("%L1").arg(rc_counters.trigOnSpill(), fieldWidth));
    tableStats->item(TRIG_STAT_COL_TRIG_RUN_SPILL,  0)->setText(QString("%L1").arg(rc_counters.trigOnRunOnSpill(), fieldWidth));
    tableStats->item(TRIG_STAT_COL_TRIG_ACCEPTED,   0)->setText(QString("%L1").arg(rc_counters.trigAccepted(), fieldWidth));

    const bool hasDeadTimeControl = (status.getFwVer() >= fw_version_t(1, 0, 25367));
    ui->labelTrigSourceDeadTime->setVisible(hasDeadTimeControl);
    ui->doubleSpinBoxTrigRandomDeadTime->setVisible(hasDeadTimeControl);

    ui->labelDeadTime->setText(QString("Dead Time %1 %").arg(status.statCounters.rc_counters.deadTimeRatio()*100.0, 0, 'f', 2));
    ui->labelDeadTime->setStyleSheet(digitStyle);

    ui->widgetSpill->setSpillStatus(status.spillStatus);

    if(!status.conn.isReady()){ //todo move this check to TrigDeviceController
        qInfo() << QString("%1 Device not ready - writing config").arg(status.getDeviceIdent());
        emit resetAll();
    }
}

void TrigMainForm::clearStats()
{
    prevTrigStatus->statCounters = {};
}

void TrigMainForm::spillPlotStepUpdated()
{
//    emit configChanged(*config);
}

void TrigMainForm::setRcRunId(const RcRunId &id)
{
    rcRunId = id;
}

void TrigMainForm::setTrigSrcMask(quint16 mask, bool checked)
{
    if (checked)
        config->rcTrigConfig.trig.trigSrc |= mask;
    else
        config->rcTrigConfig.trig.trigSrc &= ~mask;
    emit configChanged(*config);
}

void TrigMainForm::restoreGuiFromConfig()
{
    MultiSignalBlocker msb(children());

    // trigChannelConfig->updateGui();

    bool oldState6 = ui->checkBoxTrigIn->blockSignals(true);
    bool oldState7 = ui->checkBoxTrigPeriod->blockSignals(true);
    bool oldState8 = ui->checkBoxTrigRandom->blockSignals(true);
    bool oldState9 = ui->doubleSpinBoxTrigTimerFreq->blockSignals(true);
    bool oldState10 = ui->doubleSpinBoxTrigRandomFreq->blockSignals(true);
    bool oldState11 = ui->doubleSpinBoxTrigRandomDeadTime->blockSignals(true);


    ui->radioButtonTrigOpen->blockSignals(true);
    if(config->rcTrigConfig.trig.trigOpen)
        ui->radioButtonTrigOpen->setChecked(true);
    else
        ui->radioButtonTrigClose->setChecked(true);
    ui->radioButtonTrigOpen->blockSignals(false);

    bool trigIn = config->rcTrigConfig.trig.trigSrc & 0x1;
    bool trigPeriodical = config->rcTrigConfig.trig.trigSrc & TrigSourceConfig::TrigMaskTimer;
    bool trigRandom = config->rcTrigConfig.trig.trigSrc & TrigSourceConfig::TrigMaskRandom;
    bool trigLvds = config->rcTrigConfig.trig.trigSrc & TrigSourceConfig::TrigMaskLvds;
    ui->checkBoxTrigIn->setChecked(trigIn);
    trigIn = config->rcTrigConfig.trig.trigSrc & 0x2;
    ui->checkBoxTrigIn2->setChecked(trigIn);
    trigIn = config->rcTrigConfig.trig.trigSrc & 0x4;
    ui->checkBoxTrigIn3->setChecked(trigIn);
    trigIn = config->rcTrigConfig.trig.trigSrc & 0x8;
    ui->checkBoxTrigIn4->setChecked(trigIn);
    ui->checkBoxTrigPeriod->setChecked(trigPeriodical);
    ui->checkBoxTrigRandom->setChecked(trigRandom);
    ui->checkBoxTrigLvds->setChecked(trigLvds);
    ui->doubleSpinBoxTrigTimerFreq->setMinimum(TrigDevice::getTrigTimerMinFreqHz());
    ui->doubleSpinBoxTrigTimerFreq->setMaximum(std::min(25e6, TrigDevice::getTrigTimerMaxFreqHz()));
    ui->doubleSpinBoxTrigTimerFreq->setValue(config->rcTrigConfig.trigCtrlStat.trigTimerFreq);
    ui->doubleSpinBoxTrigRandomFreq->setMinimum(TrigDevice::getTrigTimerMinFreqHz());
    ui->doubleSpinBoxTrigRandomFreq->setMaximum(std::min(10e6, TrigDevice::getTrigTimerMaxFreqHz()));
    ui->doubleSpinBoxTrigRandomFreq->setValue(config->rcTrigConfig.trigCtrlStat.randomPulserFreq);

    ui->doubleSpinBoxTrigRandomFreq->setMinimum(0);
    ui->doubleSpinBoxTrigRandomFreq->setMaximum(1e6);
    ui->doubleSpinBoxTrigRandomDeadTime->setValue(config->rcTrigConfig.trigCtrlStat.programmedDeadTime * scaleDeadTime);

    updateTableSize();
    updateExtTrigGroup();

    ui->checkBoxTrigIn->blockSignals(oldState6);
    ui->checkBoxTrigPeriod->blockSignals(oldState7);
    ui->checkBoxTrigRandom->blockSignals(oldState8);
    ui->doubleSpinBoxTrigTimerFreq->blockSignals(oldState9);
    ui->doubleSpinBoxTrigRandomFreq->blockSignals(oldState10);
    ui->doubleSpinBoxTrigRandomDeadTime->blockSignals(oldState11);
}

void TrigMainForm::on_pushButtonClearStat_clicked()
{
    emit clearTrigStats(false); // todo: use difference logic instead of reseting
    clearStats();
}

void TrigMainForm::on_pushButtonResetAll_clicked()
{
    int ret = QMessageBox::question(this, "Confirm ResetAll", "Are you sure to ResetAll settings in Run Control device."
//                                    ,QMessageBox::Yes
//                                    ,QMessageBox::No
                                    );
    if(ret == QMessageBox::Yes) {
        emit resetAll();
    }
}

void TrigMainForm::configUpdated(quint16 xOff)
{
    TrigSourceConfig newConfig(config->rcTrigConfig.trig);
    newConfig.xOffMask = xOff;
    if(config->rcTrigConfig.trig == newConfig)
        return;
    config->rcTrigConfig.trig = newConfig;
    emit configChanged(*config);
}

void TrigMainForm::on_radioButtonTrigOpen_toggled(bool open)
{
    config->rcTrigConfig.trig.trigOpen = open;
    emit configChanged(*config);
}

void TrigMainForm::on_doubleSpinBoxTrigTimerFreq_valueChanged(double arg1)
{
    config->rcTrigConfig.trigCtrlStat.trigTimerFreq = arg1;
    emit configChanged(*config);
}

void TrigMainForm::on_doubleSpinBoxTrigRandomFreq_valueChanged(double arg1)
{
    config->rcTrigConfig.trigCtrlStat.randomPulserFreq = arg1;
    emit configChanged(*config);
}

void TrigMainForm::on_doubleSpinBoxTrigRandomDeadTime_valueChanged(double arg1)
{
    config->rcTrigConfig.trigCtrlStat.programmedDeadTime = arg1 / scaleDeadTime;
    emit configChanged(*config);
}

void TrigMainForm::on_checkBoxTrigIn_toggled(bool checked)
{
    setTrigSrcMask(0x1, checked);
}

void TrigMainForm::on_checkBoxTrigIn2_toggled(bool checked)
{
    setTrigSrcMask(0x2, checked);
}

void TrigMainForm::on_checkBoxTrigIn3_toggled(bool checked)
{
    setTrigSrcMask(0x4, checked);
}

void TrigMainForm::on_checkBoxTrigIn4_toggled(bool checked)
{
    setTrigSrcMask(0x8, checked);
}

void TrigMainForm::on_checkBoxTrigLvds_toggled(bool checked)
{
    setTrigSrcMask(TrigSourceConfig::TrigMaskLvds, checked);
}

void TrigMainForm::on_checkBoxTrigPeriod_toggled(bool checked)
{
    setTrigSrcMask(TrigSourceConfig::TrigMaskTimer, checked);
}

void TrigMainForm::on_checkBoxTrigRandom_toggled(bool checked)
{
    setTrigSrcMask(TrigSourceConfig::TrigMaskRandom, checked);
}

void TrigMainForm::onRcFsmStateChange(FsmState st)
{
    auto en = st != FsmState::Starting &&
            st != FsmState::Stopping;
    setEnabled(en);
}

void TrigMainForm::updateTableSize()
{
    int height=ui->tableWidgetStats->horizontalHeader()->height();
    for(int i=0;i<ui->tableWidgetStats->rowCount();++i){
        height += ui->tableWidgetStats->rowHeight(i);
    }
    height += 5;
    ui->tableWidgetStats->setMinimumHeight(height);
    QRect geo = ui->tableWidgetStats->geometry();
    geo.setHeight(height);
    ui->tableWidgetStats->setGeometry(geo);
//    qDebug()<<ui->tableWidgetStats <<" height="<<height;
}
