/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "VmeTrigsController.h"

#include <QCoreApplication>
#include <QDebug>
#include <QElapsedTimer>

#include "CoreTypes.h"
#include "RcConfig.h"
#include "RcConfigDispatcher.h"
#include "VmeTrigFsm.h"
#include "VmeTrigsController_p.h"
#include "mldiscover/DeviceDispatcher.h"

static DeviceIndexSet getDeviceIndexSetFromMap(const DeviceIndexEnabledMap &map)
{
    DeviceIndexSet set;
    for (auto i : map)
        set.insert(i.index);
    return set;
}

VmeTrigsController::VmeTrigsController(QObject *parent) :
    AbstractDeviceController(parent),
    d_ptr(new VmeTrigsControllerPrivate(this))
{
    Q_D(VmeTrigsController);
    connect(this, &VmeTrigsController::printFsmInput,
            d, &VmeTrigsControllerPrivate::printFsmInput);
}

VmeTrigsController::~VmeTrigsController()
{
    delete d_ptr;
}

QSet<DeviceId> VmeTrigsController::getDeiveIdFilter()
{
    return {DEVICE_ID_VME_TRIG};
}

void VmeTrigsController::setConfig(const RcConfig &newConfig)
{
    Q_D(VmeTrigsController);

    if (newConfig == *d->config)
        return;

    const auto oldDeviceList = d->deviceListFromConfig(*d->config);
    const auto newDeviceList = d->deviceListFromConfig(newConfig);
    *d->config = newConfig;

    // update device list from config
    if (oldDeviceList != newDeviceList)
        emit deviceListChanged(newDeviceList);

    // split out config
    for (auto it = d->config->vmeTrigs.begin(); it != d->config->vmeTrigs.end(); ++it) {
        emit trigConfigChanged(it.key(), it.value());
    }
    d->sendVmeMap();
}

void VmeTrigsController::setTrigConfig(const DeviceIndex &index, const VmeTrigConfig &config)
{
    Q_D(VmeTrigsController);

    if (config == d->config->vmeTrigs[index])
        return;
    const auto oldConfig = d->config->vmeTrigs[index];
    const auto oldDeviceList = d->deviceListFromConfig(*d->config);
    d->config->vmeTrigs[index] = config;
    const auto newDeviceList = d->deviceListFromConfig(*d->config);

    emit configChanged(*d->config);

    if (oldConfig.checkEvNum != config.checkEvNum)
        emit vmeReadoutUpdated(index, config.checkEvNum);

    // update device list from config
    if (oldDeviceList != newDeviceList){
        emit deviceListChanged(newDeviceList);
    }

    if(config.enabled) {
        d->addTrigDevice(DeviceIndexEnabled(index, config.enabled));
    } else {
        d->removeTrigThread(index);
    }
    d->sendVmeMap();
}

void VmeTrigsController::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    Q_D(VmeTrigsController);

    static DeviceIndexEnabledMap old_devices;

    const auto old_set = getDeviceIndexSetFromMap(old_devices);
    const auto new_set = getDeviceIndexSetFromMap(devices);
    const auto toRemove = old_set - new_set;
    const auto toAdd = new_set - old_set;

    for (auto i : toRemove)
        d->removeTrigDevice(i);

    for (auto i : toAdd)
        d->addTrigDevice(devices[i]);

    for (auto i : (old_set & new_set)) {
        if(devices[i].enabled) {
            auto &cfg = d->config->vmeTrigs[i];
            cfg.enabled = true;
            d->addTrigThread(i, cfg);
        } else {
            d->removeTrigThread(i);
        }
    }

    for (auto i : devices) {
            d->config->vmeTrigs[i.index].enabled = i.enabled;
            emit configChanged(*d->config);
            emit trigConfigChanged(i.index, d->config->vmeTrigs[i.index]);
    }

    old_devices = devices;
}

void VmeTrigsController::onRunStart(const RcRunId &runId)
{
    Q_D(VmeTrigsController);

    qDebug()<<"Start Vme Trig resetting...";
    d->last_got_run_id = runId;
    d->fsm->setRequestState(FsmState::Starting);
}

void VmeTrigsController::onRunStop()
{
    Q_D(VmeTrigsController);
    d->fsm->setRequestState(FsmState::Stopping);
}

void VmeTrigsController::pnpCacheUpdated(const PnpInfoCache &cache)
{
    // unused at this moment
    return;

    // map from TMWR_devIndex to "<VmeDaqHost>:<VmeDaqInterfaceNumber>";
    QMap<DeviceIndex, QString> trigHostMap;
    for(auto i=cache.cbegin(); i!=cache.cend(); ++i) {
        const auto &pd = *i;
        if(!(pd.getDaqModule() == DaqModule::Type::VmeDaq))
            return;

        bool ok;
        for(int i=0; i<24; ++i) {
            const QString optKeyId = QString("slot%1_id").arg(i, 2, 10, QChar('0'));
            const QString optKeySerial = QString("slot%1_serial").arg(i, 2, 10, QChar('0'));
            if(!pd.options.contains(optKeyId))
                continue;
            quint16 id = pd.options[optKeyId].toUShort(&ok, 16);
            if(!ok || id != DEVICE_ID_VME_TRIG)
                continue;
            if(!pd.options.contains(optKeySerial))
                continue;
            quint64 ser = pd.options[optKeySerial].toULongLong(&ok, 16);
            if(!ok)
                continue;
            if(!pd.options.contains("interfaceIndex"))
                continue;
            quint16 interfaceIndex = pd.options["interfaceIndex"].toUShort(&ok, 10);
            if(!ok)
                continue;
            DeviceIndex index(static_cast<DeviceId>(id), ser);
            trigHostMap[index] = QString("%1:%2").arg(pd.hostName).arg(interfaceIndex);
            break;
        }
    }
}
