/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "VmeTrigsModel.h"

#include <QColor>
#include <QHostAddress>
#include <QPalette>

#include "app/AppStyle.h"
#include "mldiscover/LldpDiscovery.h"
#include "VmeTrigDevice.h"
#include "RcConfig.h"

struct VmeTrigTableEntry
{
    VmeTrigConfig config;
    VmeTrigStatus status;
};

class VmeTrigTable : public QMap<DeviceIndex, VmeTrigTableEntry>
{
public:
    int rowOf(const DeviceIndex &deviceIndex) const;
    DeviceIndex indexOf(int row) const;
};

int VmeTrigTable::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(begin(), find(deviceIndex));
}

DeviceIndex VmeTrigTable::indexOf(int row) const
{
    int r = 0;
    for (auto it = begin(); it != end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

VmeTrigsModel::VmeTrigsModel(QObject *parent)
    : QAbstractTableModel(parent),
    lldpInfoCache(new LldpInfoCache()),
    devices(new VmeTrigTable())
{

}

VmeTrigsModel::~VmeTrigsModel()
{
    delete lldpInfoCache;
    delete devices;
}

void VmeTrigsModel::invalidateRow(const DeviceIndex &index)
{
    int row = devices->rowOf(index);
    if (row == -1)
        return;

    auto indexFrom = createIndex(row, 0);
    auto indexTo = createIndex(row, columnCount()-1);
    emit dataChanged(indexFrom, indexTo);
}

void VmeTrigsModel::lldpCacheUpdated(const LldpInfoCache &lldp)
{
    if (lldp == *lldpInfoCache)
        return;
    beginResetModel();
    *lldpInfoCache = lldp;
    endResetModel();
}

void VmeTrigsModel::setTrigConfig(const DeviceIndex &index, const VmeTrigConfig &config)
{
    if (config == (*devices)[index].config)
        return;
    (*devices)[index].config = config;
    invalidateRow(index);
}

void VmeTrigsModel::setTrigStatus(const DeviceIndex &index, const VmeTrigStatus &status)
{
    if (status == (*devices)[index].status)
        return;
    (*devices)[index].status = status;
    invalidateRow(index);
}

void VmeTrigsModel::addDevice(const DeviceIndex &index)
{
    beginResetModel();
    devices->insert(index, VmeTrigTableEntry());
    endResetModel();
//    invalidateRow(index);
}

void VmeTrigsModel::removeDevice(const DeviceIndex &index)
{
    beginResetModel();
    devices->remove(index);
    endResetModel();
}

void VmeTrigsModel::setEnabled(bool en)
{
    enabled = en;

    auto indexFrom = createIndex(0, 0);
    auto indexTo = createIndex(rowCount()-1, columnCount()-1);
    emit dataChanged(indexFrom, indexTo);
}

int VmeTrigsModel::rowCount(const QModelIndex &parent) const
{
    return  devices->size();
}

int VmeTrigsModel::columnCount(const QModelIndex &parent) const
{
    return COL_VME_TOTAL;
}

Qt::ItemFlags VmeTrigsModel::flags(const QModelIndex & index) const
{
    auto f = QAbstractTableModel::flags(index);
    switch (index.column()) {
    case COL_ENABLE:
    case COL_VME_CHECK_EV_NUM:
    case COL_VME_CHECK_WR:
        if(enabled)
            f |= Qt::ItemIsUserCheckable;
        else
            f &= ~Qt::ItemIsEnabled;
    default:
        return f;
    }
}

QVariant VmeTrigsModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return QVariant();

    switch (role) {
    // case Qt::EditRole: return column == COL_ENABLE || column == COL_VME_CHECK_WR;
    case Qt::BackgroundRole: return data_BackgroundRole(deviceIndex, column);
    case Qt::CheckStateRole: return data_CheckStateRole(deviceIndex, column);
    case Qt::DisplayRole: return data_DisplayRole(deviceIndex, column);
    }
    return QVariant();
}

QVariant VmeTrigsModel::data_BackgroundRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    const auto &config = device.config;
    const auto &status = device.status;

    if(!config.enabled)
        return {};

    switch (column) {
    case COL_VME_ID:
    case COL_VME_SERIAL:
        return status.connected
                ? QVariant() : AppStyle::tableBg(Severity::Error);
    case COL_VME_WR_TIME_VALID:
        return (!config.checkWr || status.isWrTimeValid())
                ? QVariant() : AppStyle::tableBg(Severity::Error);
    }
    return QVariant();
}

QVariant VmeTrigsModel::data_CheckStateRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    const auto &config = device.config;
    // const auto &status = device.status;

    switch (column) {
    case COL_ENABLE: return config.enabled ? Qt::Checked : Qt::Unchecked;
    case COL_VME_CHECK_EV_NUM: return config.checkEvNum ? Qt::Checked : Qt::Unchecked;
    case COL_VME_CHECK_WR: return config.checkWr ? Qt::Checked : Qt::Unchecked;
    }
    return QVariant();
}

QVariant VmeTrigsModel::data_DisplayRole(const DeviceIndex &index, int column) const
{
    const auto &device = (*devices)[index];
    const auto &status = device.status;
    const auto &lldpInfo = (*lldpInfoCache)[index];

    switch (column) {
    // case COL_ENABLE: return config.enabled;
    case COL_VME_ID: return index.getDeviceIdName();
    case COL_VME_SERIAL: return index.getSerialStr();
    // case COL_VME_READOUT: return config.readout;
    case COL_VME_HOST: return lldpInfo.ext.master_locked ? lldpInfo.ext.master.hostAddress().toString() : QVariant();
    case COL_VME_TEMP: return TempsInfo(status.getTemps()).brief(); // .append(QChar(0x02da)).append('C');
    case COL_VME_EV_NUM: return QString("%L1").arg(status.coreStatus.evNum);
    case COL_VME_WR_TIME_VALID: return (status.isWrTimeValid() ? "Ok" : "Invalid");
    case COL_VME_WR_SYNC_LOST: if (status.wrStatus.has_value()) return QString::number(status.wrStatus->syncFail);
    case COL_VME_WR_LINK_ERROR: if (status.wrStatus.has_value()) return (QString::number(status.wrStatus->linkError));
    case COL_VME_BUSY: return QString("%L1").arg(status.coreStatus.busyTimer);
    case COL_VME_IDLE: return QString("%L1").arg(status.coreStatus.idleTimer);
    case COL_VME_OCCUPANCY: return QString::number(100.*status.coreStatus.occupancy());
    }
    return {};
}

QVariant VmeTrigsModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<Columns>(section)) {
        case COL_ENABLE: return tr("En");
        case COL_VME_ID: return tr("ID");
        case COL_VME_SERIAL: return tr("Serial");
        case COL_VME_CHECK_EV_NUM: return tr("CheckEvNum");
        case COL_VME_HOST: return tr("Host");
        case COL_VME_TEMP: return tr("Temp");
        case COL_VME_EV_NUM: return tr("evNum");
        case COL_VME_CHECK_WR: return tr("Check WR");
        case COL_VME_WR_TIME_VALID: return tr("WR Time");
        case COL_VME_WR_SYNC_LOST: return tr("WR Sync Lost");
        case COL_VME_WR_LINK_ERROR: return tr("WR Link Error");
        case COL_VME_BUSY: return tr("Busy");
        case COL_VME_IDLE: return tr("Idle");
        case COL_VME_OCCUPANCY: return tr("Ocupancy");
        case COL_VME_TOTAL: return {};
        }
    }
    // TODO return DeviceIndexTableModel::headerData(section, orientation, role);
    return {};
}

bool VmeTrigsModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return false;

    auto &device = (*devices)[deviceIndex];
    auto &config = device.config;

    if (role != Qt::CheckStateRole)
        return false;

    switch (column) {
    case COL_ENABLE:
        config.enabled = (value == Qt::Checked);
        break;
    case COL_VME_CHECK_EV_NUM:
        config.checkEvNum = (value == Qt::Checked);
        break;
    case COL_VME_CHECK_WR:
        config.checkWr = (value == Qt::Checked);
        break;
    default:
        return false;
    }

    emit dataChanged(index, index);
    emit trigConfigChanged(deviceIndex, config);
    return true;
}
