#include "CalibrationModel.h"
#include "DeviceChModelItem.h"

#include <QFont>
#include <QBrush>
#include <QDebug>

namespace {
const int RMS_LIMIT = 20; // in adc64 100

QString getThrText(int dacValue)
{
    return QString("%1").arg(dacValue);
}
}

CalibrationModel::CalibrationModel(QObject *parent)
    :QAbstractItemModel(parent)
    , rootItem(new DeviceChModelItem())

{}

void CalibrationModel::clear()
{
    for(auto die : qAsConst(map)){
        if(die.enabled) {
            auto cal = CalResult();
            cal.index = die.index;
            cal_config[die.index] = cal;
        }
    }
    resetItems();
    emit calConfigChanged(cal_config);
}

void CalibrationModel::reset()
{
    resetItems();
}

void CalibrationModel::removeDev(const DeviceIndex &index)
{
    auto newCalConfig = cal_config;
    newCalConfig.remove(index);
//    setCalConfig(newCalConfig);
}

QVariant CalibrationModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();

    switch (role) {
    case Qt::TextAlignmentRole:
        if (index.column() > 0)
            return Qt::AlignRight;
        break;
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::ForegroundRole:
        return getForegroundData(index);
    case Qt::FontRole:
        return fontData(index);
    }
    return QVariant();
}

QVariant CalibrationModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return QVariant();
    const auto col = static_cast<TreeColumns>(section);
    if(role == Qt::DisplayRole) {
        switch (col) {
        case TreeColumns::NAME:
            return "Device / ch";
        case TreeColumns::BASELINE:
            return "Baseline";
        case TreeColumns::BASELINE_RMS:
            return "Baseline rms";
        case TreeColumns::BASELINE4:
            return "Baseline x4";
        case TreeColumns::BASELINE4_RMS:
            return "Baseline x4 rms";
        case TreeColumns::DAC:
            return "Analog Thr. (mV)";
        default:
            return QVariant();
        }
    }
    if(role == Qt::ForegroundRole) {
        if(col == TreeColumns::BASELINE_RMS ||
                col == TreeColumns::BASELINE4_RMS)
            return getHeaderForeground(col);
    }

    return QVariant();
}

QModelIndex CalibrationModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const DeviceChModelItem *parentItem = parent.isValid() ?
                static_cast<DeviceChModelItem*>(parent.internalPointer()) :
                rootItem;
    DeviceChModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex CalibrationModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const DeviceChModelItem *childItem = static_cast<DeviceChModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    DeviceChModelItem *parentItem = childItem->parent;
    if(parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int CalibrationModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
                static_cast<DeviceChModelItem*>(parent.internalPointer()) : rootItem;
    return parentItem->childCount();
}

void CalibrationModel::setCalConfig(const ConfigKey &config_key, const CalConfig &newCalConfig)
{
    Q_UNUSED(config_key);
    if(cal_config == newCalConfig)
        return;

    bool performReset = cal_config.keys() != newCalConfig.keys();
    if(!performReset) {
        for(auto &cr : newCalConfig)
            performReset |= cal_config[cr.index].nch != cr.nch;
    }

    cal_config = newCalConfig;

    if(performReset) {
        resetItems();
    } else {
        auto top = index(0,0);
        auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
        auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
        emit dataChanged(top, lastCh);
    }
}

void CalibrationModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    if(this->map == map)
        return;
    this->map  = map;
    resetItems();
}

QVariant CalibrationModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    std::optional<double> opt;
    switch (col) {
    case TreeColumns::NAME:
        if(item->dev.isNull()) {
            return "All";
        } else if(item->ch == -1) {
            return item->dev.getSerialStr();
        } else {
            return QString::number(item->ch);
        }
    case TreeColumns::DAC:
        opt = cal_config.getValue(CalConfig::Type::Dac, item->dev, item->ch);
        if (!opt)
            return {};
        return getThrText(*opt);

    case TreeColumns::BASELINE:
        opt = cal_config.getValue(CalConfig::Type::Baseline, item->dev, item->ch);
        break;
    case TreeColumns::BASELINE_RMS:
        opt = cal_config.getValue(CalConfig::Type::Rms, item->dev, item->ch);
        break;
    case TreeColumns::BASELINE4:
        opt = cal_config.getValue(CalConfig::Type::Baseline4, item->dev, item->ch);
        break;
    case TreeColumns::BASELINE4_RMS:
        opt = cal_config.getValue(CalConfig::Type::Rms4, item->dev, item->ch);
        break;
    default:
        return {};
    }
    if (!opt)
        return {};
    return QString("%1").arg(*opt, 6, 'f', 1);
}

QVariant CalibrationModel::getForegroundData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    if (!item || item->dev.isNull())
        return {};
    auto found = cal_config.find(item->dev);
    if (found == cal_config.end())
        return {};
    const auto ch = item->ch;
    const auto &cal = *found;
    if (ch == -1) {
        bool ok = isDeviceRmsOk(item->dev);
        QBrush b;
        b.setColor(ok ? Qt::black : Qt::red);
        return b;
    }

    switch (col) {
    case TreeColumns::BASELINE_RMS: {
        if (cal.baseline_x1[ch].stddev > RMS_LIMIT)
            return QBrush(Qt::red);
        break;
    }
    case TreeColumns::BASELINE4_RMS:
        if (cal.baseline_x4[ch].stddev > RMS_LIMIT)
            return QBrush(Qt::red);
        break;
    default:;
    }
    return {};
}

QVariant CalibrationModel::fontData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    QFont font;
//    if(item->dev.isNull()) {
//        font.setItalic(true);
//        font.setBold(true);
//    } else
        if(item->ch == -1) {
        font.setBold(true);
    } else {
//            fg.setItalic(true);
    }
        return font;
}

QVariant CalibrationModel::getHeaderForeground(CalibrationModel::TreeColumns col) const
{
//    QMap<int, double> CalResult::* ptr = nullptr;
//    if(col == TreeColumns::BASELINE_RMS)
//        ptr = &CalResult::rms;
//    else if(col == TreeColumns::BASELINE4_RMS)
//        ptr = &CalResult::rms4;
//    else
//        return QVariant();
//    bool ok = true;
//    for(auto &cr : cal_config) {
//        for(auto rms : cr.*ptr){
//            if(rms > RMS_LIMIT) {
//                ok = false;
//                break;
//            }
//        }
//        if(!ok)
//            break;
//    }
//    return QBrush (ok ? Qt::black : Qt::red);
    return {};
}

bool CalibrationModel::isDeviceRmsOk(const DeviceIndex &index) const
{
    Q_ASSERT(cal_config.contains(index));
    if (!cal_config.contains(index))
        return false;
    auto &cr = cal_config[index];
    for (auto v : cr.baseline_x1.rmsMap()) {
        if (v > RMS_LIMIT)
            return false;
    }
    for (auto v : cr.baseline_x4.rmsMap()) {
        if (v > RMS_LIMIT)
            return false;
    }
    return true;
}

CalibrationModel::TreeColumns CalibrationModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<TreeColumns>(index.column());
}

DeviceChModelItem *CalibrationModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<DeviceChModelItem*>(index.internalPointer());
}

int CalibrationModel::getDevRow(const DeviceIndex &di)
{
    auto allItem = rootItem->children.first();
    if(allItem) {
        for(int r=0; r<allItem->children.size(); ++r)
            if(allItem->children[r]->dev == di)
                return r;
    }
    return -1;
}

void CalibrationModel::resetItems()
{
    beginResetModel();
    rootItem->children.clear();

    for(auto devIt=cal_config.cbegin();devIt!=cal_config.cend(); ++devIt) {
        if(map.contains(devIt.key())){
            if(map[devIt.key()].enabled){
                auto devItem = new DeviceChModelItem(rootItem);
                devItem->dev = devIt.key();
                for(int ch=0; ch<devIt->nch; ++ch) {
                    auto chItem = new DeviceChModelItem(devItem);
                    chItem->ch = ch;
                    devItem->addChild(chItem);
                }
                rootItem->addChild(devItem);
            }
        }
    }
    endResetModel();
}
