#include "TrcDevice.h"

#include <QDateTime>
#include <QElapsedTimer>
#include <QFile>
#include <QTextStream>
#include <QVector>
#include <math.h>

#include "LegacyDeviceModule.h"
#include "TrcCore/PhysicalTrigger/PhysicalTriggerModule.h"
#include "TrcCore/TrcErc/TrcErcModule.h"
#include "TrcRegs.h"
#include "regio/regop.h"

TrcDevice::TrcDevice(const DeviceIndex &index, QObject *parent)
    : ModularMregDevice(index, status, parent)
    , status(index)
    , legacyDevice (createModule<LegacyDeviceModule>())
    , phyTrigMod (createModule<PhysicalTriggerModule>())
    , evReadoutMod (createModule<TrcErcModule>())
{
    qRegisterMetaType<ModularDeviceConfig>();
    qRegisterMetaType<BaseDevCountersMap>();
    setObjectName(getIdent());
}

bool TrcDevice::writeConfig()
{
    phyTrigMod->writeConfig(config.trcCore.phyTrig);
    evReadoutMod->writeConfig(config.trcCore.phyTrig.oscEn);
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg());

    return legacyDevice->regOpExecRebased(r);
}

void TrcDevice::resetAll()
{
    if(!isConnected())
        return;
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, 1);
    r.RegWrite16(REG_DEVICE_CTRL, 0);
    legacyDevice->regOpExecRebased(r);
}

void TrcDevice::softClear()
{
    if(!isConnected())
        return;

    phyTrigMod->softClear();
}

bool TrcDevice::readStatus()
{
    return
        ModularMregDevice::readStatus() &&
        readTrcStatus();
}

bool TrcDevice::updateCounters()
{
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg());
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg() | REG_DEVICE_CTRL_BIT_LOCK_COUNTERS);
    return regOpExec(r);
}

bool TrcDevice::readTrcStatus()
{
    if (!updateCounters())
        return {};

    if (!phyTrigMod->readStatus(&status.phyTrig))
        return {};
    quint64 globEvNum;
    if (!evReadoutMod->readGlobEvNum(&globEvNum))
        return {};
    status.globEvNum = globEvNum;

    emit statusUpdated(status);
    emit countersUpdated(status.phyTrig.rcCounters);
    return true;
}

void TrcDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();
    resetAll();

    auto bp = phyTrigMod->getBuildParam();
    emit deviceInfoUpdated(getDeviceIndex(), bp.auxChNum, bp.phyChNum);
}

quint16 TrcDevice::getDeviceCtrlReg() const
{
    return configCache.trcCore.phyTrig.trig.trigOpen ? REG_DEVICE_CTRL_BIT_RUN : 0;
}

void TrcDevice::preRunReset(const QVariant &startEvNum)
{
    ModularMregDevice::preRunActions(startEvNum);

    clearStatistics();
    bool ok;
    quint64 evNum = startEvNum.toULongLong(&ok);
    if(!ok)
        evNum = 1;
    evReadoutMod->setGlobEvNum(evNum);
}

void TrcDevice::clearStatistics()
{
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg() | REG_DEVICE_CTRL_BIT_SOFTCLEAR);
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg());
    legacyDevice->regOpExecRebased(r);
    status.statistic.clear();
}
