#include "TtvxsMainWindow.h"
#include "ui_TtvxsMainWindow.h"

#include <QDateTime>
#include <QDebug>
#include <QHostInfo>
#include <QJsonDocument>
#include <QScrollBar>
#include <QShortcut>
#include <QStyle>
#include <QTextCharFormat>
#include <QThread>
#include <QTimer>

#include "FeLinkDialog/FeLinkStatusDialog.h"
#include "SnmpDialog.h"
#include "TtvxsAppConfigDispatcher.h"
#include "TtvxsConfigDb.h"
#include "TtvxsDevice.h"
#include "TtvxsDeviceGlobal.h"
#include "TtvxsParams.h"
#include "base/AbstractConfigConverter.h"
#include "base/DeviceId.h"
#include "base/Globals.h"
#include "base/RunDocument.h"
#include "base/ThreadController.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PnpDiscovery.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "run-control-fsm/TrigDeviceMgrFsm.h"
#include "run-control-fsm/RcFsm.h"
#include "run-control-client-manager/RcClientManager.h"
#include "run-control-client-manager/gui/RcClientManagerWidget.h"
#include "widgets/DeviceStatusDialog.h"

namespace {

const int MAX_SLOT_NUMBER = 21;
const int CRATE_SLOT_ZERO = 10; // slot number for ch:0
const int CONFIG_TIMEOUT = 4000;

enum TAB_INDEX { TAB_TRIG_DEV=0, TAB_CLIENTS=1, TAB_MSTREAM=2, TAB_LOGS=3};

enum {
    ROW_NAME,ROW_STATUS,
    ROW_DEV_FW, ROW_DEV_MCU_VER, ROW_DEV_TEMP,
    ROW_DEV_STREAM, ROW_DEV_MS_ERR, ROW_WR_TIME,
    ROW_SPACER_0,
    ROW_UP_TIME, ROW_CUR_EV_NUM, ROW_LEMO_CNT, ROW_LEMO_RATE, ROW_SFP_CNT, ROW_SFP_RATE,
    ROW_SPACER_1,
    ROW_RUN_TIME, ROW_RUN_TRIG_CANDID, ROW_RUN_TRIG_ACCEPTED, ROW_RUN_TRIG_RATE,
    ROW_SPACER_2,
    ROW_CH_SLOT_FIRST,
    ROW_CH_SLOT_LAST = ROW_CH_SLOT_FIRST+(TTVXS_CH_NUM-1)+3, // 3 = 1st slot + two in middle
    ROW_CH_CNT
};
const QSet<int> SPACER_ROWS = {ROW_SPACER_0,ROW_SPACER_1,ROW_SPACER_2};

int chToSlot(int ch)
{
    bool odd = ch&1;
    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;

    int slot = (ch&1) ? (slotBase+(ch>>1)) : (slotBase-(ch>>1));
    if(slot<0 || slot>MAX_SLOT_NUMBER) return -1;

    return slot;
}

int slotToCh(int slot)
{
    if(slot>MAX_SLOT_NUMBER) return -1;
    if(slot>CRATE_SLOT_ZERO && slot<CRATE_SLOT_ZERO+3) return -1;
    bool odd = slot>CRATE_SLOT_ZERO;

    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;
    int ch = qAbs(slotBase-slot);
    ch <<=1;
    if(odd)
        ch |= 1;
    return ch;
}
void setLabelColor(QLabel * const lab, QColor col) {
    QPalette p = lab->palette();
    p.setColor(lab->foregroundRole(), col);
    lab->setPalette(p);
}
}

TtvxsMainWindow::TtvxsMainWindow(QWidget *parent) :
    BaseMainWindow(parent)
    , ui(new Ui::TtvxsMainWindow)
    , configuration_manager(new ConfigurationManager(this))
    , configDb(new TtvxsConfigDb(this))
    , discoverDialog(new DiscoverDialog(this))
    , deviceStatusDialog(new DeviceStatusDialog(this))
    , feLinkStatusDialog(new FeLinkStatusDialog(this))
    , clientCore(new RcClientManager(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , remoteControlServer(new RemoteControlServer(this, 0))
    , remoteConnected(false)
    , evNumChecker(new EvNumChecker(this))
    , writeRunHistoryTimer(new QTimer(this))
    , rc_fsm(new RcFsm(this))
    , trig_fsm(new TrigDeviceMgrFsm(this))
    , labelFsmStateRc(new QLabel(this))
    , labelFsmStateTrig(new QLabel(this))
    , labelFsmStateClient(new QLabel(this))
{
    qRegisterMetaType<TtvxsStatus>();
    qRegisterMetaType<QMap<int, FeLinkStatus> >();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<ModularDeviceConfig>();
    qRegisterMetaType<RUN_STATES>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<RunDocument>();
    AbstractConfigConverter<RunControlAppConfig>::registerTypes();


    AbstractConfigConverter<DeviceIndex>::registerTypes();
    static const constexpr char* k_global = "global";
    QMap<QVariant, QString> test;

    QVariant kk = QVariant::fromValue(DeviceIndex());
    test[kk] = "first";
    kk = QVariant::fromValue(QString());
    test[kk] = "second";
    //QVariant kz = QVariant::fromValue(QString("bar"));

//    qDebug() << operator==(QVariant::fromValue(di) , QVariant::fromValue(k_global));
    qInfo() << test;
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);
    labelFsmStateRc->setMinimumWidth(150);
    labelFsmStateTrig->setMinimumWidth(150);
    labelFsmStateClient->setMinimumWidth(150);
    statusBar()->addWidget(labelFsmStateRc);
    statusBar()->addWidget(labelFsmStateTrig);
    statusBar()->addWidget(labelFsmStateClient);

    widgets_to_block.append(ui->checkBox_external_ttl);
    widgets_to_block.append(ui->checkBox_timer);
    widgets_to_block.append(ui->checkBox_random);
    widgets_to_block.append(ui->spinBox_timer);
    widgets_to_block.append(ui->spinBox_random);
    widgets_to_block.append(ui->doubleSpinBox_dead_time);
    widgets_to_block.append(ui->checkBox_aux_cnt_for_trig);
    widgets_to_block.append(ui->groupBoxStat);
    widgets_to_block.append(ui->spinBoxStatPeriod);
    widgets_to_block.append(ui->comboBoxPreset);
    widgets_to_block.append(ui->actionAllow_Remote_Control);
    widgets_to_block.append(ui->radioButton_open);

    init_GUI_style();
    init_GUI_default_settings();
    initTableDev();
    initStatPresets();

    connect(writeRunHistoryTimer, &QTimer::timeout,
            this, &TtvxsMainWindow::sendRunDocument);
    writeRunHistoryTimer->setSingleShot(true);
    writeRunHistoryTimer->setInterval(CONFIG_TIMEOUT);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter({DEVICE_ID_TTVXS});
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(discoverDialog);
    lldpDiscovery.subscribe(this);

    auto configDispatcher = &TtvxsAppConfigDispatcher::instance();
    deviceDispatcher->subscribeDevices(discoverDialog);
    deviceDispatcher->subscribeDevices(configDispatcher);
    deviceDispatcher->subscribeDevices(trig_fsm);
    deviceDispatcher->subscribeDevices(feLinkStatusDialog);
    deviceDispatcher->subscribeDevices(ui->statistic_widget);

    configDispatcher->subscribeConfig(this);
    configDispatcher->subscribeConfig(ui->groupBoxBusySource);
    configDispatcher->subscribeCS(this);
    configDispatcher->subscribeConfig(configDb);
    configDispatcher->subscribeCS(configDb);
    configDispatcher->subscribeCS(configuration_manager);
    configDispatcher->subscribeRcClientConfigMap(clientCore);
    configDispatcher->subscribeRcClientConfigMap(ui->clientManagerWidget);
    {
        auto model = feLinkStatusDialog->getModel();
        connect(model, &FeLinkTreeModel::configChanged,
                configDispatcher, &TtvxsAppConfigDispatcher::setConfigFeLink);
        connect(configDispatcher, &TtvxsAppConfigDispatcher::configChanged,
                model, &FeLinkTreeModel::setConfig);
    }


    connect(rc_fsm, &RcFsm::stateChanged,
            this, &TtvxsMainWindow::onRcFsmStateChange);
//    connect(configDispatcher, &TtvxsAppConfigDispatcher::deviceListChanged,
//            trig_fsm, &TrigDeviceMgrFsm::setDeviceList);
    connect(trig_fsm, &TrigDeviceMgrFsm::stateChanged,
            this, &TtvxsMainWindow::onTrigFsmStateChange);
    connect(configDispatcher, &TtvxsAppConfigDispatcher::configChanged,
            trig_fsm, &DeviceMgrFsm::setMainSetupKey);

    {
        // ClientManagerWidget
        connect(clientCore, &RcClientManager::runStateChanged,
                this, &TtvxsMainWindow::clientManagerRunStateChanged);

        connect(clientCore, &RcClientManager::runWarn,
                this, &TtvxsMainWindow::onRunWarning);
        connect(clientCore, &RcClientManager::runError,
                this, &TtvxsMainWindow::onRunError);
        connect(clientCore, &RcClientManager::backPressure,
                this, &TtvxsMainWindow::onBackPressure);
        connect(clientCore, &RcClientManager::config_transfer,
                this, &TtvxsMainWindow::config_transfer_from_clients);

        connect(this, &TtvxsMainWindow::restartAll,
                clientCore, &RcClientManager::onRestartAll);
        connect(this, &TtvxsMainWindow::stopAll,
                clientCore, &RcClientManager::onStopAll);

        connect(evNumChecker, &EvNumChecker::evNumRequest,
                clientCore, &RcClientManager::evNumRequest);
        connect(evNumChecker, &EvNumChecker::checkFail, this, &TtvxsMainWindow::onEvNumCheckFail);
        connect(clientCore, &RcClientManager::evNumResponse,
                evNumChecker, &EvNumChecker::gotClientEvNum);
        connect(clientCore, &RcClientManager::clientMapUpdated,
                evNumChecker, &EvNumChecker::updateClientMap);

        PnpDiscovery::subscribeCacheUpdaed(clientCore);

        connect(clientCore, &RcClientManager::clientInfoUpdated,
                ui->clientManagerWidget, &RcClientManagerWidget::clientInfoUpdated);

        connect(clientCore, &RcClientManager::runStateChanged,
                ui->clientManagerWidget, &RcClientManagerWidget::onFsmStateChange);

        connect(clientCore, &RcClientManager::runStateChanged,
                ui->clientManagerWidget, &RcClientManagerWidget::onFsmStateChange);
    }

    PnpDiscovery::subscribeMyProgram(this);

    {
        auto mainWindow = this;
        connect(remoteControlServer, &RemoteControlServer::startRequest,
                mainWindow, &TtvxsMainWindow::onRemoteStartRequest);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                mainWindow, &TtvxsMainWindow::onRemoteStopRequest);
        connect(mainWindow, &TtvxsMainWindow::sendRemoteState,
                remoteControlServer, &RemoteControlServer::sendState);
        connect(rc_fsm, &RcFsm::stateChanged,
                remoteControlServer, &RemoteControlServer::onFsmStateChange);
        connect(mainWindow, &TtvxsMainWindow::sendRunError,
                remoteControlServer, &RemoteControlServer::onRunError);
        connect(mainWindow, &TtvxsMainWindow::sendRunWarn,
                remoteControlServer, &RemoteControlServer::onRunWarning);
        connect(mainWindow, &TtvxsMainWindow::sendBackPressure,
                remoteControlServer, &RemoteControlServer::onBackPressure);

        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                mainWindow, &TtvxsMainWindow::onRemoteStatusUpdated);
        connect(mainWindow, &TtvxsMainWindow::allowRemoteControl,
                remoteControlServer, &RemoteControlServer::startListen);

        connect(remoteControlServer, &RemoteControlServer::evNumCheckRequest,
                evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);
        connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
                remoteControlServer, &RemoteControlServer::evNumCheckResponse);

        connect(mainWindow, &TtvxsMainWindow::sendRunDocumentToRc,
                remoteControlServer, &RemoteControlServer::config_transfer);
    }

    connect(ui->actionPrint_FSM_inputs, &QAction::triggered,
            this, &TtvxsMainWindow::printFsmInput);
    connect(this, &TtvxsMainWindow::printFsmInput,
            rc_fsm, &RcFsm::printFsmInput);
    connect(this, &TtvxsMainWindow::printFsmInput,
            trig_fsm, &TrigDeviceMgrFsm::printFsmInput);
    connect(this, &TtvxsMainWindow::printFsmInput,
            clientCore, &RcClientManager::printFsmInput);

    new ThreadController(remoteControlServer, this);

    updateGui();

    configuration_manager->setDefaultConfig(RunControlAppConfig());
}

TtvxsMainWindow::~TtvxsMainWindow()
{
    disconnectFromDev(DeviceIndex());

    delete ui->clientManagerWidget;
    delete ui;
}

//=== Public slots:
void TtvxsMainWindow::setConfig(ConfigKey config_key, const RunControlAppConfig &config_)
{
    Q_UNUSED(config_key)
    if(this->config == config_)
        return;
    auto newSet = config_.getDeviceIndexEnabledMap().getEnabled();
    auto oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    config = config_;
    emit allowRemoteControl(config.allowRemoteControl);
    updateGui();

    //remove from devices
    for(auto index : oldSet-newSet) {
        disconnectFromDev(index);
    }

    //add devices
    auto addSet = newSet-oldSet;
    for(auto index : addSet) {
        addDevice(index);
    }
    if(!addSet.isEmpty())
        TtvxsAppConfigDispatcher::instance().sendCache();
}

void TtvxsMainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    BaseMainWindow::configSelectorChanged(conf);

    if(localProgramIndex != conf.getIndex()) {
        localProgramIndex = conf.getIndex();
        ClientIndex clIndex{Globals::getDaqModule(), localProgramIndex};
        emit closeProgram(progDesc);
        progDesc.setClientIndex(clIndex);
        emit updateProgramDescription(progDesc);
    }

    // Initialize MetricSender
    MetricSender::setBaseTag("index", localProgramIndex.toString());
}

void TtvxsMainWindow::lldpCacheUpdated(const LldpInfoCache &lldpCache)
{
    const auto lldp_keys = lldpCache.keys();
    const auto lldp_indices = makeDeviceIndexSet(lldp_keys);
    const auto my_keys = config.knownConfigs.keys();
    const auto my_indices = makeDeviceIndexSet(my_keys);
    const auto indices = lldp_indices & my_indices;

    for (const auto &i : indices)
        updateDeviceLldpInfo(i, lldpCache[i]);
}

void TtvxsMainWindow::onRcFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    logRcFsmStateChange(state, oldState);
    labelFsmStateRc->setText("RcFsm: " + fsmStateToString(state));
    updateStatStopButton();
    updateControlEnable();
    emit set_run_state(!isReadyErrorSt());

    if(state == FsmState::Error && (oldState != FsmState::Ready && oldState != FsmState::Init))
        emit printFsmInput();

    switch (state) {
    case FsmState::CloseGlobTrig:
        trig_fsm->setRequestState(trig_fsm->getState()==FsmState::Ready ?
                          FsmState::Starting : FsmState::Error);
        rc_fsm->setRequestState(FsmState::Starting);
        break;
    case FsmState::Starting:
        emit restartAll(rcRunId, startEvNum);
        rc_fsm->setRequestState(isMegamaster() ? FsmState::OpenGlobTrig : FsmState::Run);
        if(!isMegamaster())
            trig_fsm->setRequestState(FsmState::Starting);
        break;

    case FsmState::OpenGlobTrig:
        trig_fsm->setRequestState(FsmState::OpenGlobTrig);
        rc_fsm->setRequestState(FsmState::Run);
        break;

    case FsmState::Run:
        actionOnStateRun();
        break;
    case FsmState::PrepareStopping:
        Q_ASSERT(isMegamaster());
        lookingForStopEvNum = true;
        trig_fsm->setRequestState(FsmState::Stopping);
        rc_fsm->setRequestState(FsmState::Stopping);
        break;
    case FsmState::Stopping:
        evNumChecker->stopCheck();
        trig_fsm->setRequestState(FsmState::Stopping);
        emit stopAll(stopEvNum.toUInt());
        break;
    case FsmState::Error:
    {
        auto trigFsmSt = trig_fsm->getState();
        if(trigFsmSt != FsmState::Ready && trigFsmSt != FsmState::Error) {
            // force reset to ready st.
            trig_fsm->setRequestState(FsmState::Error);
        }
        if(config.defaultSetup.ttvxsCore.trig_open) {
            setTrigState(false);
        }
        evNumChecker->stopCheck();
        emit stopAll();
        break;
    }
    default:;
    }
}

void TtvxsMainWindow::onTrigFsmStateChange(const FsmState &state)
{
    setTabState(state!=FsmState::Error, TAB_TRIG_DEV);
    labelFsmStateTrig->setText("TrigFsm: " + fsmStateToString(state));
    rc_fsm->setTrigDeviceMgrState(state);

    switch (state) {
    case FsmState::Error:
    case FsmState::Ready:
        if(config.defaultSetup.ttvxsCore.trig_open)
            setTrigState(false);
        setMegamaster(false);
        break;
    case FsmState::Starting:
        if(isMegamaster())
            setTrigState(false);
        else
            setTrigState(true);
        trig_fsm->setRequestState(isMegamaster() ? FsmState::CloseGlobTrig : FsmState::Run);
        break;
    case FsmState::CloseGlobTrig:
        break;
    case FsmState::OpenGlobTrig:
        setTrigState(true);
        trig_fsm->setRequestState(FsmState::Run);
        break;
    case FsmState::Stopping:
        setTrigState(false);
        trig_fsm->setRequestState(FsmState::Ready);
        break;
    default:
    {}
    }
}

void TtvxsMainWindow::clientManagerRunStateChanged(const FsmState &state)
{
    setTabState(state!=FsmState::Error, TAB_CLIENTS);
    labelFsmStateClient->setText("ClientFsm: " + fsmStateToString(state));
    rc_fsm->setClientManagerRunState(state);
}


// Private methods
void TtvxsMainWindow::init_GUI_style()
{
    QString styleLargeFont = "font-size: 18pt";
    ui->pushButton_start->setStyleSheet(styleLargeFont);
    ui->pushButton_stop->setStyleSheet(styleLargeFont);
    ui->labelRunState->setStyleSheet(styleLargeFont);
    ui->pushButton_start->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButton_stop->setIcon(style()->standardIcon(QStyle::SP_MediaStop));

    ui->spinBox_timer->setMaximum(10000000);
    ui->spinBox_random->setMaximum(1000000);
}

void TtvxsMainWindow::init_GUI_default_settings()
{
    labelWarnTtvxs = new QLabel();
    labelWarnClients = new QLabel();
    labelWarnOther = new QLabel();

    ui->statusBar->addWidget(labelWarnTtvxs);
    ui->statusBar->addWidget(labelWarnClients);
    ui->statusBar->addWidget(labelWarnOther);

    labelWarnTtvxs->hide();
    labelWarnClients->hide();
    labelWarnOther->hide();
}

void TtvxsMainWindow::initTableDev()
{
    auto t = ui->tableWidgetDev;
    t->setRowCount(ROW_CH_CNT);

    t->setVerticalHeaderItem(ROW_NAME, new QTableWidgetItem("Name"));
    t->setVerticalHeaderItem(ROW_STATUS, new QTableWidgetItem("Status"));

    t->setVerticalHeaderItem(ROW_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    t->setVerticalHeaderItem(ROW_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));
    t->setVerticalHeaderItem(ROW_DEV_TEMP, new QTableWidgetItem("Temp"));
    t->setVerticalHeaderItem(ROW_DEV_STREAM, new QTableWidgetItem("Data Stream, kB/s"));
    t->setVerticalHeaderItem(ROW_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    t->setVerticalHeaderItem(ROW_WR_TIME, new QTableWidgetItem("WR time"));

    t->setVerticalHeaderItem(ROW_UP_TIME, new QTableWidgetItem("Up time"));
    t->setVerticalHeaderItem(ROW_CUR_EV_NUM, new QTableWidgetItem("Cur evNum"));
    t->setVerticalHeaderItem(ROW_LEMO_CNT, new QTableWidgetItem("TTL pulses"));
    t->setVerticalHeaderItem(ROW_LEMO_RATE, new QTableWidgetItem("TTL rate"));
    t->setVerticalHeaderItem(ROW_SFP_CNT, new QTableWidgetItem("SFP pulses"));
    t->setVerticalHeaderItem(ROW_SFP_RATE, new QTableWidgetItem("SFP rate"));

    t->setVerticalHeaderItem(ROW_RUN_TIME, new QTableWidgetItem("Run time"));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_CANDID, new QTableWidgetItem("Trig run cand."));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_ACCEPTED, new QTableWidgetItem("Trig run accepted"));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_RATE, new QTableWidgetItem("Run trig rate"));

    for(int row : SPACER_ROWS) {
        t->setVerticalHeaderItem(row, new QTableWidgetItem(" "));
    }

    for(int row=ROW_CH_SLOT_FIRST; row<=ROW_CH_SLOT_LAST; ++row) {
        auto item = new QTableWidgetItem(QString("Slot %2").arg(row-ROW_CH_SLOT_FIRST+1, 2, 10, QChar('0')));
        if(isSlotBlocked(row)) {
            item->setForeground(Qt::darkCyan);
        }
        t->setVerticalHeaderItem(row, item);
    }

    t->resizeRowsToContents();
    t->setWordWrap(false);
    t->setStyleSheet(":disabled {background-color:lightGray}");

    int spacer_height = 8;
    auto header = t->verticalHeader();
    header->setSectionResizeMode(QHeaderView::Fixed);
    header->setMinimumSectionSize(spacer_height);

    for(int row : SPACER_ROWS) {
        header->resizeSection(row, spacer_height);
        auto spItem = t->verticalHeaderItem(row);
        spItem->setBackground(Qt::lightGray);
    }
}

void TtvxsMainWindow::initStatPresets()
{
    ui->comboBoxPreset->blockSignals(true);
    for(auto preset : StatPresetFabric::getAllPresets()){
        ui->comboBoxPreset->addItem(preset.label, preset.type);
    }
    ui->comboBoxPreset->blockSignals(false);
}

void TtvxsMainWindow::addDevice(const DeviceIndex &index)
{
    Q_ASSERT(!index.isNull());
    if(devices.contains(index)) return;

    TtvxsDeviceGlobal * glob = new TtvxsDeviceGlobal(index);
    devices.insert(index, QSharedPointer<TtvxsDeviceGlobal>(glob));
    connect(this, &TtvxsMainWindow::resetTimestamp, glob, &TtvxsDeviceGlobal::resetTimestamp);
    connect(this, &TtvxsMainWindow::clearStat, glob, &TtvxsDeviceGlobal::clearStat);
    connect(this, &TtvxsMainWindow::evNumReset, glob, &TtvxsDeviceGlobal::evNumReset);
    connect(this, &TtvxsMainWindow::softClear, glob, &TtvxsDeviceGlobal::softClear);
    connect(this, &TtvxsMainWindow::restartAll, glob, &TtvxsDeviceGlobal::restartAll);
    connect(glob, &TtvxsDeviceGlobal::statusUpdated_, this, &TtvxsMainWindow::statusUpdated);
    connect(glob, &TtvxsDeviceGlobal::feLinkStatusUpdated, feLinkStatusDialog, &FeLinkStatusDialog::setFeLinkStatus);
    if(auto model = feLinkStatusDialog->getModel()) {
        connect(glob->getDevice<TtvxsDevice>(), &TtvxsDevice::feLinkBpUpdated,
                model, &FeLinkTreeModel::setFeLinkBp);
    }
    connect(glob, &TtvxsDeviceGlobal::configWritten, trig_fsm, &DeviceMgrFsm::setDevSetupKey); // FIXME: see BaseSignalRouter

    connect(evNumChecker, &EvNumChecker::getTrigEvNum,
            glob, &TtvxsDeviceGlobal::getTrigEvNum);
    connect(glob, &TtvxsDeviceGlobal::topEvNumResponse,
            evNumChecker, &EvNumChecker::gotTopLevelEvNum);
    connect(glob, &TtvxsDeviceGlobal::evNumResponse,
            evNumChecker, &EvNumChecker::gotDevEvNum);

    connect(glob, &TtvxsDeviceGlobal::stat_updated,
            ui->statistic_widget, &MStreamStatisticWidget::stat_updated);
    connect(glob, &TtvxsDeviceGlobal::stat_from_receiver_updated,
            ui->statistic_widget, &MStreamStatisticWidget::stat_from_receiver_updated);
    connect(this, &TtvxsMainWindow::set_run_state,
            ui->statistic_widget, &MStreamStatisticWidget::set_run_state);

    PnpDiscovery::subscribeMyProgram(glob);

    deviceDispatcher->subscribeHostAddressUpdates(index, devices[index]->getDevice<TtvxsDevice>());

    int colDev = -1;
    auto tableDev = ui->tableWidgetDev;

    if(colDev == -1){
        colDev = tableDev->columnCount();
        tableDev->setColumnCount(colDev+1);
        for(int row=0; row<tableDev->rowCount(); ++row) {
            auto item = new QTableWidgetItem();
            tableDev->setItem(row, colDev, item);
            if(row == ROW_NAME) {
                item->setFlags(item->flags() | Qt::ItemIsEditable);
                item->setText(config.custom_name[index]);
            } else {
                item->setFlags(item->flags() & ~Qt::ItemIsEditable);
            }
            if(SPACER_ROWS.contains(row)) {
                item->setBackground(Qt::lightGray);
                continue;
            }
            if(row < ROW_CH_SLOT_FIRST || row > ROW_CH_SLOT_LAST) continue;
            if(isSlotBlocked(row)){
                item->setBackground(Qt::darkCyan);
            } else {
                item->setCheckState(Qt::Unchecked);
            }
        }
    }

    const quint32 &vxsChMask = config.knownConfigs[index].ttvxsCore.vxsChMask;
    Q_ASSERT(colDev != -1);
    for(int row=ROW_CH_SLOT_FIRST; row<=ROW_CH_SLOT_LAST; ++row) {
        const int slot = row-ROW_CH_SLOT_FIRST+1;
        const int ch = slotToCh(slot);
        if(!isSlotBlocked(row)){
            tableDev->item(row, colDev)->setCheckState(vxsChMask&(1<<ch)?Qt::Checked:Qt::Unchecked);
        }
    }

    auto header_item = new QTableWidgetItem(index.getSerialStr());
    header_item->setData(Qt::UserRole, QList<QVariant>() << index.getDevId() << index.getSerial());
    header_item->setForeground(Qt::red);
    tableDev->setHorizontalHeaderItem(colDev, header_item);
    qInfo() << QString("%1 was added").arg(index.getIdent());
}

void TtvxsMainWindow::disconnectFromDev(const DeviceIndex &index)
{
    if(!index.isNull() && !devices.contains(index)) return;
    devices.remove(index);
    int col = getChCfgCol(index);
    if(col!=-1) {
        ui->tableWidgetDev->removeColumn(col);
    }

    if(!index.isNull()) return;
}

void TtvxsMainWindow::updateGui()
{
    block_widgets_signal(true);
    const auto &cfg = config.defaultSetup;

    ui->checkBox_external_ttl->setChecked(cfg.ttvxsCore.trigSrc.bit.ttl);
    ui->checkBox_feLinkSfp->setChecked(cfg.ttvxsCore.trigSrc.bit.sfp);
    ui->checkBox_timer->setChecked(cfg.ttvxsCore.trigSrc.bit.fixed_pulser);
    ui->checkBox_random->setChecked(cfg.ttvxsCore.trigSrc.bit.random_pulser);

    ui->spinBox_timer->setValue(cfg.trigCtrlStat.trigTimerFreq);
    ui->spinBox_random->setValue(cfg.trigCtrlStat.randomPulserFreq);
    ui->doubleSpinBox_dead_time->setValue(cfg.trigCtrlStat.programmedDeadTime);

    ui->checkBox_aux_cnt_for_trig->setChecked(cfg.ttvxsCore.aux_cnt_for_trig);

    ui->groupBoxStat->setChecked(cfg.statReadout.enable);
    ui->spinBoxStatPeriod->setValue(cfg.statReadout.period);
    ui->comboBoxPreset->setCurrentText(cfg.statReadout.preset.label);

    ui->actionAllow_Remote_Control->setChecked(config.allowRemoteControl);
    (cfg.ttvxsCore.trig_open ? ui->radioButton_open : ui->radioButton_close)
            ->setChecked(true);
     block_widgets_signal(false);
}


int TtvxsMainWindow::getChCfgCol(const DeviceIndex &i) const
{
    auto t = ui->tableWidgetDev;
    if(i.isNull()) return -1;
    for(int col=0; col<t->columnCount(); ++col) {
        if(getDevByTableItem(t->horizontalHeaderItem(col)) == i)
            return col;
    }
    return -1;
}

DeviceIndex TtvxsMainWindow::getDevByTableItem(const QTableWidgetItem *item) const
{
    if(!item) return DeviceIndex();

    bool ok1,ok2;
    QList<QVariant> list = item->data(Qt::UserRole).toList();
    if(list.size()>=2) {
        auto id = static_cast<DeviceId>(list.first().toUInt(&ok1));
        quint64 ser = list.at(1).toULongLong(&ok2);
        if(ok1 && ok2)
            return DeviceIndex(id, ser);
    }
    return DeviceIndex();
}

DeviceIndex TtvxsMainWindow::getDevByCol(int col) const
{
    auto item = ui->tableWidgetDev->horizontalHeaderItem(col);
    return getDevByTableItem(item);
}

bool TtvxsMainWindow::isSlotBlocked(int row)
{
    static QSet<int> BLOCKED_ROWS{ROW_CH_SLOT_FIRST,
                ROW_CH_SLOT_FIRST+CRATE_SLOT_ZERO,
                ROW_CH_SLOT_FIRST+CRATE_SLOT_ZERO+1};
    return BLOCKED_ROWS.contains(row);
}

bool TtvxsMainWindow::performStart(const RcRunId &runId)
{
    qInfo() << __PRETTY_FUNCTION__ ;
    if(rc_fsm->getState() != FsmState::Ready) {
        qWarning() << __PRETTY_FUNCTION__ << "failed";
        return false;
    }

    rcRunId = runId;
    emit configChanged(config);
    setTrigState(false);
    clearWarnings();
    rc_fsm->setRequestState(isMegamaster() ? FsmState::CloseGlobTrig : FsmState::Starting);

    return true;
}

void TtvxsMainWindow::openGlobTrig()
{
    if(!isMegamaster())
        return;

    setTrigState(true);
    trig_fsm->setRequestState(trig_fsm->getState()==FsmState::CloseGlobTrig ?
                          FsmState::OpenGlobTrig : FsmState::Error);
}

void TtvxsMainWindow::closeGlobTrig()
{
    if(!isMegamaster())
        return;
    setTrigState(false);
    trig_fsm->setRequestState(trig_fsm->getState()==FsmState::Ready ?
                          FsmState::Starting : FsmState::Error);
}

void TtvxsMainWindow::setTrigState(bool open)
{
    auto configNew = config;
    configNew.defaultSetup.ttvxsCore.trig_open = open;
    configNew.unify_config();
    emit configChanged(configNew);
}

void TtvxsMainWindow::actionOnStateRun()
{
    DeviceIndexSet devSet;
    if(!isMegamaster()) {
        devSet = config.getDeviceIndexEnabledMap().getEnabled();
    }
    trig_fsm->clearStopEvNumber();
    evNumChecker->set_dev_serials(devSet);
    if(isMegamaster())
        evNumChecker->startCheck();
    writeRunHistoryTimer->start();
    checkCanSendRunDocument();
}

void TtvxsMainWindow::perfomeStop(bool isError)
{
    qInfo() << __PRETTY_FUNCTION__ << "isError:" << isError << " state:" << rc_fsm->getState();
    bool isNotRun = rc_fsm->getState() != FsmState::Run;
    rc_fsm->setRequestState(isNotRun ? FsmState::Error
                            : (isMegamaster() ? FsmState::PrepareStopping
                                              : FsmState::Stopping));
    return;
}

bool TtvxsMainWindow::isReadForRun() const
{
    bool allOk=true;
    for(bool ok: tabOk.values())
        allOk &= ok;
    return allOk;
}

void TtvxsMainWindow::refreshProgramDescription()
{
    QString remoteState;

    progDesc.options.clear();
    progDesc.options["fsm"] = fsmStateToString(rc_fsm->getState());
    progDesc.options["run_index"] = rcRunId.runIndex;
    progDesc.options["run_number"] = QVariant(rcRunId.runNumber).toString();

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    // bool readyForRun = isReadForRun();
    for(auto i: config.clients.keys()) {
        remoteState += QString("%1; ").arg(i.toString());
    }
    emit updateProgramDescription(progDesc);
    emit sendRemoteState(remoteState);
}

void TtvxsMainWindow::updateControlEnable()
{
    const auto &debug = Globals::getDebugMode();

    discoverDialog->setEnabled(isReadyErrorSt());

    bool enCfg = isReadyErrorSt() || (rc_fsm->getState()==FsmState::Run && (debug || !remoteConnected));
    ui->groupBox_trigger_state->setEnabled(enCfg);
    ui->groupBox_trigger_source->setEnabled(enCfg);
    ui->checkBox_aux_cnt_for_trig->setEnabled(enCfg);
    ui->groupBoxStat->setEnabled(enCfg);

    bool enXOff = isReadyErrorSt() || (rc_fsm->getState()==FsmState::Run && debug);
    auto table = ui->tableWidgetDev;
    for(int col=0; col<table->columnCount(); ++col){
        for(int row=ROW_CH_SLOT_FIRST; row<=ROW_CH_SLOT_LAST; ++row) {
            if(isSlotBlocked(row))
                continue;
            auto i = table->item(row, col);
            if(enXOff) {
                i->setFlags(i->flags() | Qt::ItemIsUserCheckable | Qt::ItemIsEnabled);
            } else {
                i->setFlags(i->flags() & (~Qt::ItemIsUserCheckable) & (~Qt::ItemIsEnabled));
            }
        }
    }
}

void TtvxsMainWindow::updateStatStopButton()
{
    auto st = rc_fsm->getState();
    ui->pushButton_start->setEnabled(st==FsmState::Ready && !remoteConnected);
    ui->pushButton_stop->setEnabled(st==FsmState::Run && !remoteConnected);
}

void TtvxsMainWindow::updateGroupBoxWarn()
{
    QStringList warnStrList;
    bool hidden = true;
    if(!labelWarnTtvxs->isHidden()) {
        hidden = false;
        warnStrList << labelWarnTtvxs->text();
    }
    if(!labelWarnClients->isHidden()) {
        hidden = false;
        warnStrList << labelWarnClients->text();
    }
    if(!labelWarnOther->isHidden()) {
        hidden = false;
        warnStrList << labelWarnOther->text();
    }

    if(!hidden)
        emit sendRunWarn(rcRunId, warnStrList.join("; "));
}

void TtvxsMainWindow::clearWarnings()
{
    for(auto l : {labelWarnTtvxs, labelWarnClients, labelWarnOther}) {
        l->clear();
        l->hide();
    }
}

void TtvxsMainWindow::logRcFsmStateChange(const FsmState &state,
                                          const FsmState &oldState)
{
    QColor col;
    const auto &rcRunIdStr = rcRunId.toString();
    auto txt = QString("%1 %2").arg(fsmStateToString(state)).arg(rcRunIdStr);
    switch (state) {
    case FsmState::Ready:
        if(oldState != FsmState::Stopping)
            return;
        txt = QString("Run %1 stopped").arg(rcRunIdStr);
        col = Qt::darkGreen;
        break;
    case FsmState::Error:
        if(oldState == FsmState::Init || oldState == FsmState::Ready
            || oldState == FsmState::Unknown)
            return;
        txt.append(" failed");
        col = Qt::red;
        break;
    case FsmState::Starting:
        txt.append("...");
        col = Qt::darkYellow;
        break;
    case FsmState::Stopping:
        txt.append("...");
        col = Qt::darkYellow;
        break;
    case FsmState::Run:
        txt.append(" started");
        col = Qt::darkGreen;
        break;
    default:
        break;
    }
    (col==Qt::red ? qWarning() : qInfo()) << txt;
    ui->labelRunState->setText(txt);
    setLabelColor(ui->labelRunState, col);
}

void TtvxsMainWindow::block_widgets_signal(bool block)
{
    for(int i=0; i<widgets_to_block.size(); ++i)
        widgets_to_block[i]->blockSignals(block);
}

bool TtvxsMainWindow::isReadyErrorSt() const
{
    auto st = rc_fsm->getState();
    return st == FsmState::Error || st == FsmState::Ready;
}

void TtvxsMainWindow::checkCanSendRunDocument()
{
    if(!writeRunHistoryTimer->isActive())
        return;

    bool all_clients_config = getMissingConfigClients().isEmpty();

    if(all_clients_config)
        sendRunDocument();
}

ClientIndexSet TtvxsMainWindow::getMissingConfigClients() const
{
    return config.clients.getEnabledSet()
            - makeClientIndexSet(clients_config.keys());
}

void TtvxsMainWindow::sendRunDocument()
{
    const auto &myCfgObj = QVariant::fromValue(config).value<QJsonObject>();
    const auto &myClientIndex = configDb->getClientIndex();
    RunDocument runDoc(rcRunId, myClientIndex, myCfgObj);

    for(auto it=clients_config.cbegin();it!=clients_config.cend();++it){
        auto clRcDoc = QVariant::fromValue(*it).value<RunDocument>();
        if(!runDoc.merge(clRcDoc)) {
            qWarning() << QString("Failed merge RunDocument from client:%1")
                          .arg(it.key().toString());
        }
    }

    auto missingClients = getMissingConfigClients();
    if(!missingClients.isEmpty()) {
        QStringList sl;

        for(auto i : getMissingConfigClients())
            sl << i.toString();

        qWarning() << "Missing config (for RunDocument) from clients: "
                   << sl.join(", ");
    }

    if(isMegamaster()) {
        runDoc.setMeta(myClientIndex);
        configDb->saveRunDocument(runDoc);
    } else {
        emit sendRunDocumentToRc(runDoc);
    }
    clients_config.clear();
    writeRunHistoryTimer->stop();
}

static QColor get_status_col(SensorStatus status) {
    switch (status) {
    case SENSOR_UNKNOWN: return Qt::gray;
    case SENSOR_NORMAL: return Qt::green;
    case SENSOR_WARNING: return Qt::yellow;
    case SENSOR_CRITICAL: return Qt::red;
    default: return {};
    }
    return {};
}
// Private slots
void TtvxsMainWindow::statusUpdated(const TtvxsStatus &st, const TtvxsStatus &prev_st)
{
    const auto &index = st.getDeviceIndex();
    if(!devices.contains(index))
        return;

    deviceStatusDialog->setStatus(st);
    if(lookingForStopEvNum && st.coreStatus.trigState) {
        if(st.trigInfo) {
            stopEvNum = st.trigInfo->eventNumber;
            qInfo() << "StopEvNum detected as:" << stopEvNum;
        } else {
            stopEvNum = 0;
        }
        trig_fsm->markStopEvNumber();
        lookingForStopEvNum = false;
    }

    auto tableDev = ui->tableWidgetDev;
    const int colCh = getChCfgCol(index);

    Q_ASSERT(colCh!=-1);
    if(colCh == -1) return;

    const auto &aux_cnt = st.coreStatus.aux_cnt;
    const auto &devCfg = config.knownConfigs[index];
    const QMap<QString, QString> tags({{"serial", index.getSerialStr()}});
    auto vxsChMask = devCfg.ttvxsCore.vxsChMask;
    for(int ch=0; ch<TTVXS_CH_NUM; ++ch) {
        bool inputSt = st.coreStatus.inputState & (1<<ch);
        bool timeoutSt = st.coreStatus.timeoutState & (1<<ch);
        const int slot = chToSlot(ch);
        if(!prev_st.isNull()) {
            if(timeoutSt != ((prev_st.coreStatus.timeoutState & (1<<ch)) != 0)) {
                QString baseStr = QString("[%3] %1 %4 timeout on slot=%2")
                        .arg(index.getIdent()).arg(slot)
                        .arg(QDateTime::currentDateTime().toString());
                if(timeoutSt) {
                    qWarning() << baseStr.arg("got");
                } else {
                    qInfo() << baseStr.arg("remove");
                }
            }
        }
        const int row = ROW_CH_SLOT_FIRST+slot-1;
        auto chItem = tableDev->item(row, colCh);
        QBrush brush_green;
        brush_green.setColor(Qt::green);
        QBrush brush_transparent;
        brush_transparent.setColor(Qt::transparent);
        chItem->setBackground(inputSt ? brush_green : brush_transparent);
        if(timeoutSt != (chItem->foreground() == Qt::red)) {
            chItem->setForeground(timeoutSt ? Qt::red : QColor());
            chItem->setText(timeoutSt ? "timeout" : "");
        }
        bool ch_en = vxsChMask&(1<<ch);

        tableDev->item(row, colCh)->setText(ch_en ? aux_cnt.getBusyTime(ch) : "");
        if (ch_en) {
            auto tags2 = tags;
            tags2["channel"] = QString::number(slot);
            tags2["channel_descr"] = QString::number(slot);
            MetricSender::send(QString("rc.channel.time_xoff"), aux_cnt.getBusyTimeUs(ch) / 1e6, tags2);
        }
    }
    tableDev->item(ROW_UP_TIME, colCh)->setText(aux_cnt.getUpTime());
    MetricSender::send("uptime", aux_cnt.getUpTimeSec(), tags);
    tableDev->item(ROW_LEMO_CNT, colCh)->setText(aux_cnt.getTtlCnt());
    tableDev->item(ROW_SFP_CNT, colCh)->setText(aux_cnt.getSfpCnt());
    tableDev->item(ROW_CUR_EV_NUM, colCh)->setText(st.trigInfo.has_value()
                                                   ? QString("%L1").arg(st.trigInfo->eventNumber)
                                                   : "-");
    if(st.trigInfo.has_value())
        MetricSender::send("rc.eventNumber", st.trigInfo->eventNumber, tags);

    tableDev->item(ROW_RUN_TIME, colCh)->setText(aux_cnt.getRunTime());
    MetricSender::send("time.runtime", aux_cnt.getRunTimeSec(), tags);
    tableDev->item(ROW_RUN_TRIG_CANDID, colCh)->setText(aux_cnt.getTrigRunCandCnt());
    tableDev->item(ROW_RUN_TRIG_ACCEPTED, colCh)->setText(aux_cnt.getTrigRunAcceptCnt());

    tableDev->item(ROW_DEV_FW, colCh)->setText(st.getHwFwStr());
    TempsInfo ti(st.getTemps());
    tableDev->item(ROW_DEV_TEMP, colCh)->setText(ti.brief());
    tableDev->item(ROW_DEV_TEMP, colCh)->setToolTip(ti.full());

    auto wr_item = tableDev->item(ROW_WR_TIME, colCh);
    if (st.wrStatus.has_value()) {
        QDateTime wrTime = QDateTime::fromMSecsSinceEpoch(st.wrStatus->mSecsSinceEpoch());
        bool wrTimeValid = st.wrStatus->timeValid();
        wr_item->setText(st.wrStatus->timeValid() ? wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss") : "invalid"); // add ".zzz" for ms
        wr_item->setBackground(wrTimeValid ? Qt::green : Qt::lightGray);
    } else {
        wr_item->setText({});
        wr_item->setBackground({});
    }

    tableDev->horizontalHeaderItem(colCh)->setForeground(st.connected ? QColor() : Qt::gray);
    tableDev->item(ROW_STATUS, colCh)->setText(st.connected ? "Connected" : "Disconnected");
    tableDev->item(ROW_STATUS, colCh)->setForeground(st.connected ? Qt::darkGreen : Qt::red);
    auto mcu_ver_item = tableDev->item(ROW_DEV_MCU_VER, colCh);
    if (st.bmcStatus.has_value()) {
        const BmcStatus &bmcStatus = st.bmcStatus.value();

        mcu_ver_item->setText(bmcStatus.getVerStr());
        mcu_ver_item->setToolTip(QString("overall - %1\npower - %2\nPLL- %3")
                                     .arg(bmcStatus.get_overall_status_str())
                                     .arg(bmcStatus.get_power_status_str())
                                     .arg(bmcStatus.get_pll_status_str()));
        //    QBrush mcu_ver_brush = mcu_ver_item->background();
        //    mcu_ver_brush.setColor(QColor(255,0,0));
        //    mcu_ver_item->setBackground(mcu_ver_brush);
        mcu_ver_item->setBackground(get_status_col(bmcStatus.get_overall_status()));
    } else {
        mcu_ver_item->setText({});
        mcu_ver_item->setToolTip({});
        mcu_ver_item->setBackground({});
    }

//    lastDevStatus = st;
    const auto &aux1 = st.coreStatus.aux_cnt;
    const auto &aux2 = prev_st.coreStatus.aux_cnt;
    tableDev->item(ROW_RUN_TRIG_RATE, colCh)->setText(aux1.getRunTrigRate(aux2));
    tableDev->item(ROW_LEMO_RATE, colCh)->setText(aux1.getTtlRate(aux2));
    tableDev->item(ROW_SFP_RATE, colCh)->setText(aux1.getSfpRate(aux2));

    if(!st.connected && !isReadyErrorSt()) {
        onRunError(rcRunId,
                   QString("TTVXS #%1 s/n:%2 offline")
                   .arg(configDb->getIndex().toString())
                   .arg(st.getSerialStr()));
    }
}

void TtvxsMainWindow::updateDeviceLldpInfo(const DeviceIndex &index, const LldpInfo &lldp)
{
    const int col = getChCfgCol(index);
    if(col == -1) return;
    auto ttvxsItem = ui->tableWidgetDev->item(ROW_CH_SLOT_FIRST+lldp.ext.chassis_slot-1, col);
    if(!ttvxsItem)
        return;
//    Q_ASSERT(ttvxsItem);
    if(ttvxsItem->text() != "ttvxs") {
        ttvxsItem->setText("ttvxs");
        ui->tableWidgetDev->resizeColumnsToContents();
    }
}

int TtvxsMainWindow::get_col_by_serial(QString serial)
{
    QTableWidget *table = ui->tableWidgetDev;
    for(int col=table->columnCount()-1; col>=0; --col) {
        if(getDevByTableItem(table->horizontalHeaderItem(col)).getSerialStr() == serial)
            return col;
    }
    return -1;
}

void TtvxsMainWindow::update_table_size(QTableWidget * const table, bool widthOnly)
{
//    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

void TtvxsMainWindow::onRunError(const RcRunId &runId, QString errStr)
{
    if(isReadyErrorSt() || rcRunId != runId)
        return;

    if(remoteConnected){
        QString mess = QString("Retransmit runError to the remoteControl in Run %1")
                .arg(rcRunId.toString());
        if(!errStr.isEmpty())
            mess.append(": ").append(errStr);
        qWarning() << mess;
        emit sendRunError(runId, errStr);
    } else {
        QString mess = QString("Errors in Run %1")
                .arg(rcRunId.toString());
        if(!errStr.isEmpty())
            mess.append(": ").append(errStr);
        qWarning()<< mess;
        
        perfomeStop(true);
    }
}

void TtvxsMainWindow::onRunWarning(const RcRunId &runId, QString warnStr)
{
    if(!warnStr.isEmpty() &&
            (isReadyErrorSt() || rcRunId != runId) )
        return;

    QLabel *label = nullptr;
    QString prefix;
    warnStr = warnStr.trimmed();
    if(sender() == clientCore) {
        label = labelWarnClients;
        prefix = "Clients: ";
    }
/*
    else if(devices.values().contains(qobject_cast<TtvxsDeviceGlobal*>(sender()))){ // FIXME: WTF???
            label = labelWarnTtvxs;
            prefix = "Ttvxs: ";
        }
*/
    if(!label) {
        label = labelWarnOther;
        prefix = "Other: ";
    }
    if(!warnStr.isEmpty())
        warnStr.prepend(prefix);

    const int WARN_SIZE_LIMIT= 90;
    if(warnStr.size()>WARN_SIZE_LIMIT) {
        warnStr = warnStr.left(WARN_SIZE_LIMIT-3)+"...";
    }
    label->setText(warnStr);
    label->setHidden(false);

    updateGroupBoxWarn();
}

void TtvxsMainWindow::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(rc_fsm->getState() == FsmState::Run && rcRunId == runId) {
        if(remoteConnected) {
            QString mess = QString("Retransmit backPressure(%2) to the remoteControl in Run %1")
                    .arg(rcRunId.toString()).arg(backPressure);
            if(backPressure)
                qWarning() << mess;
            else
                qInfo() << mess;
            emit sendBackPressure(runId, backPressure);
        } else {
            if(backPressure) {
                setTrigState(false);
                qWarning()<< "BackPressure enabled";
            } else {
                setTrigState(true);
                qInfo()<< "BackPressure disabled";

            }
        }
    }
}

void TtvxsMainWindow::onEvNumCheckFail(QString mess)
{
    onRunWarning(rcRunId, mess);
}

void TtvxsMainWindow::setTabState(bool isOk, int tabIndex)
{
    QTabBar *bar = ui->tabWidget->tabBar();
    QColor c = bar->tabTextColor(tabIndex);
    if(c.isValid() == isOk) {
        bar->setTabTextColor(tabIndex, isOk ? QColor() : Qt::red);
    }
    bool prevOk = isReadForRun();
    tabOk[tabIndex] = isOk;
    bool newOk = isReadForRun();
    if(newOk != prevOk)
        refreshProgramDescription();
}

void TtvxsMainWindow::config_transfer_from_clients(const ClientIndex &id, const QJsonObject &json)
{
    if(!config.clients.contains(id))
        return;

    clients_config[id] = json;
    checkCanSendRunDocument();
}

void TtvxsMainWindow::onRemoteStartRequest(const RcRunId &runId, const QVariant &startEvNumber)
{
    qInfo() << __PRETTY_FUNCTION__ << runId << startEvNumber;
    startEvNum = startEvNumber;
    setMegamaster(false);
    bool success = performStart(runId);
    clients_config.clear();
    emit remoteStartCompleted(success, runId);

    refreshProgramDescription();
}

void TtvxsMainWindow::onRemoteStopRequest(const RcRunId &runId, const QVariant &stopEvNumber)
{
    if(isReadyErrorSt())
        return;
    stopEvNum = stopEvNumber;
    perfomeStop(false);
    refreshProgramDescription();
}

void TtvxsMainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl) == 0) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        progDesc.interfaces.push_back(interface);
    }

    ProgramInterface &i = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !status.portInUse;

    if(status.portInUse)
        i.setPeer(status.peerHost, status.peerPort);
    else
        i.peers.clear();

    refreshProgramDescription();
    bool remoteConnectedOnRun = !isReadyErrorSt() && !remoteConnected
        && status.portInUse;
    bool remoteDisconnectedOnRun = !isReadyErrorSt() && remoteConnected
        && !status.portInUse;
    remoteConnected = status.portInUse;
    if(remoteConnectedOnRun || remoteDisconnectedOnRun) {
        qWarning("Stop run due to remoteControl %sconnected",
                 remoteDisconnectedOnRun ? "dis" : "");
        perfomeStop(true);
    }
    updateStatStopButton();
    updateControlEnable();
}

//=== GUI SLOTS
void TtvxsMainWindow::on_actionChange_devices_triggered()
{
    discoverDialog->show();
}

void TtvxsMainWindow::on_actionReset_timestamp_triggered()
{
    emit resetTimestamp();
}

void TtvxsMainWindow::on_actionSoftclear_triggered()
{
    emit softClear();
}

void TtvxsMainWindow::on_actionClear_stats_triggered()
{
    emit clearStat();
}

void TtvxsMainWindow::on_actionSNMP_dialog_triggered()
{
    SnmpDialog dialog(this);
    dialog.exec();
}

void TtvxsMainWindow::on_actionEvent_number_reset_triggered()
{
    emit evNumReset();
}

void TtvxsMainWindow::on_actionAllow_Remote_Control_toggled(bool val)
{
    config.allowRemoteControl = val;
    emit configChanged(config);
    emit allowRemoteControl(config.allowRemoteControl);
    refreshProgramDescription();
}

void TtvxsMainWindow::on_pushButton_start_clicked()
{
    if(devices.size() == 1 )
        setMegamaster(true);
    else{
        qWarning() << "Start failed! More than one TTVXS device.";
        return;
    }
    clients_config.clear();

    const RcRunId newRunId(localProgramIndex.toString(),
                           ++config.pc.localRunNumber,
                           QUuid::createUuid());
    startEvNum = 1;
    bool success = performStart(newRunId);

    if(success && config.clients.isEmpty()) {
        const auto &json = QVariant::fromValue(config).value<QJsonObject>();
        emit write_history_config(newRunId, json, true);
    }
    emit configChanged(config);
}

void TtvxsMainWindow::on_pushButton_stop_clicked()
{
    perfomeStop(false);
}

void TtvxsMainWindow::on_radioButton_open_toggled(bool checked)
{
    config.defaultSetup.ttvxsCore.trig_open = checked;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_checkBox_external_ttl_toggled(bool checked)
{
    config.defaultSetup.ttvxsCore.trigSrc.bit.ttl = checked;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_checkBox_feLinkSfp_toggled(bool checked)
{
    config.defaultSetup.ttvxsCore.trigSrc.bit.sfp = checked;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_checkBox_timer_toggled(bool checked)
{
    ui->spinBox_timer->setEnabled(checked);
    config.defaultSetup.ttvxsCore.trigSrc.bit.fixed_pulser = checked;
    config.unify_config();
    emit configChanged( config);
}

void TtvxsMainWindow::on_checkBox_random_toggled(bool checked)
{
    ui->spinBox_random->setEnabled(checked);
    ui->doubleSpinBox_dead_time->setEnabled(checked);
    ui->labelDeadTime->setEnabled(checked);
    config.defaultSetup.ttvxsCore.trigSrc.bit.random_pulser = checked;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_spinBox_timer_valueChanged(int val)
{
    config.defaultSetup.trigCtrlStat.trigTimerFreq = val;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_spinBox_random_valueChanged(int val)
{
    config.defaultSetup.trigCtrlStat.randomPulserFreq = val;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_doubleSpinBox_dead_time_valueChanged(double val)
{
    config.defaultSetup.trigCtrlStat.programmedDeadTime = val;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_checkBox_aux_cnt_for_trig_toggled(bool checked)
{
    config.defaultSetup.ttvxsCore.aux_cnt_for_trig = checked;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_groupBoxStat_toggled(bool checked)
{
    config.defaultSetup.statReadout.enable = checked;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_spinBoxStatPeriod_valueChanged(int period_ms)
{
    config.defaultSetup.statReadout.period = static_cast<quint16>(period_ms);
    config.unify_config();
    emit configChanged(config);
}


void TtvxsMainWindow::on_comboBoxPreset_currentIndexChanged(int index)
{
    StatPreset stat_preset;
    int preset_type = ui->comboBoxPreset->itemData(index).toInt();
    stat_preset = StatPresetFabric::getPreset(preset_type);

    config.defaultSetup.statReadout.preset = stat_preset;
    config.unify_config();
    emit configChanged(config);
}

void TtvxsMainWindow::on_tableWidgetDev_cellChanged(int row, int column)
{
    if(row == ROW_NAME) {
        auto table = ui->tableWidgetDev;
        DeviceIndex index = getDevByCol(column);
        if(index.isNull())
            return;
        config.custom_name[index] = table->item(row, column)->text();
        emit configChanged(config);
        return;
    }
    if(row < ROW_CH_SLOT_FIRST || row > ROW_CH_SLOT_LAST) return;
    DeviceIndex index = getDevByCol(column);
    if(index.isNull()) return;
    const int slot = row-ROW_CH_SLOT_FIRST+1;
    if(isSlotBlocked(row)) return;
    const int ch = slotToCh(slot);

    Q_ASSERT(config.knownConfigs.contains(index));
    auto &cfg = config.knownConfigs[index];
    bool en = ui->tableWidgetDev->item(row, column)->checkState() == Qt::Checked;
    if(((cfg.ttvxsCore.vxsChMask & (1<<ch)) != 0) != en) {
        if(en)
            cfg.ttvxsCore.vxsChMask |= 1<<ch;
        else
            cfg.ttvxsCore.vxsChMask &= ~static_cast<quint32>(1<<ch);

        emit configChanged(config);
    }
}

void TtvxsMainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void TtvxsMainWindow::on_actionFeLinkStatusDialog_triggered()
{
    feLinkStatusDialog->show();
}

// FIXME: duplicated code
void TtvxsMainWindow::on_actionDevice_Status_triggered()
{
    deviceStatusDialog->show();
}

void TtvxsMainWindow::on_tableWidgetDev_currentItemChanged(QTableWidgetItem *current, QTableWidgetItem *previous)
{
    if (!current)
        return;
    auto deviceIndex = getDevByCol(current->column());
    deviceStatusDialog->setIndex(deviceIndex);
}

void TtvxsMainWindow::on_tableWidgetDev_itemDoubleClicked(QTableWidgetItem *item)
{
    if (!item)
        return;

    auto deviceIndex = getDevByCol(item->column());

    deviceStatusDialog->setIndex(deviceIndex);
    deviceStatusDialog->show();
}

bool TtvxsMainWindow::isMegamaster() const
{
    return megamaster;
}

void TtvxsMainWindow::setMegamaster(bool newMegamaster_flg)
{
    megamaster = newMegamaster_flg;
}
