/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "BaseMainWindow.h"

#include <QCloseEvent>
#include <QCoreApplication>
#include <QDebug>
#include <QMenuBar>
#include <QSettings>
#include <QTimer>

#include "base/Globals.h"

#include "AboutBox.h"

BaseMainWindow::BaseMainWindow(QWidget *parent) :
    QMainWindow(parent)
{
    // setupMenu();

    QTimer::singleShot(0, this, &BaseMainWindow::delayedInit);

    auto t = Globals::getInitialConfigSelector().toWindowTitle();
    if(windowTitle() != t)
        setWindowTitle(t);
}

void BaseMainWindow::delayedInit()
{
    loadMainWindowPosition();
}

void BaseMainWindow::configSelectorChanged(const ConfigIndexName &cin)
{
    auto t = ConfigSelector(Globals::getDaqModule(), cin).toWindowTitle();
    if(windowTitle() != t)
        setWindowTitle(t);
}

void BaseMainWindow::closeEvent(QCloseEvent *event)
{
    saveMainWindowPosition();
    QMainWindow::closeEvent(event);
}

void BaseMainWindow::setupMenu(QMenu *file, QMenu *help, QMenu *options, QMenu *tools)
{
    if(!menuFile)
        menuFile = file;
    if(!menuOptions)
        menuOptions = options;
    if(!menuTools)
        menuTools = tools;
    if(!menuHelp)
        menuHelp = help;

    addMenuFileActions();
    addMenuOptionsActions();
    addMenuToolsActions();
    addMenuHelpActions();

    menuBar()->setVisible(true);
}

void BaseMainWindow::createMenuIfAbsent(QMenu **menuLocal)
{
    if(*menuLocal)
        return;
    QString n;
    if(menuLocal == &menuFile) {
        n = tr("&File");
    } else if(menuLocal == &menuOptions) {
        n = tr("&Options");
    } else if(menuLocal == &menuTools) {
        n = tr("&Tools");
    } else if(menuLocal == &menuHelp) {
        n = tr("&Help");
    } else {
        n = QString("Menu #%1").arg(1); // menuBar()
    }

    *menuLocal = menuBar()->addMenu(n);
}

void BaseMainWindow::addMenuFileActions()
{
    createMenuIfAbsent(&menuFile);

    QIcon iconQuit = QIcon::fromTheme("application-exit");
    menuFile->addAction(iconQuit,
                        tr("E&xit"),
                        this,
                        &BaseMainWindow::actionQuit,
                        QKeySequence::Quit);
}

void BaseMainWindow::addMenuOptionsActions()
{
    createMenuIfAbsent(&menuOptions);

    menuOptions->addSeparator();
    menuItemNativeDialogs = new QAction(tr("Use native dialogs"), menuOptions);
    menuItemNativeDialogs->setCheckable(true);
    menuItemNativeDialogs->setChecked(Globals::instance().getUseQtDialogs());
    connect(menuItemNativeDialogs, &QAction::toggled, this, &BaseMainWindow::actionNativeDialogs);
    menuOptions->addAction(menuItemNativeDialogs);
}

void BaseMainWindow::addMenuHelpActions()
{
    createMenuIfAbsent(&menuHelp);

    QIcon iconAbout = QIcon::fromTheme("help-about");
    menuHelp->addAction(iconAbout, tr("About..."), this, &BaseMainWindow::actionAbout);
}

void BaseMainWindow::actionQuit()
{
    close();
}

void BaseMainWindow::actionAbout()
{
    AboutBox::about();
}

void BaseMainWindow::loadMainWindowPosition()
{
    QSettings settings;
    restoreGeometry(settings.value(SETTING_GROUP_MAINWINDOW + "/geometry")
                    .toByteArray());
    restoreState(settings.value(SETTING_GROUP_MAINWINDOW + "/windowState")
                 .toByteArray());
    loadAppGuiSettings(settings);
}

void BaseMainWindow::saveMainWindowPosition() const
{
    QSettings settings;
    settings.beginGroup(SETTING_GROUP_MAINWINDOW);
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();
    saveAppGuiSettings(settings);
}

void BaseMainWindow::actionNativeDialogs(bool state)
{
    Globals::instance().setUseQtDialogs(state);
}
