/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef DAQMODULE_H
#define DAQMODULE_H

#include <QDebug>
#include <QMetaType>
#include <QString>

class DaqModule
{
public:
    enum class Type {
        Default,
        Adc64,
        Adc64System,
        AfiViewer,
        Cru,
        DeviceBrowser,
        DeviceFlashProg,
        EventBuilder,
        Generator,
        GenericClient,
        HrbAnalyzer,
        HrbSystem,
        Ideas,
        Land,
        MongoViewer,
        MpdRawstat,
        MpdTlu,
        Msc16Ve,
        Msc16VeTm,
        MStream,
        PhosFec,
        PnpBeacon,
        PnpBrowser,
        RunControl,
        RunManager,
        Scheduler,
        TdcControl,
        TQDC,
        Trc,
        Ttvxs,
        T0CfgMonitor,
        VmeDaq,
        DimReceiver,
        MetricAggr
    };
    constexpr explicit DaqModule(const Type type = Type::Default) noexcept : type(type) {}
    constexpr operator Type() const { return type; }
    constexpr bool isEmpty() const { return type == Type::Default; }
    QString toString() const{ return QString::fromStdString(toStdString()); }
    std::string toStdString() const;
    static DaqModule fromStdString(const std::string &s);
    static DaqModule fromString(const QString &s);
    QString prettyName() const;

    friend bool operator==(const DaqModule &l, const DaqModule &r)
    { return l.type == r.type; }

    friend bool operator==(const DaqModule &l, const DaqModule::Type &r)
    { return l.type == r; }
    friend bool operator==(const DaqModule::Type &l, const DaqModule &r)
    { return l == r.type; }

    friend bool operator<(const DaqModule &l, const DaqModule &r)
    { return l.type < r.type; }

    friend uint qHash(const DaqModule &r, uint seed = 0) {
        return qHash(static_cast<uint>(r.type), seed);
    }
    friend QDataStream &operator<<(QDataStream &out, const DaqModule &value);
    friend QDataStream &operator>>(QDataStream &in, DaqModule &value);

private:
    Type type = Type::Default;
};

Q_DECLARE_METATYPE(DaqModule)

QDebug operator<<(QDebug dbg, const DaqModule &r);

class DaqModuleIndex
{
public:
    DaqModuleIndex(const DaqModule::Type type = DaqModule::Type::Default,
                   const QString &index = QString())
        : type(type), index(index)
    {}

    DaqModule::Type type = DaqModule::Type::Default;
    QString index;
};

inline bool operator==(const DaqModuleIndex &l, const DaqModuleIndex &r)
{
    return l.type == r.type && l.index == r.index;
}

inline bool operator<(const DaqModuleIndex &l, const DaqModuleIndex &r)
{
    if (l.type == r.type)
        return l.index < r.index;
    else
        return l.type < r.type;
}

class DaqModuleSet: public QSet<DaqModule>
{
public:
    DaqModuleSet() noexcept = default;
    DaqModuleSet(std::initializer_list<DaqModule> init) :
        QSet<DaqModule>{init} {}
};
Q_DECLARE_METATYPE(DaqModuleSet);

#endif // DAQMODULE_H
