/*
**    Copyright 2009-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "FitterGaus.h"
#include "dsplib.h"
#include "lmfit.h"

#include <cmath>
#include <cstdio>

namespace dsplib {

const double M_1_SQRT2PI = 1./sqrt(2 * M_PI);


inline double sqr(double x) { return pow(x, 2); }

double GausEstimator(double x[3])
{
    double gp[3]={0,0,0};
    FindGausApproxD(x, 3, gp);
    return gp[1]-1;
}

double FindGausDevNorm (const int *x, int len, double *par)
{
    double rms = 0;
    if (par[2]==0) {
        for (int i = 0; i < len; i++)
            rms += sqr(1.*x[i]);
        return sqrt(rms);
    }
    for (int i = 0; i < len; i++) {
        //	double expectval = par[0] * gauss (i - par[1], par[2]);
        double arg = (i - par[1]) / par[2];
        double expectval = par[0] * exp(-.5*arg*arg);
        rms += sqr (x[i] - expectval);// / sqr(expectval);
    }
    return sqrt(rms);
}

double FindGausDevNormD (const double *x, int len, double *par)
{
    int i;
    double rms = 0;
    for (i = 0; i < len; i++) {
        double expectval = par[0] * gauss (i - par[1], par[2]);
        rms += sqr (x[i] - expectval);// / sqr(expectval);
    }
    return sqrt(rms);
}

void FindGausApproxD (const double *x, int len, double *par)
{

    int i;
    double tt, val, mean, rms;
    double allcha = 0, sumx = 0, sumx2 = 0;

    if (!x || !par)
        return;

    par[0] = 0;
    par[1] = 0;
    par[2] = 0;

    for (i = 0; i < len; i++) {
        tt = i + .5;
        val = x[i];
        if (val < 0)
            val = 0;                  // FIXME
        sumx += val * tt;
        sumx2 += val * tt * tt;
        allcha += val;
        //    printf("[%d]: x %g, sumx %g, sumx2 %g, allcha %g\n", i, val, sumx, sumx2, allcha);
    }
    if (allcha > 0) {
        mean = sumx / allcha;
        rms = sqrt (sumx2 / allcha - mean * mean);
        //    printf("mean: %g, rms %g\n", mean, rms);
        //   par[0] = 0.5*allcha/(sqrt(2*pi)*rms);
        par[0] = allcha / (sqrt (2 * M_PI) * rms);
        par[1] = mean - 0.5;
        par[2] = rms;
    }
}

void FindGausApprox (const int *x, int len, double *par)
{

    int i;
    double allcha = 0, sumx = 0, sumx2 = 0;

    if (!x || !par)
        return;

    for (i = 0; i < len; i++) {
        int val = x[i];
        int tt = i;
        if (val < 0)
            val = 0;                  // FIXME
        sumx += val * tt;
        sumx2 += val * tt * tt;
        allcha += val;
        //    printf("[%d]: x %g, sumx %g, sumx2 %g, allcha %g\n", i, val, sumx, sumx2, allcha);
    }
    {
        double mean, rms;
        if (allcha==0) {
            mean = len/2;
            rms = len/2;
        } else {
            mean = sumx / allcha;
            rms = sqrt (sumx2 / allcha - mean * mean);
        }
        //    printf("mean: %g, rms %g\n", mean, rms);
        //   par[0] = 0.5*allcha/(sqrt(2*pi)*rms);
        par[0] = allcha / (sqrt (2 * M_PI) * rms);
        par[1] = mean;
        par[2] = rms;
    }
}


typedef struct {
    const double *y;
    double fp[3];
    int fix;
} fitdata_gaus_t;

void lm_evaluate_gaus (double *par, int m_dat, double *fvec, double *fjac, void *data, int *info)
/*
 *	par is an input array. At the end of the minimization, it contains
 *        the approximate solution vector.
 *
 *	m_dat is a positive integer input variable set to the number
 *	  of functions.
 *
 *	fvec is an output array of length m_dat which contains the function
 *        values the square sum of which ought to be minimized.
 *
 *	data is a read-only pointer to fitdata_gaus_t
 *
 *      info is an integer output variable. If set to a negative value, the
 *        minimization procedure will stop.
 */
{
    fitdata_gaus_t *mydata = (fitdata_gaus_t *) data;
    double m1, arg, ex, fac, p0, p1, p2, p0ex;

    if (mydata->fix) {
        int nmp=0;
        if (!(mydata->fix&1)) {p0 = par[nmp]; nmp++;} else p0 = mydata->fp[nmp];
        if (!(mydata->fix&2)) {p1 = par[nmp]; nmp++;} else p1 = mydata->fp[nmp];
        if (!(mydata->fix&4)) {p2 = par[nmp]; nmp++;} else p2 = mydata->fp[nmp];
    } else {
        p0 = par[0];
        p1 = par[1];
        p2 = par[2];
    }

    m1 = p2 ? 1. / p2 : 1;

    // duplicated code: optimized for speed
    if (fjac) {
        for (int i = 0; i < m_dat; i++) {
            arg = (i - p1) * m1;
            ex = exp(-0.5*arg*arg);
            p0ex = -p0*ex;
            fvec[i] = mydata->y[i] +p0ex;
            fac = p0ex*arg;
            fjac[0*m_dat+i] = -ex;
            fjac[1*m_dat+i] = fac*m1;
            fjac[2*m_dat+i] = fac*m1*arg;
        }
    } else {
        for (int i = 0; i < m_dat; i++) {
            arg = (i - p1) * m1;
            ex = exp(-0.5*arg*arg);
            p0ex = -p0*ex;
            fvec[i] = mydata->y[i] +p0ex;
        }
    }
    /*
*** Original gaussian() code for reference (not optimized)
  double arg, ex, fac, y;
  arg = (x - a[1]) / a[2];
  ex = exp (-arg * arg);
  fac = 2.0 * a[0] * ex * arg;
  y = a[0] * ex;
  jac[0] = ex;
  jac[1] = fac / a[2];
  jac[2] = jac[1] * arg;
  return y;
*/
    *info = *info;                /* to prevent a 'unused variable' warning */
    /* if <parameters drifted away> { *info = -1; } */
}

void lm_print_default( int n_par, const double* par, int m_dat, const double* fvec,
                      const void *data, int iflag, int iter, int nfev )
/*
 *       data  : for soft control of printout behaviour, add control
 *                 variables to the data struct
 *       iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
 *       iter  : outer loop counter
 *       nfev  : number of calls to *evaluate
 */
{
    //    double f, y, t;
    int i;
    //    const fitdata_t *mydata = (fitdata_t*)data;

    if (iflag==2) {
        printf ("trying step in gradient direction\n");
    } else if (iflag==1) {
        printf ("determining gradient (iteration %d)\n", iter);
    } else if (iflag==0) {
        printf ("starting minimization\n");
    } else if (iflag==-1) {
        printf ("terminated after %d evaluations\n", nfev);
    }

    printf( "  par: " );
    for( i=0; i<n_par; ++i )
        printf( " %12g", par[i] );
    printf ( " => norm: %12g\n", lm_enorm( m_dat, fvec ) );
    /*
    if ( iflag == -1 ) {
        printf( "  fitting data as follows:\n" );
        for( i=0; i<m_dat; ++i ) {
//            t = (mydata->t)[i];
            y = (mydata->y)[i];
//            f = mydata->user_func( t, par );
            printf( "    t[%2d]=%12g y=%12g fit=%12g residue=%12g\n",
                    i, t, y, f, y-f );
        }
    }
*/
}

FitterGaus::FitterGaus(unsigned int fit_points)
    : FitterBase(fit_points)
{
    ws = lm_alloc_ws (fit_points, 3);
}

int FitterGaus::fit(int fixpar, const double *x, int len, double *par, fitstatus_t *fs)   // approximation
{
    int i, nmp;
    double mp[3];
    lm_control_type control;
    fitdata_gaus_t fitdata;

    if (!x || !par || (len<=0))
        return (-1);

    if (!ws) return(-1);

    if ((unsigned int)len > getMaxFitPoints()) len = getMaxFitPoints();

    for (i = 0; i < 3; i++) {
        par[i] = 0;
        //    err[i] = 0;
    }

    lm_initialize_control (&control);
    fitdata.y = x;
    fitdata.fix = fixpar;
    if (fs) control.maxcall = fs->niter;
    FindGausApproxD (x, len, par);
    nmp=0;
    fitdata.fix=fixpar;
    if (!(fixpar&1)) {mp[nmp] = par[0]; nmp++;} else {fitdata.fp[0] = par[0];}
    if (!(fixpar&2)) {mp[nmp] = par[1]; nmp++;} else {fitdata.fp[1] = par[1];}
    if (!(fixpar&4)) {mp[nmp] = par[2]; nmp++;} else {fitdata.fp[2] = par[2];}
    lm_minimize (ws, len, nmp, mp, lm_evaluate_gaus, lm_print_none, &fitdata, &control);
    //  lm_minimize (ws, len, nmp, mp, lm_evaluate_gaus, lm_print_default, &fitdata, &control);
    nmp=0;
    if (!(fixpar&1)) {par[0] = mp[nmp]; nmp++;}
    if (!(fixpar&2)) {par[1] = mp[nmp]; nmp++;}
    if (!(fixpar&4)) {par[2] = mp[nmp]; nmp++;}
    if (fs) {
        fs->fnorm = control.fnorm;
        fs->niter= control.nfev;
    }
    //  printf ("status: %s after %d evaluations, fnorm=%e\n", lm_shortmsg[control.info], fs->niter, fs->fnorm);
    if (control.info <= 0) return(2);
    //  lm_free_ws(ws);
    return (0);
}

int QuickFitGaus (const double *x, int len, double *par, fitstatus_t *fs) {

    lm_control_type lmc;
    fitdata_gaus_t lmfd;
    struct lm_ws_t *lmws;

    if (len<=0) return(-1);

    lmws = lm_alloc_ws (len, 3);
    if (!lmws) return(-1);

    lm_initialize_control(&lmc);
    lmfd.y=x;
    lmfd.fix=0;

    FindGausApproxD(x,len,par);
    lm_minimize (lmws, len, 3, par, lm_evaluate_gaus, lm_print_none, &lmfd, &lmc);
    lm_free_ws(lmws);
    if (fs) {
        fs->fnorm = lmc.fnorm;
        fs->niter= lmc.nfev;
    }

    return(0);
}

// New interface

double GausFcn(const GausPar &par, double x)
{
    return par.a * gauss(x-par.x0, par.sigma);
}

double GaussianApproximator::FindGausDevNorm (const std::vector<int> &x, const GausPar &par)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausDevNorm(xd, par);
}

double GaussianApproximator::FindGausDevNorm (const std::vector<double> &x, const GausPar &par)
{
    const size_t len = x.size();
    double rms = 0;
    for (size_t i = 0; i < len; i++) {
        rms += pow(x[i] - GausFcn(par, i), 2);
    }
    return sqrt(rms);
}

double GaussianApproximator::FindGausDevNormShifted (const std::vector<int> &x, const GausPar &par)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausDevNormShifted(xd, par);
}

double GaussianApproximator::FindGausDevNormShifted (const std::vector<double> &x, const GausPar &par)
{
    const size_t len = x.size();
    double rms = 0;
    for (size_t i = 0; i < len; i++) {
        rms += pow(x[i] - GausFcn(par, i+0.5), 2);
    }
    return sqrt(rms);
}

/** @brief Gaussian function approximation, x shifted by 0.5
 *
 *  @par[in] x data values defined at 0.5, 1.5, 2.5, ...
 */
GausPar GaussianApproximator::FindGausApproxShifted (const std::vector<double> &x)
{
    GausPar par;
    const size_t len = x.size();
    double allcha = 0, sumx = 0, sumx2 = 0;

    for (size_t i = 0; i < len; i++) {
        const double binCenter = i + 0.5;
        const double &val = x[i];
        sumx += val * binCenter;
        sumx2 += val * binCenter * binCenter;
        allcha += val;
    }
    if (allcha > 0) {
        const double mean = sumx / allcha;
        double rms = sqrt (sumx2 / allcha - mean * mean);
        if (rms < M_1_SQRT2PI) rms = M_1_SQRT2PI; // force lower limit
        par.a = M_1_SQRT2PI * allcha / rms;
        par.x0 = mean;
        par.sigma = rms;
    } else {
        par.a = 0;
        par.x0 = len/2 + 0.5;
        par.sigma = len/2;
    }
    return par;
}

GausPar GaussianApproximator::FindGausApproxShifted (const std::vector<int> &x)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausApproxShifted(xd);
}

/** @brief Gaussian function approximation
 *
 *  @par[in] x data values defined at 0, 1, 2, ...
 */
GausPar GaussianApproximator::FindGausApprox (const std::vector<int> &x)
{
    const std::vector<double> xd(x.begin(), x.end());
    return FindGausApprox(xd);
}

GausPar GaussianApproximator::FindGausApprox (const std::vector<double> &x)
{
    GausPar par;
    const size_t len = x.size();
    double allcha = 0, sumx = 0, sumx2 = 0;
    for (size_t i = 0; i < len; i++) {
        const double val = x[i];
        const double binCenter = i;
        sumx += val * binCenter;
        sumx2 += val * pow(binCenter, 2);
        allcha += val;
    }
    {
        if (allcha > 0) {
            const double mean = sumx / allcha;
            double rms = sqrt (sumx2 / allcha - mean * mean);
            if (rms < M_1_SQRT2PI) rms = M_1_SQRT2PI; // force lower limit
            par.a = M_1_SQRT2PI * allcha / rms;
            par.x0 = mean;
            par.sigma = rms;
        } else {
            par.a = 0;
            par.x0 = len/2;
            par.sigma = len/2;
        }
    }
    return par;
}

} // namespace dsplib
