#include "SdbTypes.h"

#include "sdb_util.h"
#include "sdb_rom.h"

namespace {

constexpr const uint64_t VENDOR_ID = 0x414649; // 'AFI'

enum mlink_bus_type {BUS_REGIO = 1, BUS_MEMIO = 2};

QString nameToStr(const uint8_t *name, size_t size)
{
    char buf[size+1];
    buf[size] = '\0';
    sdb_copy_printable(buf, name, size, ' ');
    return QString::fromLatin1(buf).trimmed();
}

#define SDBTEXT_TO_STRING(x) nameToStr(x, sizeof(x))

QString dateToStr(uint32_t date)
{
    char buf[20];
    snprint_sdb_date(buf, sizeof(buf), date);
    return QString::fromLatin1(buf);
}

QString versionToStr(uint32_t version)
{
    char buf[20];
    snprint_sdb_version(buf, sizeof(buf), version);
    return QString::fromLatin1(buf);
}

} // namespace

static SdbProduct parse_sdb_product(const struct sdb_product &sdb_product)
{
    SdbProduct product;
    product.vendor_id = sdb_product.vendor_id;
    product.device_id = sdb_product.device_id;
    product.version = versionToStr(sdb_product.version);
    product.date = dateToStr(sdb_product.date);
    product.name = SDBTEXT_TO_STRING(sdb_product.name);
    return product;
}

sdb_t sdb_t::parse_rom_data(const sdb_rom_t &r)
{
    sdb_rom_t rom = r;
    sdb_t sdb;
    sdb.crc_present = sdb_checksum_present(&r);
    sdb.crc_valid = sdb_validate_checksum(&r);

    auto &ic = rom.ic;
    sdb.data.ic.valid = (ic.sdb_magic != SDB_MAGIC);
    auto &syn = rom.syn;
    sdb_interconnect_fix_endian(&ic);
    sdb_synthesis_fix_endian(&syn);
    const auto deviceCount = std::min(std::max(0, ic.sdb_records-1), (int)MAX_SDB_DEVICE_COUNT);
    for (int i=0; i<deviceCount; i++) {
        sdb_device &dev = rom.device[i];
        if (dev.sdb_component.product.record_type == sdb_type_device) {
            sdb_device_fix_endian(&dev);
            SdbDevice sdbDevice;
            sdbDevice.abi_ver_major = dev.abi_ver_major;
            sdbDevice.abi_ver_minor = dev.abi_ver_minor;
            sdbDevice.bus_specific = dev.bus_specific;
            sdbDevice.component.addr_first = dev.sdb_component.addr_first;
            sdbDevice.component.addr_last = dev.sdb_component.addr_last;
            sdbDevice.component.product = parse_sdb_product(dev.sdb_component.product);
            sdb.data.devices.push_back(sdbDevice);
        }
    }
    sdb.data.ic.sdb_version = ic.sdb_version;
    sdb.data.ic.sdb_bus_type = ic.sdb_bus_type;
    sdb.data.ic.component.addr_first = ic.sdb_component.addr_first;
    sdb.data.ic.component.addr_last = ic.sdb_component.addr_last;
    sdb.data.ic.component.product = parse_sdb_product(ic.sdb_component.product);

    sdb.data.syn.commit_id = SDBTEXT_TO_STRING(syn.commit_id);
    sdb.data.syn.user_name = SDBTEXT_TO_STRING(syn.user_name);
    sdb.data.syn.date = dateToStr(syn.date);
    sdb.data.syn.valid = (syn.record_type == sdb_type_synthesis);

    sdb.valid = sdb.ic().valid;
    return sdb;
}

bool sdb_t::isValid() const
{
    return valid;
}

bool sdb_t::crcPresent() const
{
    return crc_present;
}

bool sdb_t::crcOk() const
{
    return crc_valid;
}

SdbInterconnect sdb_t::ic() const
{
    return data.ic;
}

QVector<SdbDevice> sdb_t::devices() const
{
    return data.devices;
}

SdbSynthesis sdb_t::syn() const
{
    return data.syn;
}

QList<AFI_SDB_ID> sdb_t::getDeviceIds() const
{
    QList<AFI_SDB_ID> result;
    for (const auto &device : data.devices) {
        auto id = static_cast<AFI_SDB_ID>(device.component.product.device_id);
        result.push_back(id);
    }
    return result;
}

bool sdb_t::hasSdbDevice(AFI_SDB_ID id) const
{
    if (!isValid())
        return false;
    for (const auto &device : data.devices)
        if (device.component.product.device_id == id)
            return true;

    return false;
}

QPair<quint16, quint16> sdb_t::getSdbAddrRange(AFI_SDB_ID id, int seqNum) const
{
    if (!isValid())
        return {0, 0};

    for (const auto &device : data.devices) {
        if (device.component.product.device_id == id && seqNum-- == 0) {
            const auto &sdb_comp = device.component;
            return {static_cast<quint16>(sdb_comp.addr_first >> 1),
                    static_cast<quint16>(sdb_comp.addr_last >> 1)};
        }
    }
    return {0, 0};
}

quint16 sdb_t::getSdbBaseAddr(AFI_SDB_ID id, int seqNum) const
{
    return getSdbAddrRange(id, seqNum).first;
}

QString sdb_t::getProductSignature() const
{
    if (!ic().isValid())
        return {};
    auto type = ic().productName().toLower();
    return type.replace('-', '_');
}

QList<SdbDevice> sdb_t::filterDevices(const QList<AFI_SDB_ID> &search) const
{
    if (!isValid()) {
        return {};
    }
    QList<SdbDevice> result;
    for (const auto &device : data.devices) {
        auto id = static_cast<AFI_SDB_ID>(device.component.product.device_id);
        if (!search.contains(id))
            continue;
        result.push_back(device);
    }
    return result;
}
