#include "AdcSpiAds52J90Module.h"

#include <QDebug>
#include <QTime>
#include <QThread>

#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace  {
enum {
    REG_ID                  = 0x00,
    REG_ADC_CH_INFO         = 0x02,
    REG_ADC_BITS            = 0x03,

    REG_SPI_ADDR            = 0x8,
    REG_SPI_WR_DATA         = 0x9,
    REG_SPI_READ_DATA       = 0xA,
    REG_SPI_ACTIVE_SELECT   = 0xB,
    REG_SPI_ACTIVE_CSB      = 0xC
};

enum {
    REG_AD52J90_READ_EN = 0x0,
    REG_AD52J90_INPUT_MODE = 0x1,
    REG_AD52J90_OPERATION_MODE = 0x2,
    REG_AD52J90_SER_DATA_RATE = 0x3,
    REG_AD52J90_ADC_RES = 0x4,
    REG_AD52J90_CUSTOM_PATTERN = 0x5,
    REG_AD52J90_AUTO_OFFSET_REMOVAL = 0xA,
};
}

AdcSpiAds52J90Module::AdcSpiAds52J90Module(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        if (getDeviceId() == DEVICE_ID_PHOS_FEC)
            setBaseAddr(0x4100, 0xFF);
    }
}

QList<AFI_SDB_ID> AdcSpiAds52J90Module::getSupportedSdbIds()
{
    return {AFI_SDB_ADS52J90_SPI};
}

void AdcSpiAds52J90Module::post_init()
{
    quint16 chInfo, bits;
    RegOpVector r;
    r.RegRead16(REG_ADC_CH_INFO, &chInfo);
    r.RegRead16(REG_ADC_BITS, &bits);
    if(!regOpExecRebased(r))
        return;
    adcChPerChip = chInfo >> 8;
    adcChips = chInfo & 0xFF;
    adcDataBits = bits;

    setupAdc();
}

bool AdcSpiAds52J90Module::spi_test_impl(TestResult *result)
{
    TestResult testResult;

    std::vector<quint16> testData(adcChips);
    for (int i = 0; i < adcChips; i++) {
        testData[i] = std::rand() & 0xFFFF; //need to mask by reg size
        testResult.map[i] = true;
    }

    for (int adcId = 0; adcId < adcChips; ++adcId) {
        quint16 active_adc_mask = 1 << adcId;

        quint16 rand_val = testData[adcId];
        if (!spi_write(REG_AD52J90_CUSTOM_PATTERN, rand_val, active_adc_mask))
            return {};
        quint16 readregval;
        if (!spi_read(REG_AD52J90_CUSTOM_PATTERN, &readregval, active_adc_mask))
            return {};
        if (rand_val != readregval) {
            testResult.map[adcId] = false;
            qWarning() << __PRETTY_FUNCTION__ << getDeviceIndex().getIdent()
                       << hex << adcId << active_adc_mask << rand_val << readregval << (rand_val ^ readregval);
        }
    }
    if (result)
        *result = testResult;
    return true;
}

bool AdcSpiAds52J90Module::setupAdc()
{
    bool ok = true;
    qDebug() << "ADS52J90 setup";

    ok &= spi_test();
    if(!ok)
        return ok;

    const quint16 &adc_mask=getAdcMask();
    ok &= spi_write(REG_AD52J90_AUTO_OFFSET_REMOVAL, 0x3000, adc_mask); //after reset
    quint16 data1h = 0;
    if (adcChPerChip == 0)
        adcChPerChip = 16; // default
    switch (adcChPerChip) {
    case 8: data1h = 0x0094; break;
    case 16: data1h = 0x0014; break;
    case 32: data1h = 0x0000; break;
    default:;
    }
    ok &= spi_write(REG_AD52J90_INPUT_MODE, data1h, adc_mask);
    ok &= spi_write(REG_AD52J90_OPERATION_MODE, 0x1000, adc_mask); // normal data operation

    quint16 adcSerialization;
    quint16 adcResolution;
    switch(adcDataBits) {
    case 10:
        adcSerialization = 0x6000; // 10X
        adcResolution = 0x3;
        break;
    case 14:
        adcSerialization = 0x2000; // 14X
        adcResolution = 0x11;
        break;
    case 16:
        adcSerialization = 0x8000; // 16X
        adcResolution = 0x1;
        break;
    case 12:
    default:
        adcSerialization = 0; // 12X
        adcResolution = 0;
    }
    ok &= spi_write(REG_AD52J90_SER_DATA_RATE, adcSerialization, adc_mask);
    ok &= spi_write(REG_AD52J90_ADC_RES, adcResolution, adc_mask);

    //setCustomPatternMode(true);
    //setCustomPattern(0x40);
    return ok;
}

bool AdcSpiAds52J90Module::setCustomPatternMode(bool mode)
{
    return spi_write(REG_AD52J90_OPERATION_MODE, mode ? 0x1180 : 0x1000, 0xFF);
}

bool AdcSpiAds52J90Module::setCustomPattern(quint16 data)
{
    data <<= 16-adcDataBits;
    return spi_write(REG_AD52J90_CUSTOM_PATTERN, data, 0xFF);
}

bool AdcSpiAds52J90Module::setPRBSMode(bool en)
{
    quint16 regOld, regNew;
    auto ret = spi_read(REG_AD52J90_ADC_RES, &regOld, 1); // get original reg val from firt ADC
    if(!ret)
        return false;
    constexpr quint16 PRBS_EN_BIT = 0x20;
    regNew = en ? (regOld | PRBS_EN_BIT) : (regOld & ~PRBS_EN_BIT);
    if(regOld == regNew)
        return true;

    bool ok = spi_write(REG_AD52J90_ADC_RES, regNew, getAdcMask());
    ok &= spi_write(REG_AD52J90_OPERATION_MODE, en ? 0x11C0 : 0x1000, getAdcMask());
    return ok;
}

bool AdcSpiAds52J90Module::spi_write(quint16 addr, quint16 data, quint16 device_mask)
{
    bool ok = true;

    RegOpVector r;
    r.RegWrite16(REG_SPI_ACTIVE_SELECT, device_mask);
    r.RegWrite16(REG_SPI_WR_DATA, data);
    r.RegWrite16(REG_SPI_ADDR, addr);
    ok &= regOpExecRebased(r);

    ok &= spi_csb_check();
    return ok;
}

bool AdcSpiAds52J90Module::spi_read(quint16 addr, quint16 *result, quint16 device_mask)
{
    bool ok = true;
    quint16 readval;
    ok &= spi_write(REG_AD52J90_READ_EN, 0x2, device_mask);
    ok &= regWrite16(REG_SPI_ADDR, addr);
    ok &= spi_csb_check();
    ok &= regRead16(REG_SPI_READ_DATA, &readval);
    ok &= spi_write(REG_AD52J90_READ_EN, 0x0, device_mask);

    if (ok && result)
        *result = readval;
    return ok;
}

bool AdcSpiAds52J90Module::spi_csb_check()
{
    quint16 check_val;
    const quint16 full_mask = (1<<adcChips)-1;
    for (int i=0; i<10; i++) {
        regRead16(REG_SPI_ACTIVE_CSB, &check_val);
        if (check_val == full_mask)
            return true;
        usleep(i*1000);
    }
    qWarning() << "SPI CSB CHECK ERROR" << hex << check_val;
    return false;
}

