#include "ClockControlModule.h"

#include "ClockControl/ClockControlConfig.h"
#include "regio/RegOpStream.h"
#include "regio/regop.h"

#include "ClockStatus.h"
#include "sdb/AfiSdbId.h"

enum {
    REG_CLOCK_CONTROL   = 0x00, // R/W
    REG_CLOCK_STATUS    = 0x01, // RO
    REG_CLOCK_SOURCE    = 0x02, // RO

    REG_V2_PLL_BYPASS   = 0x04, // R/W
    REG_V2_CONTROL      = 0x05, // R/W
    REG_V2_STATUS       = 0x06, // RO
    REG_V2_SYS_SOURCE   = 0x07, // RO
    REG_V2_CAPABILITY   = 0x08, // RO Static
};

QList<AFI_SDB_ID> ClockControlModule::getSupportedSdbIds()
{
    return { AFI_SDB_Clock_Control };
}

void ClockControlModule::post_init()
{
    readBuildParams();
}

bool ClockControlModule::writeConfig(const ClockControlConfig &config)
{
    RegOpVector r;
    quint16 pll_bypass = config.pll_bypass ? 1 : 0;
    if(isV2()) {
        auto source = static_cast<quint16>(config.source)
                & buildParams->clockCapability;
        if(!buildParams || (source == 0))
            return {};
        r.RegWrite16(REG_V2_PLL_BYPASS, pll_bypass);
        r.RegWrite16(REG_V2_CONTROL, source);
    } else {
        r.RegWrite16(REG_CLOCK_CONTROL, pll_bypass);
    }

    return regOpExecRebased(r);
}

mlink::RegOpVector ClockControlModule::getStatusRegOp() const
{
    RegOpVector r;
    r.RegRead16(REG_CLOCK_CONTROL);
    r.RegRead16(REG_CLOCK_STATUS);
    r.RegRead16(REG_CLOCK_SOURCE);
    return r;
}

std::optional<ClockStatus> ClockControlModule::readStatus()
{
    ClockStatus st;
    if(isV2()) {
        RegOpVector r;
        quint16 val[2];
        r.RegRead16(REG_V2_STATUS, &val[0]);
        r.RegRead16(REG_V2_SYS_SOURCE, &val[1]);

        if (regOpExecRebased(r)) {
            st.selected_clock = static_cast<ClockControlType>(val[1] & 0x7FFF);
            st.pll_bypass = val[1] & 0x8000;
            return st;
        }
    } else {
        auto r = getStatusRegOp();
        if (!regOpExecRebased(r))
            return {};
        auto status = decodeStatus(r);
        if (!status.has_value())
            return {};
        return status;
    }
    return {};
}

RegOpStream &operator>>(RegOpStream &in, ClockStatus &status)
{
    reg_clock_control clock_control = {};
    in >> clock_control.all;
    reg_clock_status clock_status = {};
    in >> clock_status.all;
    reg_clock_source clock_source = {};
    in >> clock_source.all;

    // decode
    status.pll_bypass = clock_control.bit.pll_bypass;
    status.clock_valid[ClockControlRef41_6] = clock_status.bit.ttc_locked;
    status.clock_valid[ClockControlVXS] = clock_status.bit.vxs_locked;
    if (clock_source.bit.vxs)
        status.selected_clock = ClockControlVXS;
    else if (clock_source.bit.ttc)
        status.selected_clock = ClockControlRef41_6;
    else
        status.selected_clock = ClockControlLocal;

    return in;
}
std::optional<ClockStatus> ClockControlModule::decodeStatus(const RegOpVector &result, int posOffset)
{
    const auto &chunk = RegOpVector(result.begin() + posOffset, result.end());
    RegOpStream in(chunk);
    ClockStatus status;
    in >> status;
    if (!in.isValid())
        return {};
    return status;
}

bool ClockControlModule::readBuildParams()
{
    buildParams.reset();
    if(isV2()) {
        ClockBuildParams bp;
        RegOpVector r;
        r.RegRead16(REG_V2_CAPABILITY, &bp.clockCapability);

        if(!regOpExecRebased(r))
            return {};

        buildParams = bp;
    }
    return true;
}

bool ClockControlModule::isV2() const
{
    return majorVersion() >= 2;
}
