
#include <QDebug>
#include <QHostAddress>

#include "NetworkPortStatus.h"
#include "base/fw_version.h"

template<int N>
static QString formatCounters(const std::array<uint32_t, N> &arr,
                              const std::array<const char *, N> &labels,
                              QString title)
{
    int maxlen = 12;
    for (int i = 0; i < N; i++) {
        int len = QString(labels[i]).length();
        if (len > maxlen)
            maxlen = len;
    }

    QStringList cl;
    for (int i = 0; i < N; i++) {
        cl << QString("%1 %2").arg(labels[i], -maxlen).arg(arr[i], -11);
    }
    QStringList lines;
    int n = 0;
    QString indent = "  ";
    QStringList line;
    for (const auto &s : cl) {
        line << s;
        if (n % 4 == 3) {
            lines << (indent + line.join(""));
            line.clear();
        }
        n++;
    }
    if (!line.empty())
        lines << (indent + line.join(""));

    QString result = lines.join("\n");
    result = "<span style=\"font-family:monospace;white-space:pre;font-size:0.9em\">" + result + "</span>";
    if (!title.isNull())
        result = title + "\n" + result;
    return result;
}

static const char *hilight_error = "<span style=\"color:red;\">";
static const char *hilight_warn = "<span style=\"color:yellow;\">";
static const char *hilight_close = "</span>";

QString formatHtmlErrorNumber(int number)
{
    auto s = QString::number(number);
    if (number > 0)
        s = hilight_error + s + hilight_close;
    return s;
}

QString PhyPortStatus::toString() const
{
    if (!isValid())
        return "[invalid]";
    QStringList sl;
    // sl << "status: " + QString("%1").arg(status, 8, 16, QChar('0'));
    sl << " SFP (sfp_present: " + QString::number(status_reg().sfp_present)
            + ", osd: " + QString::number(status_reg().sfp_osd)
            + ", tx_fault: " + QString::number(status_reg().sfp_tx_fault)
            + ", tx_disable: " + QString::number(status_reg().tx_disable) + ")";
    sl << " GTX (speed: "
            + QString::number(status_reg().speed)
            // + ", rx_error: " + QString::number(status_reg().gtx_rx_error)
            + ", phy_reset: " + QString::number(status_reg().gtx_phy_reset)
            + ", resetdone: " + QString::number(status_reg().gtx_resetdone)
            + ", mmcm_locked: " + QString::number(status_reg().gtx_mmcm_locked) + ")";
    return sl.join("\n");
}

QDebug operator<<(QDebug dbg, const PhyPortStatus &r)
{
    dbg << "phy: " + r.toString() + ";";
    return dbg;
}

static QString decode_speed(int speed)
{
    switch (speed) {
    case 0: return "10M";
    case 1: return "100M";
    case 2: return "1G";
    case 3: return "10G";
    default: return "unknown";
    }
}

QString PcspmaPortStatus::toString() const
{
    if (!isValid())
        return "[invalid]";
    QStringList sl;
    // sl << QString("status: %1").arg(status, 8, 16, QChar('0'));
    sl << " PCS (osd: " + QString::number(status_reg().rx_sig_det)
            // + ", pcs_rx_locked_ll: " + QString::number(status_reg().pcs_rx_locked_ll)
            + ", pcs_rx_locked: " + QString::number(status_reg().pcs_rx_locked)
            // + ", pcs_rx_link_status: " + QString::number(status_reg().pcs_rx_link_status)
            + ", teng_pcs_rx_link_status: " + QString::number(status_reg().teng_pcs_rx_link_status)
            + ")";
    sl << " PMA (speed: " + decode_speed(status_reg().speed)
            // + ", duplex: " + QString::number(status_reg().duplex)
            // + ", pma_link_status: " + QString::number(status_reg().pma_link_status)
            + ", pcs_hiber: " + formatHtmlErrorNumber(status_reg().pcs_hiber)
            + ", pcs_rx_hiber_lh: " + formatHtmlErrorNumber(status_reg().pcs_rx_hiber_lh)
            + ", pcs_ber_count: " + formatHtmlErrorNumber(status_reg().pcs_ber_count)
            + ", pcs_test_patt_err_count: " + formatHtmlErrorNumber(status_reg().pcs_test_patt_err_count)
            + ")";

    sl << formatCounters<counters_size>(counters,
                                        {"signal_loss", "link_down", "bit_error", "pcs_unlock"},
                                        {});
    sl << formatCounters<timers_size>(timers,
                                      {"uptime", "signal_flap", "link_flap", "bit_error"},
                                      "Timers");
    return sl.join("\n");
}

QDebug operator<<(QDebug dbg, const PcspmaPortStatus &r)
{
    dbg << "pcspma: " + r.toString() + ";";
    return dbg;
}

QString MacPortStatus::toString() const
{
    if (!isValid())
        return "[invalid]";
    QStringList sl;
    // sl << QString("status: %1").arg(status, 8, 16, QChar('0'));
    sl << QString("mtu: %1").arg(mtu);
    sl << formatCounters<counters_size>(
        counters,
        {
            "rx_valid", "rx_error", "tx_valid", "tx_error",
            "crc_error", "rx_pause", "local_fault", "remote_fault",
            "rx_overrun", "rx_underrun", "tx_overrun", "tx_underrun",
            "rx_overrun", "tx_overrun"
        },
        {});
    return sl.join("\n");
}

QDebug operator<<(QDebug dbg, const MacPortStatus &r)
{
    dbg << "mac: " + r.toString() + ";";
    return dbg;
}

QString HwipPortStatus::toString() const
{
    if (!isValid())
        return "[invalid]";
    QStringList sl;
    // sl << QString("status: %1").arg(status, 8, 16, QChar('0'));
    // Network addresses
    sl << QString("device ip: %1, mac: %2")
              .arg(QHostAddress(host_ip).toString())
              .arg(host_mac, 12, 16, QChar('0'));
    sl << QString("master peer: %1:%2, mstream peer: %3:%4")
              .arg(QHostAddress(master_ip).toString()).arg(master_port)
              .arg(QHostAddress(mstream_ip).toString()).arg(mstream_port);
    // Bus errors
    if (regio_error_code & 0x003F) {
        QStringList el;
        QString rw_str = (regio_error_code & 0x0040) ? "write": "read";
        QString addr_str = QString("%1").arg(regio_error_addr, 8, 16, QChar('0'));
        QString descr = rw_str + " address " + addr_str;
        bool master_error = regio_error_code & 0x0007;
        bool slave_error = regio_error_code & 0x0018;
        bool aborted_error = regio_error_code & 0x0020;
        if (master_error) {
            QStringList ll;
            if (regio_error_code & 0x0001) ll << "write";
            if (regio_error_code & 0x0002) ll << "addr";
            if (regio_error_code & 0x0003) ll << "wdata";
            el << QString("Bus master error at %1: %2 changed while strobe")
                      .arg(descr, ll.join(", "));
        }
        if (slave_error)
            el << QString("Bus slave error at %1:%2%3")
                      .arg(descr)
                      .arg((regio_error_code & 0x0008) ? " rdata changed while strobe" : "")
                      .arg((regio_error_code & 0x0010) ? " rdy deasserted while strobe" : "");
        if (aborted_error)
            el << QString("Bus error: cycle aborted at %1")
                      .arg(descr);
        if (!el.empty())
            sl << hilight_error + el.join("\n") + hilight_close;
    }
    // Configuration
    sl << QString("mlink_config: %1")
              .arg(mlink_config, 8, 16, QChar('0'));
    // Counters
    sl << formatCounters<counters_size>(counters, {"reg_error", "reg_timeout", "retx_overrun"}, {});
    // Timers
    sl << formatCounters<timers_size>(
        timers,
        {"uptime", "ip_assigned", "master_lock", "mstream_lock",
            "reg_error", "reg_timeout", "retx_overrun"},
        "Timers");
    return sl.join("\n");
}

QDebug operator<<(QDebug dbg, const HwipPortStatus &r)
{
    dbg << "hwip: " + r.toString() + ";";
    return dbg;
}

int MstreamPortStatus::getRetxBufCount() const
{
    return 1 << (retx_config & 0x000F);
}

int MstreamPortStatus::getRetxBufSize() const
{
    return 4 * (1 << ((retx_config >>4) & 0x000F));
}

int MstreamPortStatus::getAckSizeLimit() const
{
    if (retx_config == 0)
        return 16; // legacy compatibility
    return (mtu - 20 - 8 - 16 - 4) / 4; // ip+udp+mlink+1
}

QString MstreamPortStatus::toString() const
{
    if (!isValid())
        return "[invalid]";
    QStringList sl;
    sl << QString("MStream %1, MTU: %2 (%3-%4), max_packet_size %5, defragmentation timeout: %6 ms")
              .arg(fw_version_t(version(), 0).toString())
              .arg(mtu).arg(mtu_min).arg(mtu_max)
              .arg(max_packet_size).arg(defrag_timeout * 0.001);

    // Configuration
    QString retx_details;
    if (retx_config) {
        int nbuf = getRetxBufCount();
        int bufsize = getRetxBufSize();
        retx_details = QString("%1 buffers x %2 bytes")
                           .arg(nbuf)
                           .arg(bufsize);
    }
    auto s = retx_details;
    if (clk_khz)
        s += ", " + QString("retx_ack_time: %1 ms, est_rtt: %2 ms").arg(retx_ack_time*1.0/clk_khz).arg(retx_est_rtt*1.0/clk_khz);
    sl << s;
    // Counters
    sl << formatCounters<counters_size>(
        counters,
        {"in_words", "out_words", "in_pkts", "split_pkts",
         "merge_pkts", "slow_pkts", "pass_pkts", "ready_pkts",
         "sent_pkts", "noack_pkts", "rx_ack_ok", "retx_timeout",
         "retx_error", "in_error", "len_error", "rx_ack_bad"},
        {});
    // Timers
    sl << formatCounters<timers_size>(
        timers,
        {"uptime", "in_pkts", "merge_pkts", "out_pkts",
         "rx_act", "retx_error", "retx_timeout", "rx_hdr_error"},
        "Timers");
    return sl.join("\n");
}

QDebug operator<<(QDebug dbg, const MstreamPortStatus &r)
{
    dbg << "mstream: " + r.toString() + ";";
    return dbg;
}

bool NetworkPortStatus::isValid() const
{
    return phy.isValid() && pcspma.isValid() && mac.isValid() && hwip.isValid();
}

QString NetworkPortStatus::toString() const
{
    QStringList sl;
    sl << phy.toString();
    sl << pcspma.toString();
    sl << mac.toString();
    sl << hwip.toString();
    sl << mstream.toString();
    return sl.join("\n");
}

QDebug operator<<(QDebug dbg, const NetworkPortStatus &r)
{
    dbg << "NetworkPort:" + r.toString();
    return dbg;
}
