/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigCtrlStatModule.h"

#include "TrigCtrlStat/TrigCtrlStatConfig.h"
#include "base/DeviceId.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
enum {
    REG32_ID                       = 0x00,
    REG_CONTROL                    = 0x02,
    REG_LOCK                       = 0x03,
    REG32_TRIG_SRC_MASK            = 0x04,
    REG32_TRIG_SRC_CAPS            = 0x06,
    REG32_TRIG_POL_MASK            = 0x08,
    REG32_TRIG_POL_CAPS            = 0x0A,
    REG32_TIMER_CLOCK_HZ           = 0x0C,
    REG32_PERIODIC_PULSER_PERIOD   = 0x0E,
    REG32_RANDOM_PULSER_PERIOD     = 0x10,
    REG32_RANDOM_PULSER_MIN_PERIOD = 0x12,
    REG32_PROGRAM_DEADTIME         = 0x14,
    REG_TRIG_QUEUE                 = 0x16,
    REG_BUSY_STATUS                = 0x17,
    REG_DIG_THR_DELAY              = 0x18, // 8ns step, 1us max
    REG_ANALOG_THR_DELAY           = 0x19, // 8ns step, 1us max
    REG64_DIG_THR_CH_MASK          = 0x20,
    REG64_DIG_THR_CH_POL           = 0x28,
    REG64_ANALOG_THR_CH_MASK       = 0x30,
    REG64_ANALOG_THR_CH_POL        = 0x38
};

inline constexpr const uint16_t REG_CONTROL_BIT_RESET = 1;
inline constexpr const uint16_t REG_LOCK_BIT_LOCK = 1;

} // namespace

TrigCtrlStatModule::TrigCtrlStatModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        if (getDeviceId() == DEVICE_ID_PHOS_FEC)
            setBaseAddr(0x100, 0xFF);
    }
}

QList<AFI_SDB_ID> TrigCtrlStatModule::getSupportedSdbIds()
{
    return {AFI_SDB_TDC_Trigger_Control};
}

void TrigCtrlStatModule::post_init()
{
    TrigCtrlBuildParams bp;
    RegOpVector r;
    r.RegRead32(REG32_TRIG_SRC_CAPS, &bp.sourceCaps);
    r.RegRead32(REG32_TRIG_POL_CAPS, &bp.polarityCaps);
    r.RegRead32(REG32_TIMER_CLOCK_HZ, &bp.timerClock);
    if(regOpExecRebased(r))
        buildParam = bp;
    else
        buildParam.reset();
}

bool TrigCtrlStatModule::softreset()
{
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, REG_CONTROL_BIT_RESET);
    r.RegWrite16(REG_CONTROL, 0);
    return regOpExecRebased(r);
}

bool TrigCtrlStatModule::writeConfig(const TrigCtrlStatConfig &newConfig)
{
    if(!buildParam)
        return false;

    auto timerClockHz = buildParam->timerClock;
    RegOpVector r;

    double periodD = timerClockHz / newConfig.trigTimerFreq;
    quint32 periodicPulserPeriod = std::clamp(periodD, 1.0, double(UINT32_MAX));
    r.RegWrite32(REG32_PERIODIC_PULSER_PERIOD, periodicPulserPeriod);
    periodD = newConfig.randomPulserFreq * (quint32)(-1) / timerClockHz;
    quint32 randomPulserPeriod = std::clamp(periodD, 1.0, double(UINT32_MAX));
    r.RegWrite32(REG32_RANDOM_PULSER_PERIOD, randomPulserPeriod);
    r.RegWrite32(REG32_RANDOM_PULSER_MIN_PERIOD, newConfig.randomPulserMinPeriod * timerClockHz);
    r.RegWrite32(REG32_PROGRAM_DEADTIME, newConfig.programmedDeadTime * timerClockHz);
    r.RegWrite16(REG_ANALOG_THR_DELAY, std::clamp(newConfig.analogThrDelay, (uint16_t)0, (uint16_t)0xFFFF));
    r.RegWrite16(REG_DIG_THR_DELAY, std::clamp(newConfig.digThrDelay, (uint16_t)0, (uint16_t)0xFFFF));
    auto trig_src = newConfig.source_mask;
    r.RegWrite32(REG32_TRIG_SRC_MASK, trig_src);
    r.RegWrite64(REG64_DIG_THR_CH_MASK, newConfig.dig_thr_channel_en);
    r.RegWrite64(REG64_ANALOG_THR_CH_MASK, newConfig.analog_thr_channel_en);
    return regOpExecRebased(r);
}

std::optional<TrigCtrlStatStatus> TrigCtrlStatModule::readStatus()
{
    TrigCtrlStatStatus status;
    if (!readStatus(&status))
        return {};
    statusCache = status;
    return status;
}

bool TrigCtrlStatModule::readStatus(TrigCtrlStatStatus *result)
{
    TrigCtrlStatStatus tmp;
    RegOpVector r;
    r.RegRead32(REG32_ID, &tmp.id);
    r.RegWrite16(REG_LOCK, 0); // update status
    r.RegWrite16(REG_LOCK, REG_LOCK_BIT_LOCK); // lock status
    r.RegRead16(REG_TRIG_QUEUE, &tmp.trigQueue);
    r.RegRead16(REG_BUSY_STATUS, &tmp.busyStatus);

    if (!regOpExecRebased(r))
        return {};
    statusCache = tmp;
    if (result)
        *result = tmp;
    return true;
}
