/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AbstractThresholdCalibrator.h"
#include <cmath>
#include <optional>

#include <QDebug>
#include <QLoggingCategory>

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc)
Q_LOGGING_CATEGORY(lc, "afi.cal")

static const int FAIL_WIDE_RANGE = 10; // max number of DAC bins of transition region
static const double eps = 1e-38;

/// Linear Least Squares fit
void LinearLSqFit(const double *x, const double *y, int n, double *a, double *b, double *c)
{
    // Find a,b for array of (x,y) of length n
    // y = ax + b
    int i;
    double xa = 0, ya = 0, stt = 0, xy = 0, aa, bb, chi2 = 0;

    if (n < 1)
        return;

    for (i = 0; i < n; i++) {
        xa += x[i];
        ya += y[i];
        //    printf("%d: %f, %f\n", i, x[i], y[i]);
    }
    xa /= n;
    ya /= n;
    for (i = 0; i < n; i++)
        stt += pow(x[i] - xa, 2);
    for (i = 0; i < n; i++)
        xy += (x[i] - xa) * y[i];
    if (stt != 0)
        bb = xy / stt;
    else
        bb = 0;
    aa = ya - xa * bb;
    for (i = 0; i < n; i++)
        chi2 += pow(y[i] - aa - bb * x[i], 2);
    //  printf("Line fit: a=%f, b=%f, chi2=%f\n", aa, bb, chi2);
    if (a)
        *a = bb;
    if (b)
        *b = aa;
    if (c)
        *c = chi2;
}

struct FindMediumResult
{
    bool isValid() const { return numPoints >= 2 && a < 0 && (chi2 / numPoints) < 0.5; }
    double dacValue() const { return (fabs(a) > eps) ? (0.5 - b) / a : 0; }

    int numPoints = 0;
    double a = 0;
    double b = 0;
    double chi2 = 0;
};

FindMediumResult findMediumValue(const QMap<int, double> &values)
{
    if (values.size() < 2)
        return {};
    QVector<double> px, py;
    for (auto it = values.cbegin(); it != values.cend(); ++it) {
        px.push_back(it.key());
        py.push_back(it.value());
    }
    // trim left
    while (py.size() > 2 && abs(py[0] - py[1]) < eps) {
        px.pop_front();
        py.pop_front();
    }
    // trim right
    while (py.size() > 2 && abs(py[py.size() - 1] - py[py.size() - 2]) < eps) {
        px.pop_back();
        py.pop_back();
    }
    FindMediumResult result;
    result.numPoints = py.size();
    LinearLSqFit(&px[0], &py[0], px.size(), &result.a, &result.b, &result.chi2);
    return result;
}

} // namespace

AbstractThresholdCalibrator::AllChannelState AbstractThresholdCalibrator::calcProbabilities(
    const QVector<quint64> &chStates, int nch)
{
    int num_samples = chStates.size();
    if (!num_samples)
        return {};
    AllChannelState states;
    for (int ch=0; ch<nch; ch++) {
        int sum = 0;
        for (int i=0; i<num_samples; i++)
            sum += (chStates[i] & (1ull<<ch)) ? 1 : 0;
        states.push_back((double)sum / num_samples);
    }
    return states;
}

AbstractThresholdCalibrator::AllChannelState AbstractThresholdCalibrator::searchStep(Workspace *workspace,
                                                                                     int dacVal)
{
    auto &chStates = workspace->dacToChState[dacVal];
    QMap<int, int> chThr;
    for (int ch = 0; ch < workspace->nch; ++ch)
        chThr[ch] = dacVal;
    if (!writeLevels(chThr))
        return {};
    if (!readStates(&chStates) || chStates.empty())
        return {};
    // qDebug() << dacVal << hex << chStates;
    return chStates;
}

int AbstractThresholdCalibrator::findNumberOfChannels()
{
    AllChannelState chStates;
    if (!readStates(&chStates))
        return {};
    return chStates.size();
}

AbstractThresholdCalibrator::RawResult AbstractThresholdCalibrator::Workspace::makeResult()
{
    AbstractThresholdCalibrator::RawResult rawResults;
    for (int ch = 0; ch < nch; ++ch) {
        QMap<int, double> values;
        for (auto it = dacToChState.cbegin(); it != dacToChState.cend(); ++it) {
            const quint16 dacCode = it.key();
            const double &state = it.value()[ch];
            values[dacCode] = state;
        }
        auto median = findMediumValue(values);
        if (median.isValid() && median.numPoints < FAIL_WIDE_RANGE) {
            rawResults[ch] = median.dacValue();
        } else {
            qWarning(lc).noquote() << QString("DAC Calibration failed for channel %1").arg(ch);
        }
    }
    if (rawResults.size() < nch)
        qCritical(lc).noquote() << "DAC calibration incomplete";

    return rawResults;
}

AbstractThresholdCalibrator::RawResult AbstractThresholdCalibrator::runSearch(Workspace *workspace,
                                                                              int minRawValue,
                                                                              int maxRawValue,
                                                                              int step)
{
    const int DAC_MIN_VALUE = minRawValue;
    const int DAC_MAX_VALUE = maxRawValue;
    const int SEARCH_START = (minRawValue + maxRawValue) / 2;

    for (int dacVal = SEARCH_START; dacVal < DAC_MAX_VALUE; dacVal += step) {
        const auto &chStates = searchStep(workspace, dacVal);
        // stop condition: all states high
        auto minValue = *std::min_element(chStates.cbegin(), chStates.cend());
        if (minValue > (1 - eps))
            break;
        if (abs(dacVal - SEARCH_START) > 100)
            step = 10;
    }

    for (int dacVal = SEARCH_START; dacVal >= DAC_MIN_VALUE; dacVal -= step) {
        const auto &chStates = searchStep(workspace, dacVal);
        if (dacVal == 0)
            break;
        // stop condition: all states low
        auto maxValue = *std::max_element(chStates.cbegin(), chStates.cend());
        if (maxValue < eps)
            break;
        if (abs(dacVal - SEARCH_START) > 100)
            step = 10;
    }
    return workspace->makeResult();
}

DacCalResult AbstractThresholdCalibrator::calibrateDac(int minRawValue, int maxRawValue, int step)
{
    auto nch = findNumberOfChannels();
    if (!nch)
        return {};
    Workspace workspace(nch);
    const auto &rawResults = runSearch(&workspace, minRawValue, maxRawValue, step);

    DacCalResult dacCalResult;
    for (auto it = rawResults.cbegin(); it != rawResults.cend(); ++it) {
        dacCalResult[it.key()] = convert_raw_to_mV(it.value());
    }
    return dacCalResult;
}
