#include "HptdcStatusModel.h"
#include "HptdcStatusModelItem.h"
#include "HptdcStatusMetrics.h"
#include "HptdcStatusMetrics.cpp"
#include <QBrush>
#include <QDebug>
#include <QFont>

HptdcStatusModel::HptdcStatusModel(QObject *parent)
    : QAbstractItemModel(parent)
    , rootItem(new HptdcStatusModelItem())
{
}

QVariant HptdcStatusModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::ForegroundRole:
        return getForegroundData(index);
    case Qt::FontRole:
        return fontData(index);
    }
    return QVariant();
}

QVariant HptdcStatusModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation == Qt::Horizontal
            && role == Qt::DisplayRole) {
        if(section == 0)
            return "Device serial / metric";
        else
            return "HPTDC #" + QString::number(section-1);
    }
    return QVariant();
}

QModelIndex HptdcStatusModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();

    const HptdcStatusModelItem *parentItem = parent.isValid() ?
        static_cast<HptdcStatusModelItem*>(parent.internalPointer()) :
        rootItem;
    HptdcStatusModelItem *childItem = parentItem->child(row);
    if(childItem) {
        QModelIndex index = createIndex(row, column, childItem);
        return index;
    }
    return QModelIndex();
}

QModelIndex HptdcStatusModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const HptdcStatusModelItem *childItem = static_cast<HptdcStatusModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    HptdcStatusModelItem *parentItem = childItem->parent;
    if(parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int HptdcStatusModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
                static_cast<HptdcStatusModelItem*>(parent.internalPointer()) : rootItem;
    return parentItem->childCount();
}

int HptdcStatusModel::columnCount(const QModelIndex &) const
{
    return maxChipNum + 1;
}

QModelIndex HptdcStatusModel::indexOf(const DeviceIndex &deviceIndex) const
{
    for (int i=0; i<rowCount(); i++) {
        auto idx = index(i, 1);
        if (!idx.isValid()) continue;
        const auto item = getItemByIndex(idx);
        if (!item)
            continue;
        if (item->dev == deviceIndex)
            return idx;
    }
//    const auto children = rootItem->children;
//    for (auto it = children.begin(); it != children.end(); ++it) {
//        HptdcStatusModelItem *item = *it;
//        if (item->dev == deviceIndex)
//            return {};
//    }
    return {};
}

void HptdcStatusModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    if(enabled_cache == map.getEnabled())
        return;

    enabled_cache = map.getEnabled();
    for(DeviceIndex index : hptdcStatusDeviceMap.keys()){
        if(!enabled_cache.contains(index))
           hptdcStatusDeviceMap.remove(index);
    }

    reset();
}

void HptdcStatusModel::setHptdcStatusMap(const DeviceIndex &index, const TdcCoreStatus &status)
{
    bool need_reset = false;
    bool need_display = false;
    if(hptdcStatusDeviceMap.keys().contains(index)){
        if(hptdcStatusDeviceMap[index] == status)
            return;
        need_display = true;
    } else
        need_reset = true;

    hptdcStatusDeviceMap[index] = status;

    int maxChipNumNew = 0;
    for(DeviceIndex index : hptdcStatusDeviceMap.keys())
        maxChipNumNew = std::max(maxChipNumNew, hptdcStatusDeviceMap[index].hptdcStatus.size());

    if(maxChipNum != maxChipNumNew) {
        maxChipNum = maxChipNumNew;
        need_reset = true;
    }

    if(need_reset)
        reset();
    if(need_display){
        if(!rootItem->children.isEmpty()) {
            for(int r=0; r<rootItem->children.size(); ++r) {
                if(rootItem->children[r]->dev == index){
                    if(!rootItem->children[r]->children.isEmpty()){
                        auto top = rootItem->children[r]->children.first();
                        auto top_index = createIndex(top->childCount(), 0, top);
                        auto bottom = rootItem->children[r]->children.last();
                        auto bottom_index = createIndex(bottom->childCount(), 0, bottom);
                        emit dataChanged(top_index, bottom_index);
                    }
                }
            }
        }
    }
}

QVariant HptdcStatusModel::displayData(const QModelIndex &index) const
{
    const HptdcStatusModelItem  *item = getItemByIndex(index);

    if(item->metric.getMetric() == HptdcStatusMetrics::Metric::Default){
        if(index.column() ==0)
            return item->dev.getSerialStr();
    } else {
        if(index.column() ==0)
            return item->metric.toString();
        else{
            auto devStat = hptdcStatusDeviceMap[item->dev];
            int id = index.column()-1;
            if(id >= devStat.hptdcStatus.size())
                return {};
            const auto &reg = devStat.hptdcStatus[id].regValues;
            switch (item->metric.getMetric()) {
            case HptdcStatusMetrics::Metric::Default:
                return -1;
            case HptdcStatusMetrics::Metric::VernierError:
                return (bool(reg[0] & (1<<0)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::CoarseError:
                return (bool(reg[0] & (1<<1)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::ChannelSelectError:
                return (bool(reg[0] & (1<<2)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::L1BufParityError:
                return (bool(reg[0] & (1<<3)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::TriggerFIFOParityError:
                return (bool(reg[0] & (1<<4)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::TriggerMatchingStateError:
                return (bool(reg[0] & (1<<5)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::ReadoutFIFOParityError:
                return (bool(reg[0] & (1<<6)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::ReadoutStateError:
                return (bool(reg[0] & (1<<7)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::SetupParityError:
                return (bool(reg[0] & (1<<8)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::ControlParityError:
                return (bool(reg[0] & (1<<9)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::JtagInstructionError:
                return (bool(reg[0] & (1<<10)) ? "error" : "no");
            case HptdcStatusMetrics::Metric::TdcHaveReadoutToken:
                return bool(reg[0] & (1<<11));
            case HptdcStatusMetrics::Metric::ReadoutFIFOFull:
                return bool(reg[0] & (1<<12));
            case HptdcStatusMetrics::Metric::ReadoutFIFOEmpty:
                return bool(reg[0] & (1<<13));
            case HptdcStatusMetrics::Metric::TriggerFIFOFull:
                return bool(reg[0] & (1<<14));
            case HptdcStatusMetrics::Metric::TriggerFIFOEmpty:
                return bool(reg[0] & (1<<15));
            case HptdcStatusMetrics::Metric::ReadoutFIFOOccupancy:
                return int(reg[1] & (127<<0));
            case HptdcStatusMetrics::Metric::TriggerFIFOOccupancy:
                return int((reg[1] >> 8) & 0xF);
            case HptdcStatusMetrics::Metric::DllLock:
                return bool(reg[1] & (1<<12));
            case HptdcStatusMetrics::Metric::InvertedSetupBit601:
                return bool(reg[1] & (1<<13));
            case HptdcStatusMetrics::Metric::OccupancyOfL1BufferOfChannelGroup0:
                return int(reg[2] & (127<<0));
            case HptdcStatusMetrics::Metric::OccupancyOfL1BufferOfChannelGroup1:
                return int(reg[2]>>8 & 0xFF);
            case HptdcStatusMetrics::Metric::OccupancyOfL1BufferOfChannelGroup2:
                return int(reg[3] & 0xFF);
            case HptdcStatusMetrics::Metric::OccupancyOfL1BufferOfChannelGroup3:
                return int(reg[3] >> 8);
            }
        }
    }
    return QVariant();
}

QVariant HptdcStatusModel::getForegroundData(const QModelIndex &index) const
{
    static const QBrush b_black(Qt::black);
    static const QBrush b_red(Qt::red);

    const HptdcStatusModelItem  *item = getItemByIndex(index);

    if(item->metric.getMetric() == HptdcStatusMetrics::Metric::Default
            || index.column() == 0){
        return b_black;
    }
    auto devStat = hptdcStatusDeviceMap[item->dev];
    int id = index.column()-1;
    if(id >= devStat.hptdcStatus.size())
        return {};
    const auto &reg = devStat.hptdcStatus[id].regValues;
    switch(item->metric.getMetric()) {
    case HptdcStatusMetrics::Metric::VernierError:
        return (bool(reg[0] & (1<<0)) ? b_red : b_black);
    case HptdcStatusMetrics::Metric::CoarseError:
        return (bool(reg[0] & (1<<1)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::ChannelSelectError:
        return (bool(reg[0] & (1<<2)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::L1BufParityError:
        return (bool(reg[0] & (1<<3)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::TriggerFIFOParityError:
        return (bool(reg[0] & (1<<4)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::TriggerMatchingStateError:
        return (bool(reg[0] & (1<<5)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::ReadoutFIFOParityError:
        return (bool(reg[0] & (1<<6)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::ReadoutStateError:
        return (bool(reg[0] & (1<<7)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::SetupParityError:
        return (bool(reg[0] & (1<<8)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::ControlParityError:
        return (bool(reg[0] & (1<<9)) ? b_red :  b_black);
    case HptdcStatusMetrics::Metric::JtagInstructionError:
        return (bool(reg[0] & (1<<10)) ? b_red :  b_black);
    default:
        return b_black;
    }
}

QVariant HptdcStatusModel::fontData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    QFont font;
    if(item->metric.getMetric() == HptdcStatusMetrics::Metric::Default)
        font.setBold(true);
    return font;
}

HptdcStatusModelItem *HptdcStatusModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<HptdcStatusModelItem*>(index.internalPointer());
}

void HptdcStatusModel::reset()
{
    beginResetModel();
    rootItem->children.clear();
    for(DeviceIndex index : hptdcStatusDeviceMap.keys()) {
        HptdcStatusModelItem *devItem = new HptdcStatusModelItem(rootItem);
        devItem->dev = index;
        for(HptdcStatusMetrics::Metric metric : HPTDC_STATUS_METRICS.keys()) {
            if(metric != HptdcStatusMetrics::Metric::Default){
                auto row_item = new HptdcStatusModelItem(devItem);
                row_item->metric.setMetric(metric);
                devItem->addChild(row_item);
            }
        }
        rootItem->addChild(devItem);
    }
    endResetModel();
}
