#include "WrStatusTableHelper.h"

#include <QTableWidget>

#include "device-modules/WrStatus/WrStatus.h"
#include "app/AppStyle.h"

QVariant data_DisplayRole(const WrStatus &wrStatus, int col) {
    switch (col) {
    case WrStatusTableHelper::COL_UPTIME:
        return wrStatus.uptime;
    case WrStatusTableHelper::COL_TIME:
    {
        if (!wrStatus.statusCode.bit.sfpOk) {
            return "No SFP";
        } else if (wrStatus.timeNull() || !wrStatus.statusCode.bit.linkUp) {
            return wrStatus.statusCode.bit.linkUp ? QString() : "No link";
        } else {
            const auto &wrTime = QDateTime::fromMSecsSinceEpoch(wrStatus.mSecsSinceEpoch());
            return wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss"); // add ".zzz" for ms
        }
    }
    case WrStatusTableHelper::COL_TIME_OFFSET:
    {
        // logic from adc64-system with acquired time
        double wrTimeOffsetMs = wrStatus.mSecsSinceEpoch() - wrStatus.acquiredAtMs;
        QString strOffset;
        if (wrTimeOffsetMs/1000. > -60 && wrTimeOffsetMs/1000. < 60)
            return QString(" %1").arg(wrTimeOffsetMs * 1e-3, 6, 'f', 3);
        return {};
    }
    case WrStatusTableHelper::COL_LINK_ERROR:
        return QString::number(wrStatus.linkError);
    case WrStatusTableHelper::COL_SYNC_LOST:
        return QString::number(wrStatus.syncFail);
    case WrStatusTableHelper::COL_RX_CODE_ERROR:
        return QString::number(wrStatus.codeError);
    default: return {};
    }
}

QVariant data_BackgroundRole(const WrStatus &wrStatus, int col) {
    static const QVariant bgColorDefault;
    static const QVariant bgColorWarning = AppStyle::tableBg(Severity::Warning);
    static const QVariant bgColorError = AppStyle::tableBg(Severity::Error);

    switch (col) {
    case WrStatusTableHelper::COL_TIME:
    {
        if (!wrStatus.statusCode.bit.sfpOk) {
            return bgColorDefault;
        } else if (wrStatus.timeNull() || !wrStatus.statusCode.bit.linkUp) {
            return bgColorWarning;
        } else {
            return wrStatus.timeValid() ? bgColorDefault : bgColorError;
        }
    }
    case WrStatusTableHelper::COL_TIME_OFFSET:
    {
        const bool wrTimeNull = wrStatus.timeNull();

        if (!wrStatus.statusCode.bit.sfpOk) {
            return bgColorDefault;
        } else if (wrTimeNull || !wrStatus.statusCode.bit.linkUp) {
            return bgColorWarning;
        } else {
            // logic from adc64-system with acquired time
            int wrTimeCheck = wrStatus.wrTimeCheck();
            if (wrTimeCheck == 1)
                return bgColorWarning;
            if (wrTimeCheck > 1)
                return bgColorError;
            return bgColorDefault;
        }
    }
    default: return {};
    }
}

QVector<int> WrStatusTableHelper::optionalColumns(int colOffset)
{
    return {colOffset + COL_LINK_ERROR,
            colOffset + COL_SYNC_LOST,
            colOffset + COL_RX_CODE_ERROR};
}

QVector<WrStatusTableHelper::Columns> WrStatusTableHelper::hiddenColumns()
{
    return {
        COL_UPTIME,
        COL_LINK_ERROR,
        COL_SYNC_LOST,
        COL_RX_CODE_ERROR
    };
}

void WrStatusTableHelper::setHeaderWrStatus(QTableWidget *table, int colOffset)
{
    for(auto c=0; c<COL_TOTAL; ++c) {
        const auto &str = headerData(c, Qt::DisplayRole).toString();
        table->setHorizontalHeaderItem(colOffset+c, new QTableWidgetItem(str));
    }
    table->hideColumn(colOffset+COL_TIME_OFFSET);
}

void WrStatusTableHelper::updateDevWrStatus(QTableWidget *table, int colOffset, int row, const std::optional<WrStatus> &wrStatus)
{
    for(auto c=0; c<COL_TOTAL; ++c) {
        auto i = table->item(row, colOffset+c);
        i->setText(data(wrStatus, c, Qt::DisplayRole).toString());
        auto bg = data(wrStatus, c, Qt::BackgroundRole);
        QBrush br;
        if(bg.isValid())
            br = bg.value<QColor>();
        i->setBackground(br);
    }
}

QVariant WrStatusTableHelper::data(const std::optional<WrStatus> &status, int col, int role)
{
    if(!status.has_value()) return QVariant();
    switch (role) {
    case Qt::BackgroundRole: return data_BackgroundRole(status.value(), col);
    case Qt::DisplayRole: return data_DisplayRole(status.value(), col);
    default: return QVariant();
    }
}

QVariant WrStatusTableHelper::headerData(int col, int role)
{
    if (role != Qt::DisplayRole)
        return QVariant();

    switch (col) {
    case COL_UPTIME: return "WR UpTime";
    case COL_TIME: return "WR Time";
    case COL_TIME_OFFSET: return "WR Offset";
    case COL_LINK_ERROR: return "WR Link";
    case COL_SYNC_LOST: return "WR Unlock";
    case COL_RX_CODE_ERROR: return "WR RX Error";
    default: return {};
    }
}
