#include "HistMngrWidget.h"
#include "ui_HistMngrWidget.h"

#include <cmath>
#include <cstdlib>

#include <QDateTime>
#include <QDebug>
#include <QFileDialog>
#include <QHash>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>
#include <QMetaType>
#include <QTimer>
#include <QSemaphore>

#include "qwt_plot.h"
#include "qwt_interval_data.h"
#include "qwt_plot_curve.h"
#include "qwt_plot_marker.h"
#include "qwt_scale_engine.h"

#include "DeviceData.h"
#include "DeviceHelper.h"

namespace  {
const int TDC_MAX_LAT_NS = 100;
const int SHRINK_MAX_SIZE = 1000;

void shrinkIntervals(QwtArray<QwtDoubleInterval> &intervals, QwtArray<double> &y_values){
    const int oldSize = intervals.size();
    Q_ASSERT(y_values.size()==oldSize);

    int group_size = oldSize/SHRINK_MAX_SIZE;
    if(group_size<=1)
        return;
    if(oldSize%SHRINK_MAX_SIZE)
        ++group_size;
    int newSize = oldSize/group_size;
    if(oldSize%group_size)
        ++newSize;
    QwtArray<QwtDoubleInterval> xVals(newSize);
    QwtArray<double> yVals(newSize);
    QwtDoubleInterval curInterval;
    double curVal=0;
    int step=0;
    int newBin=0;
    for(int oldBin=0; oldBin<oldSize; ++oldBin) {
        if(step==0) {
            curInterval = intervals[oldBin];
            curVal = y_values[oldBin];
        } else {
            curVal += y_values[oldBin];
        }
        if(++step == group_size){
            curInterval.setMaxValue(intervals[oldBin].maxValue());
            xVals[newBin] = curInterval;
            yVals[newBin] = curVal;
            ++newBin;
            curVal=0;
            step=0;
        }
    }
    if(oldSize%group_size) {
        curInterval.setMaxValue(intervals.last().maxValue());
        xVals[newBin] = curInterval;
        yVals[newBin] = curVal;
    }

    intervals = xVals;
    y_values = yVals;
}

void saveHist(const QString &path, const QHash<HistKey, Histogramm>::const_iterator &histIt)
{
    const auto &histKey = histIt.key();
    const auto &hist = histIt.value();
    bool isBase = HistGroupCfg::isBaseHist(histKey.hist_id);
    QString filePath = QString("%1/hist_%2").arg(path)
            .arg(getStrKeyForHistId(histKey.hist_id));
    if(!isBase)
        filePath += QString("_%1_%2_%3")
                .arg(histKey.index.getDeviceIdName())
                .arg(histKey.index.getSerialStr())
                .arg(histKey.ch);
    filePath += ".json";
    QFile file(filePath);
    if(!file.open(QIODevice::WriteOnly)) {
        qCritical() << "Failed to open file for saving histogram:"
                    << filePath << file.errorString();
        return;
    }
    QJsonObject histObj;
    QJsonArray bins;

    histObj["hist_id"] = getStrKeyForHistId(histKey.hist_id);
    if(!isBase) {
        histObj["index"] = histKey.index.getSerialStr();
        histObj["channel"] = histKey.ch;
    }

    histObj["n_bins"] = hist.n_bins;
    histObj["min"] = hist.min;
    histObj["max"] = hist.max;
    histObj["under_flow"] = int(hist.under_flow);
    histObj["over_flow"] = int(hist.over_flow);
    for(auto v : hist.bins) {
        bins << v;
    }
    histObj["bins"] = bins;

    QJsonDocument doc(histObj);
    file.write(doc.toJson(QJsonDocument::Compact));
    file.close();
}

inline void setComboBoxEdgeIndex(QComboBox *cb, bool lead, bool trail) {
    int i = lead && trail ? 2 : (lead ? 0 : 1);
    cb->setCurrentIndex(i);
}
}

HistMngrWidget::HistMngrWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::HistMngrWidget),
    refreshTimer(new QTimer(this))
{
    qRegisterMetaType<HistMngrCfg>();
    qRegisterMetaType<Histogramm>();
    qRegisterMetaType<InlResults>();
    qRegisterMetaType<QMap<DeviceIndex, int> >();
    qRegisterMetaType<QHash<HistKey, Histogramm> >();

    connect(refreshTimer, &QTimer::timeout, [this]() {refreshCurrentPlot();} );
    refreshTimer->setInterval(500);
    ui->setupUi(this);
    ui->spinBoxMwMax->setMaximum(HIST_MATCHWIN_MAX);
    ui->spinBoxMwMax->setMinimum(HIST_MATCHWIN_MIN);
    ui->spinBoxMwMin->setMaximum(HIST_MATCHWIN_MAX);
    ui->spinBoxMwMin->setMinimum(HIST_MATCHWIN_MIN);

    static const QStringList EDGE_ITEMS{"Lead", "Trail", "Lead+Trail"};
    for(auto cb : {ui->comboBoxCurEdge, ui->comboBoxStopEdge, ui->comboBoxHitTimeEdge})
        cb->addItems(EDGE_ITEMS);
}

HistMngrWidget::~HistMngrWidget()
{
    delete ui;
}

void HistMngrWidget::hist_updated(QHash<HistKey, Histogramm> histogramms_,
                                         bool forceRefresh)
{
    histogramms = histogramms_;
    refreshCurrentPlot(forceRefresh);
}

void HistMngrWidget::refreshCurrentPlot(bool force)
{
    if(!force && lastRefreshTime.isValid() && lastRefreshTime.elapsed()<500) {
        return;
    }
    const auto &cur_hist_id = config.cur_hist_id;
    auto hictCfgIt = config.hist_groups.constFind(cur_hist_id);
    if(hictCfgIt == config.hist_groups.cend())
        return;

    HistKey hist_key = config.get_cur_hist_key();
    if(hist_key.hist_id == HistId::mscPlot) {
        for(auto i : histogramms.keys()) {
            if(i.hist_id == HistId::mscPlot){
                hist_key = i;
                break;
            }
        }
    }
    if(!histogramms.contains(hist_key))
        return;

    const auto &hist = histogramms[hist_key];
    const int nBins = hist.n_bins;

    QwtArray<QwtDoubleInterval> intervals(nBins);
    QwtArray<double> y_values(hist.bins);
    QwtArray<double> x_values(nBins);
    int leftEdge=-1,rightEdge=-1;
    qint64 sum = 0;
    int binMin=0, binMax=0;
    bool isInlCurve = cur_hist_id == HistId::inl && !config.inl_hist_code_dens;

    if(config.cur_hist_id == HistId::tdcHitEv && config.hide_zero_bin
            && !y_values.isEmpty())
        y_values[0]=0;

//    int min_bin = isInlCurve ? 0 : hist.findBin(hist.min_filled_val);
//    int max_bin = isInlCurve ? nBins : hist.findBin(hist.max_filled_val)+1;
//    for(int bin = min_bin; bin < max_bin; ++bin)
    for(int bin = 0; bin < nBins; ++bin)
    {
        if(isInlCurve) {
            x_values[bin] = hist.getBinCenter(bin);
        } else {
            QwtDoubleInterval &i = intervals[bin];
            i.setMinValue(hist.getBinLower(bin));
            i.setMaxValue(hist.getBinUpper(bin));
        }
        int gbin = static_cast<int>(y_values[bin]);

        if(sum) {
            binMin = qMin(binMin, gbin);
            binMax = qMax(binMax, gbin);
        } else {
            binMin = gbin;
            binMax = gbin;
        }
        sum += gbin;
        if(gbin) {
            if(leftEdge==-1)
                leftEdge=bin;
            rightEdge=bin;
        }
    }

    if(leftEdge != -1 &&
            config.cur_hist_id != HistId::inl &&
            config.cur_hist_id != HistId::mscPlot &&
            config.cur_hist_id != HistId::tdcCh){
        intervals = intervals.mid(leftEdge, rightEdge-leftEdge+1);
        y_values = y_values.mid(leftEdge, rightEdge-leftEdge+1);
    }

    auto &plotItemHash = isInlCurve ? histQwtCurve : histQwtItems;
    auto plotItem = plotItemHash.value(hist_key, nullptr);
    if(!plotItem){
        addDeviceChQwtItems(hist_key.hist_id, hist_key.index, hist_key.ch);
        plotItem = plotItemHash.value(hist_key, nullptr);
        if(!plotItem)
            return;
    }

    if(isInlCurve) {
        QwtPlotCurve *inlCurv = static_cast<QwtPlotCurve*>(plotItem);
        if(y_values.size()==1024 && !config.inl_hist_code_dens){
            y_values = makeDNLCorrectionTable(y_values);
        }

        inlCurv->setData(x_values, y_values);
        inlCurv->itemChanged();
    } else {
        HistogramItem *qwtHist = static_cast<HistogramItem*>(plotItem);
        shrinkIntervals(intervals, y_values);

        qwtHist->setData(QwtIntervalData(intervals, y_values));
        updateFit(hictCfgIt.value(), hist);
        qwtHist->itemChanged();
    }
    ui->qwtPlotHists->setAxisAutoScale(QwtPlot::xBottom);

    QwtScaleEngine *scale_engine = ui->qwtPlotHists->axisScaleEngine(QwtPlot::xBottom);
    scale_engine->setAttribute(QwtScaleEngine::Floating, true);

    ui->qwtPlotHists->replot();

    QStringList histInfo;
    if(!isInlCurve) {
        histInfo << QString("under:%1; over:%2")
                        .arg(hist.getUnderFlow())
                        .arg(hist.getOverFlow());
        if(hist.getOverFlow() || hist.getUnderFlow()) {
            histInfo << QString("full_range: [%1; %2]")
                            .arg(hist.min)
                            .arg(hist.max);
        }
        if(!intervals.isEmpty()){
            histInfo << QString("fill_range: [%1; %2]")
                            .arg(intervals.first().minValue())
                            .arg(intervals.last().maxValue());
        }
        else{
            histInfo << QString("range: [0; 0]");
        }

    } else {
        histInfo << QString("min/max: %1-%2").arg(binMin).arg(binMax);
    }
    histInfo << QString("count: %1").arg(hist.getEntries());
    lastRefreshTime.start();

    ui->labelGroupInfo->setText(histInfo.join('\n'));
}

void HistMngrWidget::on_comboBoxHistGroup_currentIndexChanged(int i)
{
    config.cur_hist_id = getHistId(i);

    onCurGroupChanged();
    emit histConfigChanged(config);
}

void HistMngrWidget::on_checkBoxHistGroupEn_toggled(bool checked)
{

    auto hCfgIt = config.hist_groups.find(config.cur_hist_id);
    if(hCfgIt == config.hist_groups.end())
        return;
    QString text = hCfgIt->getHistGroupInfo().title;
    if(!checked)
        text.append(" (off)");
    hCfgIt->hist_group_en = checked;
    ui->comboBoxHistGroup->setItemText(ui->comboBoxHistGroup->currentIndex(), text);
    enable_hist_group(checked);

    emit histConfigChanged(config);
}

void HistMngrWidget::on_comboBoxCurDev_currentIndexChanged(int i)
{
    config.cur_dev_index = getDevIndex(false, i);
    updateSpinBoxMaxCh(false);
    onCurGroupChanged();
    emit histConfigChanged(config);
}

void HistMngrWidget::on_spinBoxCurCh_valueChanged(int i)
{
    config.cur_ch = i;
    onCurGroupChanged();
    emit histConfigChanged(config);
}

void HistMngrWidget::on_comboBoxCurEdge_currentIndexChanged(int index)
{
    ++index;
    config.curEdgeLead = index&1;
    config.curEdgeTrail = index&2;
    emit histConfigChanged(config);
}

void HistMngrWidget::on_groupBoxStopCh_toggled(bool val)
{
    config.use_stop_ch = val;
    emit histConfigChanged(config);
}

void HistMngrWidget::on_comboBoxStopDev_currentIndexChanged(int i)
{
    config.stop_dev_index = getDevIndex(true, i);
    updateSpinBoxMaxCh(true);
    emit histConfigChanged(config);
}

void HistMngrWidget::on_spinBoxStopCh_valueChanged(int ch)
{
    config.stop_ch = static_cast<quint16>(ch);
    emit histConfigChanged(config);
}

void HistMngrWidget::on_comboBoxStopEdge_currentIndexChanged(int index)
{
    ++index;
    config.stopEdgeLead = index&1;
    config.stopEdgeTrail = index&2;
    emit histConfigChanged(config);
}

void HistMngrWidget::on_spinBoxMwMin_valueChanged(int min)
{
    config.mw_min = min;
    emit histConfigChanged(config);
    ui->spinBoxMwMax->setMinimum(min);
}

void HistMngrWidget::on_spinBoxMwMax_valueChanged(int max)
{
    config.mw_max = max;
    emit histConfigChanged(config);
    ui->spinBoxMwMin->setMaximum(max);
}

void HistMngrWidget::on_comboBoxHitTimeEdge_currentIndexChanged(int index)
{
    ++index;
    config.hitTimeEdgeLead = index&1;
    config.hitTimeEdgeTrail = index&2;
    emit histConfigChanged(config);
}

void HistMngrWidget::on_radioButtonInl_toggled(bool checked)
{
    config.inl_hist_code_dens = !checked;
    emit histConfigChanged(config);
    onCurGroupChanged();
    if(config.cur_hist_id == HistId::inl) {
        refreshCurrentPlot(true);
    }
}

void HistMngrWidget::on_radioButtonAdcAmplMin_toggled(bool checked)
{
    config.ampl_hist_with_max = !checked;
    emit histConfigChanged(config);
}

void HistMngrWidget::on_pushButtonSave_clicked()
{
    const auto &key = config.get_cur_hist_key();
    if(key.isEmpty())
        return;
    auto it = histogramms.constFind(key);
    if(it == histogramms.cend())
        return;
    // Refresh iterator due to QDialog::exec in getExportPath
    it = histogramms.constFind(key);
    if(it == histogramms.cend())
        return;

    auto dirPath = getExportPath();
    if(dirPath.isEmpty())
        return;

    saveHist(dirPath, it);
}

void HistMngrWidget::on_pushButtonSaveAll_clicked()
{
    auto dirPath = getExportPath();
    if(dirPath.isEmpty())
        return;

    for(auto it=histogramms.cbegin(); it!=histogramms.cend(); ++it)
        saveHist(dirPath, it);
}

void HistMngrWidget::on_pushButtonClear_clicked()
{
    const auto &key = config.get_cur_hist_key();
    if(!key.isEmpty())
        emit clearHist(key);
}

void HistMngrWidget::on_pushButtonClearAll_clicked()
{
    emit clearHist(HistKey());
}

void HistMngrWidget::on_groupBoxGlobalEnable_toggled(bool checked)
{
    config.global_enabled = checked;
    emit histConfigChanged(config);
    if(config.global_enabled)
        refreshTimer->start();
    else {
        refreshTimer->stop();
        refreshCurrentPlot(true);
    }
}

void HistMngrWidget::on_groupBoxFit_toggled(bool checked)
{
    auto hCfgIt = config.hist_groups.find(config.cur_hist_id);
    if(hCfgIt == config.hist_groups.end())
        return;
    hCfgIt->show_fit = checked;

    if(checked) {
        const auto &histIt = histogramms.constFind(config.get_cur_hist_key());
        if(histIt != histogramms.cend())
            updateFit(hCfgIt.value(), histIt.value());
    }

    emit histConfigChanged(config);
    refreshCurrentPlot(true);
}

void HistMngrWidget::on_checkBoxLogY_toggled(bool checked)
{
    config.log_y_axis = checked;
    emit histConfigChanged(config);
    refreshCurrentPlot(true);
}

void HistMngrWidget::inlCheckRequest()
{
    if(!config.is_inl_enabled) return;

    InlCheckResults res;
    res.stopDev = config.stop_dev_index;
    res.stopCh = config.stop_ch;

    for(auto hIt=histogramms.cbegin(); hIt!=histogramms.cend(); ++hIt) {
        if(hIt.key().hist_id != HistId::inlCheck)
            continue;
        DeviceCalibration devCal(hIt.key().index);
        InlCheckCh &chRes = res.devices[hIt.key().index].channels[hIt.key().ch];
        chRes.ch = hIt.key().ch;
        chRes.dist = hIt->getDistribution();
    }
    emit inlCheckUpdated(res);
}

QVector<double> HistMngrWidget::makeDNLCorrectionTable(const QVector<double> code_density)
{
    if (code_density.size() != 1024) {
        qWarning() << "makeDNLCorrectionTable: invalid input vector size";
        return QVector<double>();
    }
    double sumv = 0;
    for(int i=0; i<1024; i++) {
        double v = code_density[i];
        sumv += v;
    }
    QVector<double> inl(1024, 0.0);
    double avgv = sumv / 1024.;
    double sumrelv = 0;
    double inlsum = 0;
    for(int i=0; i<1024; i++) {
        double relv = (code_density[i] - avgv) / avgv;
        sumrelv += relv;
        inl[i] = sumrelv;
        inlsum += sumrelv;
    }

    QVector<double> corr(1024, 0.0);
    for(int i=0; i<1024; i++) {
        corr[i] = inl[i] - inlsum/1024.;
        corr[i] = round(corr[i]*100.0) / 100.0; // quantize
    }
    return corr;
}

void HistMngrWidget::updateInlCorr(QMap<DeviceIndex, DeviceCalibration> allCal)
{
    inlFromConfig = allCal;
}

void HistMngrWidget::updateInlCorr(const DeviceIndex &index,
                                   const QMap<int, QVector<double> > &corrMap)
{
    for(auto ch: corrMap.keys())
        inlFromConfig[index].inl[ch] = corrMap[ch];
}

void HistMngrWidget::updateFit(const HistGroupCfg &histCfg, const Histogramm &hist) const
{
    QString text;
    if(histCfg.show_fit) {
        if(hist.getEntries() && hist.n_bins>3) {
            const auto &hist_dist = hist.getDistribution();
            text = QString("Mean:  %1\nSigma: %2")
                    .arg(hist_dist.mean, 0, 'f', 3)
                    .arg(hist_dist.rms, 0, 'f', 3);
        } else {
            text = "No fit data\n";
        }
    } else {
        text = "fit disabled";
    }
    ui->labelFitInfo->setText(text);
}

HistId HistMngrWidget::getCurHistId() const
{
    return getHistId(ui->comboBoxHistGroup->currentIndex());
}

HistId HistMngrWidget::getHistId(int index) const
{
    const auto &histId = ui->comboBoxHistGroup->itemData(index).value<HistId>();
    return config.hist_groups.contains(histId) ? histId : HistId::unknown;
}

DeviceIndex HistMngrWidget::getDevIndex(bool getStopDev, int idInComboBox) const
{
    const auto cb = getStopDev ? ui->comboBoxStopDev : ui->comboBoxCurDev;
    DeviceIndex dev;
    if(idInComboBox == -1)
        dev = cb->currentData().value<DeviceIndex>();
    else
        dev = cb->itemData(idInComboBox).value<DeviceIndex>();
    return  dev;
}

DeviceIndex HistMngrWidget::getCurDev(bool getStopDev) const
{
    return getDevIndex(getStopDev);
}

int HistMngrWidget::getCurCh() const
{
    return ui->spinBoxCurCh->value();
}

void HistMngrWidget::onCurGroupChanged()
{
    auto cfgIt = config.hist_groups.constFind(config.cur_hist_id);

    if (cfgIt == config.hist_groups.end())
        return;
    const auto &cfgHistId = cfgIt->hist_id;
    enable_hist_group(cfgIt->hist_group_en);

    ui->qwtPlotHists->detachItems(QwtPlotItem::Rtti_PlotItem, false);

    QwtPlotItem *newPlotItem = nullptr;
    if(cfgHistId == HistId::inl && !config.inl_hist_code_dens)
        newPlotItem = histQwtCurve.value(config.get_cur_hist_key(), nullptr);
    else
        newPlotItem = histQwtItems.value(config.get_cur_hist_key(), nullptr);
    if(newPlotItem)
        newPlotItem->attach(ui->qwtPlotHists);
    const auto &histMarkersIt = histMarkers.constFind(cfgHistId);
    if(histMarkersIt != histMarkers.cend())
        for(auto m : histMarkersIt.value())
            m->attach(ui->qwtPlotHists);

    ui->qwtPlotHists->setTitle(cfgIt->getHistGroupInfo().title);
    //bring fit curv to foreground, curve will be drawn on next updateFit(...)
    refreshCurrentPlot(true);

    ui->qwtPlotHists->axisScaleEngine(QwtPlot::xBottom)->setMargins(1, 1);

    // Update GUI for this group
    updateHistToolTip();

    block_gui_signals(true);
    ui->groupBoxMatchWin->setVisible(cfgHistId == HistId::tdcTime);
    ui->checkBoxZeroBin->setVisible(cfgHistId == HistId::tdcHitEv);
    ui->groupBoxInl->setVisible(cfgHistId == HistId::inl);
    ui->groupBoxFit->setVisible(cfgHistId != HistId::inl);
    ui->groupBoxFit->setChecked(cfgIt->show_fit);
    ui->groupBoxStopCh->setVisible(cfgHistId == HistId::tdcDoublePulseTime ||
                                   cfgHistId == HistId::inlCheck);
    ui->groupBoxCurDevCh->setVisible(!cfgIt->isBaseHist());
    ui->groupBoxFillWith->setVisible(cfgHistId == HistId::chAmpl);
    ui->checkBoxHistGroupEn->setChecked(cfgIt->hist_group_en);

    block_gui_signals(false);
    emit histConfigChanged(config);
}

void HistMngrWidget::updateDeviceComboBoxItems(bool stopDev)
{
    QComboBox *box = stopDev ? ui->comboBoxStopDev : ui->comboBoxCurDev;
    DeviceIndexSet devSet;
    if(stopDev ) {
        const auto &list = tdcChMap.keys();
        devSet = {list.begin(), list.end()};
    } else {
        auto histId = getCurHistId();
        for(auto i=histogramms.cbegin(); i!=histogramms.cend(); ++i) {
            if(i.key().hist_id == histId)
                devSet << i.key().index;
        }
    }

    auto devFromCfg = stopDev ? config.stop_dev_index : config.cur_dev_index;
    if(!devFromCfg.isNull())
        devSet << devFromCfg;

    // remove extra
    for(int i=0; i<box->count(); ) {
        const auto &devIndex = box->itemData(i).value<DeviceIndex>();
        if(devSet.contains(devIndex)) {
            ++i;
            continue;
        }
        box->removeItem(i);
    }

    // add missing
    for(auto devIndex : devSet) {
        int itemIndex = box->findData(QVariant::fromValue(devIndex));
        if(itemIndex != -1)
            continue;
        // todo insert in correct position
        box->insertItem(0, devIndex.getIdent(),
                        QVariant::fromValue(devIndex));
    }
    updateSpinBoxMaxCh(stopDev);
}

void HistMngrWidget::updateHistGroupComboBoxItems()
{
    Q_ASSERT(!config.hist_groups.contains(HistId::unknown));
    block_gui_signals(true);
    int cb_pos = 0;
    auto comboBox = ui->comboBoxHistGroup;
    for(auto it=config.hist_groups.cbegin(); it!=config.hist_groups.cend(); ++it) {
        const HistGroupCfg &group_config = *it;
        const HistGroupInfo &group_info = it->getHistGroupInfo();
        if(!config.is_inl_enabled
                && (group_config.hist_id == HistId::inl
                    || group_config.hist_id == HistId::inlCheck))
            continue;
        const auto &cbText = group_info.title + (group_config.hist_group_en?"" : " (off)");
        const auto &cbData = QVariant::fromValue(group_config.hist_id);
        int index = comboBox->findData(cbData);
        if(index == -1) {
            comboBox->insertItem(cb_pos, cbText, cbData);
        } else if(comboBox->itemText(index) != cbText) {
            comboBox->setItemText(index, cbText);
        }

        if(config.cur_hist_id == group_config.hist_id) {
            index = comboBox->findData(cbData);
            Q_ASSERT(index!=-1);
            comboBox->setCurrentIndex(index);
        }
        ++cb_pos;
//        ui->checkBoxHistGroupEn->setChecked(group_config.hist_group_en); // this should be in onCurGroupChanged();
    }

    for(int i=0; i<comboBox->count();) {
        const auto &groupId = getHistId(i);
        if(!config.hist_groups.contains(groupId))
            comboBox->removeItem(i);
        else
            ++i;
    }
    block_gui_signals(false);
    Q_ASSERT(!config.hist_groups.contains(HistId::unknown));
}

void HistMngrWidget::devListUpdated(const DevChMap &tdcChMap_,
                                    const QHash<HistKey, Histogramm> &histogramms_)
{
    histogramms = histogramms_;

    // Update ComboBox StopCh
    updateDeviceComboBoxItems(true);
    updateDeviceComboBoxItems(false);

    if(tdcChMap != tdcChMap_) {
        tdcChMap = tdcChMap_;
        QMap<int, QPair<DeviceIndex, int>> sortedMap;
        for(auto it=tdcChMap.cbegin(); it!=tdcChMap.cend(); ++it) {
            Q_ASSERT(!sortedMap.contains(it->first));
            Q_ASSERT(it->second);
            sortedMap[it->first] = {it.key(), it->second};
        }

        // Update hist 'Hits by dev/ch'
        QStringList tdcChTtStrList;
        auto &tdcChMarkers = histMarkers[HistId::tdcCh];
        while(!tdcChMarkers.isEmpty())
            delete tdcChMarkers.takeFirst();
        for(auto it=sortedMap.cbegin(); it!=sortedMap.cend(); ++it) {
            const auto lab = QString("%1-%2: %3").arg(it.key())
                    .arg(it.key()+it->second)
                    .arg(it->first.getIdent());

            tdcChTtStrList << lab;

            auto * marker = new QwtPlotMarker();
            marker->setLineStyle(QwtPlotMarker::VLine);
            marker->setXValue(it.key());
            marker->label().font().italic();
            QwtText label(lab);
            QFont font = label.font();
            font.setItalic(true);
            label.setFont(font);
            marker->setLabel(label);
            marker->setLabelOrientation(Qt::Vertical);
            marker->setLabelAlignment(Qt::AlignRight | Qt::AlignTop);
            if(getCurHistId() == HistId::tdcCh)
                marker->attach(ui->qwtPlotHists);
            tdcChMarkers.push_back(marker);
        }
        histToolTips[HistId::tdcCh] = '\n'+tdcChTtStrList.join('\n');

        updateHistToolTip();
    }
}

void HistMngrWidget::setHistConfig(HistMngrCfg config_)
{
    block_gui_signals(true);

    const bool histGroupChanged = config.hist_groups != config_.hist_groups;
    config = config_;

    // update Qwt items for new Config
    if(histGroupChanged) {
        removeExtraQwtItems();

        for(auto hCfg : config.hist_groups) {
            if(hCfg.isBaseHist()) {
                addBaseQwtItems(hCfg);
                continue;
            }
//            for(auto index : knownDevices) {
//                const uint nch = QtMregDevice::getDeviceChNum(index);
//                for(uint ch=0; ch<nch; ++ch)
//                    addDeviceChQwtItems(hCfg.hist_id, index, ch);
//            }
        }
    }

    if(config.cur_hist_id == HistId::unknown ||
            !config.hist_groups.contains(config.cur_hist_id)){
        if(config.hist_groups.isEmpty()) {
            config.cur_hist_id = HistId::unknown;
        } else {
            const auto &cfg = config.hist_groups.first();
            config.cur_hist_id = cfg.hist_id;
        }
    }

    // Update GUI from top to bottom
    updateHistGroupComboBoxItems();

    ui->groupBoxGlobalEnable->setChecked(config.global_enabled);
    if(config.global_enabled)
        refreshTimer->start();

    updateDeviceComboBoxItems(false);
    // checkBoxHistGroupEn in onCurGroupChanged();

    updateSpinBoxMaxCh(false);
    ui->spinBoxCurCh->setValue(config.cur_ch);
    setComboBoxEdgeIndex(ui->comboBoxCurEdge,
                         config.curEdgeLead, config.curEdgeTrail);

    ui->groupBoxStopCh->setChecked(config.use_stop_ch);
    updateDeviceComboBoxItems(true);
    updateSpinBoxMaxCh(true);
    ui->spinBoxStopCh->setValue(config.stop_ch);
    setComboBoxEdgeIndex(ui->comboBoxStopEdge,
                         config.stopEdgeLead, config.stopEdgeTrail);

    ui->radioButtonInl->setChecked(!config.inl_hist_code_dens);
    ui->radioButtonAdcAmplMin->setChecked(!config.ampl_hist_with_max);

    ui->spinBoxMwMin->setMaximum(config.mw_max);
    ui->spinBoxMwMax->setMinimum(config.mw_min);
    ui->spinBoxMwMin->setValue(config.mw_min);
    ui->spinBoxMwMax->setValue(config.mw_max);
    setComboBoxEdgeIndex(ui->comboBoxHitTimeEdge,
                         config.hitTimeEdgeLead, config.hitTimeEdgeTrail);

    ui->checkBoxLogY->setChecked(config.log_y_axis);
    ui->checkBoxLogY->hide();

    ui->qwtPlotHists->replot();

    onCurGroupChanged();

    block_gui_signals(false);
}

void HistMngrWidget::taiReperUpdated(HistKey hKey, const TaiTime &taiReperSec)
{
    auto toolTip = histToolTips[hKey];
    if(!taiReperSec.isNull())
        toolTip = QString(" (reper: %2)")
                .arg(QDateTime::fromTime_t(taiReperSec.sec).toString("dd/MM/yy hh:mm:ss"));
    else
        toolTip.clear();
    updateHistToolTip();
}

QVector<double> HistMngrWidget::getInl(const TqdcEvent &ev, const TdcChHit &tdcEv) const
{
    return getInl(ev.index, tdcEv.ch);
}

QVector<double> HistMngrWidget::getInl(const DeviceIndex &index, int ch) const
{
    return inlFromConfig.value(index).inl.value(ch, QVector<double>());
}

void HistMngrWidget::updateHistToolTip()
{
    QString toolTip = config.getHistGroupInfo(config.cur_hist_id).tool_tip;
    toolTip += histToolTips.value(config.get_cur_hist_key());

    ui->comboBoxHistGroup->setToolTip(toolTip);
    ui->labelGroupDescr->setText(toolTip);
    ui->labelGroupDescr->setVisible(!toolTip.isEmpty());
}

void HistMngrWidget::updateSpinBoxMaxCh(bool stopDev)
{
    const auto &devIndex = getDevIndex(stopDev);
    int devChNum = 0;
    if(stopDev) {
        devChNum = tdcChMap.value(devIndex, {0,0}).second;
        if(devIndex == config.stop_dev_index)
            devChNum = std::max(devChNum, int(config.stop_ch));
    } else {
        const auto &histId = getCurHistId();
        for(auto hi : histogramms.keys())
            if(hi.hist_id == histId && hi.index == devIndex)
                devChNum = std::max(devChNum, hi.ch);
    }
    devChNum = std::max(devChNum, 0);
    auto sb = stopDev ? ui->spinBoxStopCh : ui->spinBoxCurCh;
    sb->setMaximum(devChNum);
}

void HistMngrWidget::addBaseQwtItems(const HistGroupCfg &group_config)
{
    if(!group_config.isBaseHist())
        return;

    HistKey hist_key(group_config.hist_id);
    if(histQwtItems.contains(hist_key))
        return;

    HistogramItem *histItem =  new HistogramItem(group_config.getHistGroupInfo().title);
    histItem->setColor(Qt::darkCyan);
    histQwtItems[hist_key] = histItem;
}

void HistMngrWidget::addDeviceChQwtItems(HistId id, const DeviceIndex &index, int ch)
{
    auto cfgIt = config.hist_groups.constFind(id);
    if(cfgIt == config.hist_groups.cend())
        return;

    HistKey hist_key(id, index, ch);

    const auto &title = cfgIt->getHistGroupInfo().title;
    if(id == HistId::inl && !histQwtCurve.contains(hist_key)) {
        QwtPlotCurve *hist_curve = new QwtPlotCurve(title);
        histQwtCurve[hist_key] = hist_curve;
    }

    if(histQwtItems.contains(hist_key))
        return;

    auto *hist_item = new HistogramItem(title);
    hist_item->setColor(Qt::darkCyan);
    histQwtItems[hist_key] = hist_item;
}

void HistMngrWidget::removeExtraQwtItems()
{
    for(bool isCurve : {true, false}) {
        auto &map = isCurve ? histQwtCurve : histQwtItems;

        for(auto it=map.begin();it!=map.end(); ++it) {
            auto hist_key = it.key();
            auto item = it.value();

            const auto &cfgHistIt = config.hist_groups.constFind(hist_key.hist_id);
            bool toDel = cfgHistIt == config.hist_groups.cend();
            if(!toDel && !cfgHistIt->isBaseHist()
                    && !histogramms.contains(hist_key))
                toDel = true;
            if(isCurve)
                toDel &= hist_key.hist_id != HistId::inl;

            if(toDel) {
                map.remove(hist_key);
                item->detach();
                delete item;
            } else {
                const auto &title = cfgHistIt->getHistGroupInfo().title;
                if(item->title() != title) {
                    item->setTitle(title);
                }
            }
        }
    }
}

void HistMngrWidget::on_checkBoxZeroBin_clicked(bool checked)
{
    config.hide_zero_bin = checked;
}

void HistMngrWidget::block_gui_signals(bool block)
{
    if(block)
        ++block_signal_counter;
    else
        --block_signal_counter;
    Q_ASSERT(block_signal_counter>=0);

    if(block_signal_widgets_list.isEmpty()) {
        block_signal_widgets_list.append(ui->groupBoxGlobalEnable);

        block_signal_widgets_list.append(ui->comboBoxHistGroup);
        block_signal_widgets_list.append(ui->checkBoxHistGroupEn);

        block_signal_widgets_list.append(ui->comboBoxCurDev);
        block_signal_widgets_list.append(ui->spinBoxCurCh);

        block_signal_widgets_list.append(ui->groupBoxStopCh);
        block_signal_widgets_list.append(ui->comboBoxStopDev);
        block_signal_widgets_list.append(ui->spinBoxStopCh);

        block_signal_widgets_list.append(ui->spinBoxMwMin);
        block_signal_widgets_list.append(ui->spinBoxMwMax);

        block_signal_widgets_list.append(ui->radioButtonInl);
        block_signal_widgets_list.append(ui->radioButtonAdcAmplMin);

        block_signal_widgets_list.append(ui->groupBoxFit);
        block_signal_widgets_list.append(ui->checkBoxLogY);
        block_signal_widgets_list.append(ui->checkBoxZeroBin);
    }

    if(block_signal_counter > 1)
        return;

    for(int i=0; i<block_signal_widgets_list.size(); ++i){
        block_signal_widgets_list[i]->blockSignals(block_signal_counter);
    }
}

void HistMngrWidget::enable_hist_group(bool enable)
{
    ui->labelGroupDescr->setEnabled(enable);
    ui->labelGroupInfo->setEnabled(enable);
    ui->groupBoxFit->setEnabled(enable);
    ui->groupBoxCurDevCh->setEnabled(enable);
    ui->groupBoxStopCh->setEnabled(enable);
    ui->groupBoxMatchWin->setEnabled(enable);
    ui->groupBoxInl->setEnabled(enable);
    ui->groupBoxFillWith->setEnabled(enable);
    ui->checkBoxZeroBin->setEnabled(enable);
    ui->checkBoxLogY->setEnabled(enable);
}

QString HistMngrWidget::getExportPath()
{
    auto dirPath =
        QFileDialog::getExistingDirectory(this, "Export hist to directory...",
                                          config.histExportPath);
    if(!dirPath.isEmpty()) {
        config.histExportPath = dirPath;
        emit histConfigChanged(config);
    }
    return dirPath;
}

DeviceIndexSet HistMngrWidget::getDeviceIndexSet(const HistId &id) const
{
    DeviceIndexSet ret;
    for(auto it=histogramms.cbegin(); it!=histogramms.cend(); ++it) {
        if(it.key().hist_id == id)
            ret << it.key().index;
    }
    return ret;
}

QSet<int> HistMngrWidget::getDeviceChSet(const HistId &id, const DeviceIndex &index) const
{
    QSet<int> ret;
    for(auto it=histogramms.cbegin(); it!=histogramms.cend(); ++it) {
        if(it.key().hist_id == id && it.key().index == index)
            ret << it.key().ch;
    }
    return ret;
}
