#ifndef BASEDEVICEAPPCONFIG_H
#define BASEDEVICEAPPCONFIG_H

#include <QDebug>
#include <QMap>
#include <QMetaType>

#include "ModularDeviceConfig.h"
#include "base/DeviceIndex.h"
#include "base/DeviceIndexEnabled.h"
#include "zmq-lib/ZmqTypes.h"

struct BaseDeviceAppConfig
{
    friend bool operator==(const BaseDeviceAppConfig &, const BaseDeviceAppConfig &) = default;

    DeviceIndexEnabledMap getDeviceIndexEnabledMap() const
    {
        DeviceIndexEnabledMap map;
        for (auto it = knownConfigs.cbegin(); it != knownConfigs.cend(); ++it) {
            const auto &index = it.key();
            bool enabled = it.value().enabled;
            map.insert(index, {index, enabled});
        }
        return map;
    }

    bool hasEnabledDevices() const
    {
        for (const auto &s : knownConfigs)
            if (s.enabled)
                return true;
        return false;
    }

    void fixAfterDbRead()
    {
        defaultSetup.trcCore.fix();
        setRunState(false);

        for(FeLinkChConfig &chCfg : defaultSetup.feLink.chCfg) {
            chCfg.reset = false;
            chCfg.resetSoftPll = false;
            chCfg.remoteReset = false;
        }

        for(auto &devCfg : knownConfigs) {
            for(FeLinkChConfig &chCfg : devCfg.feLink.chCfg) {
                chCfg.reset = false;
                chCfg.resetSoftPll = false;
                chCfg.remoteReset = false;
            }
        }
    }

    void setRunState(bool run)
    {
        defaultSetup.run = run;
        defaultSetup.trcCore.phyTrig.trig.trigOpen = run;
        for (auto &ds : knownConfigs) {
            ds.run = run && ds.enabled;
            ds.trcCore.phyTrig.trig.trigOpen = run && ds.enabled;
        }
    }

    bool isRun() const
    {
        int on = 0, off = 0;
        for (auto &deviceConfig : knownConfigs) {
            if (deviceConfig.enabled)
                ++(deviceConfig.isRun() ? on : off);
        }
        if (on && off)
            qWarning() << "Config has mix (running/stopped) devices";
        return on;
    }

    void unify_config()
    {
        for (auto &i : knownConfigs) {
            // if (i.enabled)
            i.unify_config(defaultSetup);
        }
    }

    ModularDeviceConfig &findDevice(const DeviceIndex &index)
    {
        if (!knownConfigs.contains(index))
            knownConfigs.insert(index, defaultSetup);
        return knownConfigs[index];
    }
    const ModularDeviceConfig &findDevice(const DeviceIndex &index) const
    {
        auto found = knownConfigs.find(index);
        if (found == knownConfigs.end())
            return defaultSetup;
        else
            return *found;
    }
    DeviceIndex firstDeviceKey() const
    {
        if (knownConfigs.empty())
            return {};
        else
            return knownConfigs.firstKey();
    }
    bool contains(const DeviceIndex &index) const
    {
        return knownConfigs.contains(index);
    }
    QList<DeviceIndex> deviceKeys() const
    {
        return knownConfigs.keys();
    }
    virtual void removeInvalidEntries() {
        knownConfigs.remove(DeviceIndex{});
        auto it = knownConfigs.begin();
        while (it != knownConfigs.end()) {
            if (it.value().index != it.key())
                knownConfigs.erase(it++);
            else
                ++it;
        }
    }
    bool isEnabled(const DeviceIndex &deviceIndex) const {
        auto found = knownConfigs.find(deviceIndex);
        if (found == knownConfigs.end())
            return false;
        return found->enabled;
    }

    bool allowRemoteControl = true;
    bool errOnTrigOnXOff = true; // emits run err on TrigOnXOff
    bool errOnSerDesUnlock = true; // emits run err on SerDes unlock
    ZmqConfig devStatusZmqPublisher;

//protected:
    QMap<DeviceIndex, ModularDeviceConfig> knownConfigs;
    ModularDeviceConfig defaultSetup;
};

Q_DECLARE_METATYPE(BaseDeviceAppConfig)

#endif // BASEDEVICEAPPCONFIG_H
