/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceStatusTableView.h"

#include <QTimer>
#include <QHeaderView>
#include <QRegularExpression>
#include <QScrollBar>

#include "mvc/DeviceIndexTableModel.h"
#include "mvc/DeviceEnableTableModel.h"
#include "mvc/MergeColumnsTableModel.h"
#include "base/DaqModuleTraits.h"
#include "base/DeviceIndex.h"
#include "base/Globals.h"
#include "ModularDeviceStatus.h"
#include "device-discover/LldpDiscoveryTableModel.h"
#include "ModularDeviceStatus/ModularDeviceStatusModel.h"
#include "mldiscover/LldpDiscoveryDeviceListFilter.h"
#include "m-stream-output/MStreamStatisticTableModel.h"

namespace {
const auto k_DeviceStatusTableView = "DeviceStatusTableView";
const auto k_visible = "visible";
} // namespace

DeviceStatusTableView::DeviceStatusTableView(QWidget *parent)
    : QTableView(parent)
    , mergeModel(new MergeColumnsTableModel(this))
    , proxyModel(new QSortFilterProxyModel(this))
{
    proxyModel->setSourceModel(mergeModel);
    setModel(proxyModel);
    setSortingEnabled(true);

    addSourceModel(createEnableModel());
    addSourceModel(createDiscoverModel());
    addSourceModel(createStatusModel());
    addSourceModel(createMStreamModel());

    verticalHeader()->hide();
    // show vertical header of this model
    // mergeModel->setPrimarySourceModel(statusModel);

    setAlternatingRowColors(true);

    QTimer::singleShot(10, this, &DeviceStatusTableView::deferredInit);

    connect(proxyModel, &QAbstractItemModel::modelReset,
            this, &DeviceStatusTableView::updateSizeHint);

    auto sp = sizePolicy();
    sp.setVerticalPolicy(QSizePolicy::Maximum);
    setSizePolicy(sp);
}

QSize DeviceStatusTableView::sizeHint() const
{
    auto size = QTableView::sizeHint();
    if(!proxyModel)
        return size;

    int h = 2;
    if(auto hh = horizontalHeader())
        if(hh->isVisible())
            h += hh->sizeHint().height(); // 26

    for(auto row=0; row<proxyModel->rowCount(); ++row) {
        h += rowHeight(row); // 30 per row
    }

    if(auto sb = horizontalScrollBar())
        h += sb->sizeHint().height(); // 14

    size.setHeight(h);
    return size;
}

void DeviceStatusTableView::deferredInit()
{
    setDefaultColumnState();

    for (int i=0; i<mergeModel->columnCount(); i++) {
        auto var = loadColumnState(i);
        if (!var.canConvert<bool>())
            continue;
        QTableView::setColumnHidden(i, !var.toBool());
    }
}

void DeviceStatusTableView::updateSizeHint()
{
    updateGeometry();
}

void DeviceStatusTableView::resetColumnState()
{
    // clear settings
    setDefaultColumnState();
}

DeviceIndexTableModel *DeviceStatusTableView::createEnableModel()
{
    auto model = new DeviceEnableTableModel(this);
    model->setObjectName("Config");
    connect(this, &DeviceStatusTableView::setDeviceList,
            model, &DeviceEnableTableModel::setDeviceList);
    connect(model, &DeviceEnableTableModel::deviceListChanged,
            this, &DeviceStatusTableView::deviceListChanged);
    return model;
}

QMap<int, bool> DeviceStatusTableView::defaultColumnState() const
{
    QMap<int, bool> v;
    for (auto subModel : subModels) {
        auto defaultStates = subModel->defaultColumnState();
        for (int i=0; i<subModel->columnCount(); i++) {
            int submodelColumn = i;
            int column = mergeModel->getColumnFromSource(subModel, submodelColumn);
            if (column >= 0 && defaultStates.contains(submodelColumn))
                v[column] = defaultStates[submodelColumn];
        }
    }
    return v;
}

void DeviceStatusTableView::setDefaultColumnState()
{
    const auto &states = defaultColumnState();
    for (auto it = states.begin(); it != states.end(); ++it) {
        int column = it.key();
        bool shown = it.value();
        QTableView::setColumnHidden(column, !shown);
    }
}

DeviceIndexTableModel *DeviceStatusTableView::createDiscoverModel()
{
    auto model = new LldpDiscoveryTableModel(this);
    model->setObjectName("LLDP Info");
    auto &lldpDiscovery = LldpDiscovery::instance();
    auto filter = new LldpDiscoveryDeviceListFilter(this);
    connect(this, &DeviceStatusTableView::setDeviceList,
            filter, &LldpDiscoveryDeviceListFilter::setDeviceList);
    connect(&lldpDiscovery, &LldpDiscovery::cacheUpdated,
            filter, &LldpDiscoveryDeviceListFilter::lldpCacheUpdated);
    connect(filter, &LldpDiscoveryDeviceListFilter::lldpCacheUpdatedFiltered,
            model, &LldpDiscoveryTableModel::lldpCacheUpdated);
    return model;
}

DeviceIndexTableModel *DeviceStatusTableView::createStatusModel()
{
    auto model = new ModularDeviceStatusModel(this);
    model->setObjectName("Basic Status");
    connect(this, &DeviceStatusTableView::setStatus,
            model, &ModularDeviceStatusModel::setStatus);
    connect(this, &DeviceStatusTableView::setMultiStatus,
            model, &ModularDeviceStatusModel::setMultiStatus);
    connect(this, &DeviceStatusTableView::setDeviceList,
            model, &ModularDeviceStatusModel::setDeviceList);
    return model;
}

DeviceIndexTableModel *DeviceStatusTableView::createMStreamModel()
{
    if (!DaqModuleTraits::hasMstream(Globals::getDaqModuleType()))
        return {};
    auto model = new MStreamStatisticTableModel(this);
    model->setObjectName("MStream Status");
    connect(this, &DeviceStatusTableView::setDeviceList,
            model, &MStreamStatisticTableModel::setDeviceList);
    connect(this, &DeviceStatusTableView::stat_updated,
            model, &MStreamStatisticTableModel::stat_updated);
    connect(this, &DeviceStatusTableView::stat_from_receiver_updated,
            model, &MStreamStatisticTableModel::stat_from_receiver_updated);
    connect(this, &DeviceStatusTableView::mstreamConnectionStateChanged,
            model, &MStreamStatisticTableModel::mstreamConnectionStateChanged);
    return model;
}

void DeviceStatusTableView::addSourceModel(DeviceIndexTableModel *model)
{
    if (!model)
        return;
    if (!subModels.contains(model))
        subModels.push_back(model);
    mergeModel->addSourceModel(model,  Qt::UserRole);
}

DeviceIndex DeviceStatusTableView::deviceIndexOf(const QModelIndex &index) const
{
    auto sourceIndex = index;
    if (sourceIndex.isValid() && sourceIndex.model() == proxyModel)
        sourceIndex = proxyModel->mapToSource(sourceIndex);
    if (sourceIndex.isValid() && sourceIndex.model() == mergeModel)
        sourceIndex = mergeModel->mapToSource(sourceIndex);
    if (!sourceIndex.isValid())
        return {};

    for (auto m : subModels) {
        if (m == sourceIndex.model())
            return m->deviceIndexOf(sourceIndex.row());
    }
    return {};
}

void DeviceStatusTableView::currentChanged(const QModelIndex &current, const QModelIndex &previous)
{
    QTableView::currentChanged(current, previous);
    const auto &sourceIndex = proxyModel->mapToSource(current);

    const auto &deviceIndex = deviceIndexOf(sourceIndex);
    emit currentDeviceIndexChanged(deviceIndex);
}

void DeviceStatusTableView::hideModelColumn(QAbstractItemModel *model, int column)
{
    QTableView::hideColumn(mergeModel->getColumnFromSource(model, column));
}

void DeviceStatusTableView::showModelColumn(QAbstractItemModel *model, int column)
{
    QTableView::showColumn(mergeModel->getColumnFromSource(model, column));
}

QString DeviceStatusTableView::columnSettingsKey(int column)
{
    const auto &source = mergeModel->mapColumnToSource(column);
    if (!source.isValid())
        return {};
    QRegularExpression rx("[^A-za-z0-9]");
    QString modelName = source.model->objectName();
    QString columnName = source.model->headerData(source.col, Qt::Horizontal).toString();
    modelName.replace(rx, "_");
    columnName.replace(rx, "_");
    QString key = QString(k_DeviceStatusTableView) + "/" + modelName + "/" + columnName + "/" + k_visible;
    return key;
}

void DeviceStatusTableView::saveColumnState(int column, bool shown)
{
    auto key = columnSettingsKey(column);
    if (key.isEmpty())
        return;
    QSettings settings;
    settings.setValue(key, shown);
}

QVariant DeviceStatusTableView::loadColumnState(int column)
{
    const QSettings settings;
    return settings.value(columnSettingsKey(column));
}

void DeviceStatusTableView::hideColumn(int column)
{
    QTableView::hideColumn(column);
    saveColumnState(column, false);
}

void DeviceStatusTableView::showColumn(int column)
{
    QTableView::showColumn(column);
    saveColumnState(column, true);
}
