/*
**    Copyright 2010-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef QTMREGDEVICE_H
#define QTMREGDEVICE_H

#include <vector>

#include <QMap>
#include <QObject>
#include <QSharedPointer>
#include <QVector>

#include "base/DeviceIndex.h"
#include "MlinkStateFsm.h"

enum AFI_SDB_ID : uint32_t;
enum DeviceId : uint8_t;

namespace mlink {
class MregDevice;
class RegIOBase;
class RegOpVector;
class RegOp;
}

struct fw_version_t;
class MlinkCsrModule;
class QHostAddress;
class QTimer;
class sdb_t;

class QtMregDevice : public QObject
{
    Q_OBJECT
    Q_DISABLE_COPY(QtMregDevice)

public:

    explicit QtMregDevice(const DeviceIndex &index, QObject *parent = nullptr);
    ~QtMregDevice() override;

    bool isConnected() const;
    bool isFailed() const;
    DeviceId getDeviceId() const;
    fw_version_t getDeviceFwVersion() const;
    fw_version_t getDeviceHwVersion() const;
    DeviceIndex getDeviceIndex() const { return deviceIndex; }
    QString getIdent() const;
    QString getFirmwareVersionStr() const;
    QString getAddressStr() const;
    QMap<QString, double> getFpgaTemps() const;
    sdb_t getSdb() const;

    void setDeleteFlag();
    bool hasDeleteFlag() const;
    void setLiveMagicCheck(bool state) { enableLiveMagicCheck = state; }
    bool regOpExec(mlink::RegOpVector &data);
    bool memReadBlk(int startWord, size_t wordCount, std::vector<quint32> *result);
    bool memWriteBlk(int wordNumber, const std::vector<quint32> &data);

Q_SIGNALS:
    void deviceStateChanged(const DeviceIndex &, MlinkStateFsm::State);
    void deviceConnected(const DeviceIndex &);
    void deviceDisconnected(const DeviceIndex &);
    void liveMagicFailed(const DeviceIndex &);

public Q_SLOTS:
    void setHostAddress(const DeviceIndex &index, const QHostAddress &address, int ttl);

protected:
    quint64 getDeviceSerialNumber() const;
    virtual bool readStatus() { return true; }
    virtual bool writeConfig() { return true; }

protected Q_SLOTS:
    virtual void on_deviceConnected() {}
    virtual void on_deviceDisconnected() {}

private Q_SLOTS:
    void statusPoll();
    void fsmPoll();
    void on_linkStateChanged(MlinkStateFsm::State oldState, MlinkStateFsm::State newState);
    void onProtocolMessage(const QString &errorStr);
    void onProtocolTimeout();

private:
    bool writeLiveMagic();
    bool checkLiveMagic();
    QHostAddress getAddress() const;
    void forceDisconnect();
    bool detect();
    quint16 nextLiveMagic() const;
    bool detect_mlink_csr();
    QSharedPointer<MlinkCsrModule> createMlinkCsrModule();
    qint64 failureRecoveryTimeoutMs() const;

    mlink::RegIOBase * const link;
    std::atomic<bool> deleteFlag{false};
    const DeviceIndex deviceIndex;
    mlink::MregDevice * const mregDevice; // dev must be after link
    QSharedPointer<MlinkCsrModule> mlinkCsrModule; // Note: ModularMregDevice contains its own MlinkCsrModule
    QTimer * const timerFsmPoll;
    QTimer * const timerStatusPoll;
    MlinkStateFsm linkFSM;
    int linkFailureCount = 0;
    quint16 liveMagic = 0;
    bool enableLiveMagicCheck = true;
    int detectRetryCount = 0;
};

#endif // QTMREGDEVICE_H
