/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RegIo16Codec.h"

#include <QByteArray>

#include "regiopacket.h"
#include "regop.h"

namespace mlink {

static const quint32 READ_BIT = 0x80000000;
// static const quint32 WRITE_BIT = 0x00000000;
static const quint32 ADDR_MASK = 0x7FFF;
static const quint32 DATA_MASK = 0xFFFF;

int RegIo16Codec::encode_regio16_op(const RegOp &op, PacketData *packet)
{
    int count = 0;
    int opSize = RegIo16Codec::d16Count(op);
    // auto e = op.endianness();
    bool read = op.isRead();
    bool write = op.isWrite();
    if (read || write) {
        quint64 wrdata = op.wdata;
        for (int i = 0; i < opSize; i++) {
            quint32 data = RegIo16Codec::encode_regio16_op(read, op.addr+i, wrdata & 0xFFFF);
            wrdata >>= 16;
            packet->push_back(data);
            count++;
        }
    }
    return count;
}

quint32 RegIo16Codec::encode_regio16_op(bool read, quint16 addr, quint16 data)
{
    quint32 d = 0;
    if (read) d |= READ_BIT;
    d |= (addr & ADDR_MASK) << 16;
    d |= (data & DATA_MASK);
    return d;
}

int RegIo16Codec::d16Count(const RegOp &op)
{
    switch (op.dataWidth()) {
    case ByteCount::NONE: return 0;
    case ByteCount::D16: return 1;
    case ByteCount::D32: return 2;
    case ByteCount::D64: return 4;
    }
    return 0;
}

bool RegIo16Codec::encode(const RegOpVector &v, QByteArray &buf)
{
    for (auto &op : v) {
        int opSize = d16Count(op);
        // auto e = op.endianness();
        bool read = op.isRead();
        bool write = op.isWrite();
        if (read || write) {
            quint64 wrdata = op.wdata;
            for (int i = 0; i < opSize; i++) {
                quint32 data = encode_regio16_op(read, op.addr+i, wrdata & 0xFFFF);
                wrdata >>= 16;
                buf.append(data, sizeof(data));
            }
        }
    }
    return true;
}

bool RegIo16Codec::decode(const PacketData &rx, RegOpVector::iterator first, RegOpVector::iterator last)
{
    int pos = 0;
    for (auto it = first; it != last; ++it) {
        auto &op = *it;
        int opSize = d16Count(op);
        auto e = op.endianness();
        bool read = op.isRead();
        bool write = op.isWrite();
        if (read || write) {
            quint64 rddata = 0;
            for (int i = 0; i < opSize; i++) {
                if (pos > rx.size())
                    return false;
                const quint32 &data = rx[pos++];
                quint16 rd_addr = (data >> 16) & ADDR_MASK;
                if (rd_addr != op.addr + i)
                    return false;
                if (e == Endianness::BE)
                    rddata = (rddata << 16) | (data & 0xFFFF);
                else
                    rddata |= (quint64)(data & 0xFFFF) << (i*16);
            }
            op.rdata = rddata;
        }
        op.result.ok = true;
        op.decode();

    }
    return true;
}

bool RegIo16Codec::decode(const PacketData &buf, RegOpVector &v)
{
    return decode(buf, v.begin(), v.end());
}

} // namespace mlink
