
#include "regiobase.h"

#include <algorithm>
#include <cmath>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <sys/time.h>

#include <QCoreApplication>
#include <QDebug>

#include "mlink/frame_types.h"

namespace mlink {

double getCurrentTimestamp()
{
    struct timeval tv{};
    gettimeofday(&tv, nullptr);
    return tv.tv_sec + 1e-6*tv.tv_usec;
}

void RegIOBase::setDeleteFlag()
{
    deleteFlag.store(true, std::memory_order_relaxed);
}

std::string RegIOBase::getDevName() const
{
    return "mlink-netdev";
}

unsigned long RegIOBase::getDriverVersion() const
{
    return 0;
}

unsigned long RegIOBase::getDriverRevision() const
{
    return 0;
}

void RegIOBase::setHostAddress(const QHostAddress &address, int ttl)
{
    deviceAddress = address;
    deviceAddressTimer.setRemainingTime(ttl);
}

//const double protocolTimeout = 1.0;

bool RegIOBase::packetTypeIsData(int type)
{
    if (type == FRAME_TYPE_ADC_DATA) return true;
    if (type == FRAME_TYPE_STAT) return true;
    return false;
}

bool RegIOBase::hasDeleteFlag() const
{
    return deleteFlag.load(std::memory_order_relaxed);
}

void RegIOBase::sendPackets(const PacketList &pl)
{
    for(PacketList::const_iterator it = pl.begin(); it != pl.end(); ++it) {
        rawSend(*it);
    }
}

void RegIOBase::reset()
{
    rxBufCtrl.clear();
    //    rxBufData.clear();
}

double RegIOBase::getDeviceThroughput() const
{
    return 1e6; // default value
}

size_t RegIOBase::getCtrlPacketCount() const
{
    return rxBufCtrl.size();
}

bool RegIOBase::getCtrlPacketBySeq(RegIoPacket *dest, int seq, double timeoutSec)
{
    double tv1 = getCurrentTimestamp();
    int attempt = 0;
    const int attemptLimit = 10;
    while(!hasDeleteFlag())
    {
        auto waitTimeMs = timeoutSec*1e3/pow(2,attemptLimit-(attempt++));
        updateRxBuf(waitTimeMs);
//        qDebug() << rxBufCtrl.size();
        for(PacketList::const_iterator it = rxBufCtrl.begin(); it != rxBufCtrl.end(); ++it) {
//            qDebug() << "seq = " << it->seq << ", expected " << seq;
            if (it->seq == seq) {
                if (dest)
                    *dest = *it;
                rxBufCtrl.clear();
//                qDebug() << "Ok: rxBufCtrl=" << rxBufCtrl;
                return true;
            }
        }
        double tv2 = getCurrentTimestamp();
        double dt = tv2 - tv1;
        if (dt > timeoutSec) {
//            qWarning() << "Protocol timeout, dt=" << dt << " s (getCtrlPacketBySeq(" << seq << ", " << timeout << "))";
//            qDebug() << "rxBufCtrl=" << rxBufCtrl;
            break;
        }
    }
//    const PacketList pl(rxBufCtrl);
//    rxBufCtrl.clear();
//    qDebug()<<attempt<< " "<<qPow(0.5,attemptLimit-attempt+1);
    return false;
}

PacketList RegIOBase::getCtrlPackets(size_t packetCount, double timeout)
{
    double tv1 = getCurrentTimestamp();
    while((getCtrlPacketCount() < packetCount) && !hasDeleteFlag())
    {
        //        receive();
        if (getCtrlPacketCount() >= packetCount) break;
        double tv2 = getCurrentTimestamp();
        double dt = tv2 - tv1;
        if (dt > timeout) {
            qWarning() << "Protocol timeout, dt=" << dt << " s (getControlPackets(" << packetCount << ", " << timeout << "))";
            qDebug() << "rxBufCtrl=" << rxBufCtrl;
            break;
        }
        qApp->processEvents();
    }
    const PacketList pl(rxBufCtrl);
    rxBufCtrl.clear();
    return pl;
}

} // namespace mlink
