/*
**    Copyright 2013-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef REGOP_H
#define REGOP_H

#include <optional>

#include <QDebug>
#include <QVector>
#include <QtGlobal>
#include <QIODevice>

namespace mlink {

enum class OpMode {
    NOP,
    R16LE, R16=R16LE, R16BE,
    W16LE, W16=W16LE, W16BE,
    R32LE, R32=R32LE, R32BE,
    W32LE, W32=W32LE, W32BE,
    R64LE, R64=R64LE, R64BE,
    W64LE, W64=W64LE, W64BE
};

enum class ByteCount {NONE, D16, D32, D64};

enum class Endianness {LE, BE};

class OpResult {
public:
    bool ok = false;
};

class RegOp
{
public:
    friend class RegIo16Codec;
    friend class RegOpStream;
    friend class RegOpVector;

    constexpr RegOp(const RegOp &r) noexcept = default;
    constexpr RegOp(RegOp &&r) noexcept = default;
    RegOp &operator=(const RegOp &) noexcept = default;
    RegOp &operator=(RegOp &&r) noexcept = default;

    [[nodiscard]]
    constexpr RegOp(OpMode mode, quint64 addr, quint64 wrdata, bool *ok) noexcept
        : mode(mode), addr(addr), wdata(wrdata), ok(ok) {}
    [[nodiscard]]
    constexpr RegOp(OpMode mode, quint64 addr, void *rdptr, bool *ok) noexcept
        : mode(mode), addr(addr), rdptr(rdptr), ok(ok) {}

    friend QDebug operator <<(QDebug debug, const mlink::RegOp &v);

private:
    RegOp() noexcept = delete;

private:
    ByteCount dataWidth() const;
    Endianness endianness() const;
    bool isWrite() const;
    bool isRead() const;
    void decode();

    OpMode mode;
    quint64 addr;
    quint64 wdata;
    quint64 rdata;
    OpResult result;

    void *rdptr = nullptr;
    bool *ok = nullptr;
};

class RegOpVector: public std::vector<RegOp>
{
public:
    RegOpVector() noexcept = default;
    RegOpVector(const RegOp &op) : RegOpVector({op}) {}
    RegOpVector(std::initializer_list<RegOp> init) :
        std::vector<RegOp>{init} {}
    RegOpVector(const RegOpVector::const_iterator &first, const RegOpVector::const_iterator &last)
        : std::vector<RegOp>(first, last) {}

    RegOpVector &operator<<(RegOp &&r) noexcept {
        emplace_back(std::move(r));
        return *this;
    }
    RegOpVector &operator<<(const RegOpVector &r) noexcept {
        for (auto it = std::begin(r); it != std::end(r); ++it)
            push_back(*it);
        return *this;
    }
    void RegRead16(quint16 addr, quint16 *data = nullptr, bool *ok = nullptr) {
        emplace_back(OpMode::R16, addr, data, ok);
    }
    void RegRead32(quint16 addr, quint32 *data = nullptr, bool *ok = nullptr) {
        emplace_back(OpMode::R32, addr, data, ok);
    }
    void RegRead64(quint16 addr, quint64 *data = nullptr, bool *ok = nullptr) {
        emplace_back(OpMode::R64, addr, data, ok);
    }
    void RegWrite16(quint16 addr, quint16 data, bool *ok = nullptr) {
        emplace_back(OpMode::W16, addr, data, ok);
    }
    void RegWrite32(quint16 addr, quint32 data, bool *ok = nullptr) {
        emplace_back(OpMode::W32, addr, data, ok);
    }
    void RegWrite64(quint16 addr, quint64 data, bool *ok = nullptr) {
        emplace_back(OpMode::W64, addr, data, ok);
    }
    void rebase(quint32 offset) {
        offset -= base;
        base = offset;
        for (auto it = begin(); it != end(); ++it)
            (*it).addr += base;
    }
protected:
    quint32 base = 0;
};

} // namespace mlink

using mlink::RegOpVector;

#endif // REGOP_H
