/*
**    Copyright 2019-2020 Ivan Filippov
**    Copyright 2019-2020 Timur Gandjelashvili
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "BsonQtConverter.h"

#include <QDataStream>
#include <QDateTime>
#include <QDebug>
#include <QJsonDocument>
#include <QJsonValue>
#include <QJsonArray>
#include <QJsonObject>

#include <bsoncxx/types/value.hpp>
#include <bsoncxx/builder/basic/array.hpp>
#include <bsoncxx/builder/basic/document.hpp>
#include <bsoncxx/types.hpp>
#include <bsoncxx/document/value.hpp>

#include <bsoncxx/builder/core.hpp>
#include <bsoncxx/types.hpp>
#include <bsoncxx/json.hpp>

using bsoncxx::builder::basic::kvp;
using namespace bsoncxx;
using namespace bsoncxx::builder;

QVariant BsonQtConverter::bsoncxx_to_qvariant(const bsoncxx::types::value &value)
{
    switch (value.type()) {
    case bsoncxx::type::k_utf8: {
        auto view = value.get_utf8().value;
        return QString::fromStdString(view.to_string());
    }
    case bsoncxx::type::k_int32:
        return static_cast<int>(value.get_int32());
    case bsoncxx::type::k_int64:
        return static_cast<qint64>(value.get_int64());
    case bsoncxx::type::k_double:
        return static_cast<double>(value.get_double());
    case bsoncxx::type::k_bool:
        return static_cast<bool>(value.get_bool());
    case bsoncxx::type::k_date: {
        auto value_date = value.get_date().to_int64();
        return QDateTime::fromMSecsSinceEpoch(value_date, Qt::OffsetFromUTC);
    }
    case bsoncxx::type::k_array: {
        qDebug() << "TODO: BSON type array not implemented";
        return QVariant();
    }
    case bsoncxx::type::k_binary: {
        auto binary_data = value.get_binary();
        if (binary_data.sub_type < bsoncxx::binary_sub_type::k_user) {
            qCritical() << "Unsuported BSON binary subtype"
                        << static_cast<int>(binary_data.sub_type);
            return QVariant();
        }
        QByteArray value_binary(reinterpret_cast<const char *>(binary_data.bytes),
                                static_cast<int>(binary_data.size));
        QDataStream in(value_binary);
        QVariant var = QVariant();
        in >> var;
        if (!var.isValid())
            qCritical() << "Invalid QVariant";
        // Q_ASSERT(var.isValid());
        if (in.status() == QDataStream::Ok)
            return var;
        qCritical() << "Can't convert BSON binary to QVariant";
        Q_ASSERT(false);
    }
    default:
        qCritical() << "Unsuported BSON type" << static_cast<int>(value.type());
        Q_ASSERT(false);
    }
    return QVariant();
}

void BsonQtConverter::append_qvar_to_document(bsoncxx::builder::basic::document &doc_to_return,
                                              const std::string &key,
                                              const QVariant &qvar)
{
    switch(qvar.type()){
    case QVariant::Bool:{
        bsoncxx::types::b_bool tmp {qvar.toBool()};
        doc_to_return.append(kvp(key, tmp));
        break;
    }
    case QVariant::Int:
    case QMetaType::UShort: // 36
    case QMetaType::UChar: // 37
    {
        bsoncxx::types::b_int32 tmp {qvar.toInt()};
        doc_to_return.append(kvp(key, tmp));
        break;
    }
    case QVariant::UInt:
    case QVariant::LongLong:
    // case QVariant::ULongLong: no BSON support for unsigned 64-bit
    case QMetaType::ULong: // 35
    {
        bsoncxx::types::b_int64 tmp {qvar.toLongLong()};
        doc_to_return.append(kvp(key, tmp));
        break;
    }
    case QVariant::Double:{
        bsoncxx::types::b_double tmp {qvar.toDouble()};
        doc_to_return.append(kvp(key, tmp));
        break;
    }
    case QVariant::Date:
    case QVariant::DateTime:
    {
        auto msec = qvar.toDateTime().toMSecsSinceEpoch();
        std::chrono::milliseconds duration{msec};
        std::chrono::system_clock::time_point tp{duration};
        doc_to_return.append(kvp(key, bsoncxx::types::b_date(tp)));
        break;
    }
    case QVariant::String:
    {
        const auto &utf_str = qvar.toString().toUtf8();
        bsoncxx::types::b_utf8 tmp{utf_str};
        doc_to_return.append(kvp(key, tmp));
        break;
    }
    default:{
        qDebug() << "Converting" << qvar.type() << "to BSON binary";
        QByteArray buf;
        QDataStream stream(&buf, QIODevice::WriteOnly);
        stream << qvar;
        if(stream.status() == QDataStream::Ok) {
            bsoncxx::types::b_binary tmp = {};
            tmp.sub_type = bsoncxx::binary_sub_type::k_user;
            tmp.bytes = reinterpret_cast<const uint8_t*>(buf.data());
            tmp.size = static_cast<uint32_t>(buf.size());
            doc_to_return.append(kvp(key, tmp));
        } else {
            qWarning() << "Can't stream QVariant " << qvar <<  ". Please, add type cast";
            Q_ASSERT(false);
        }
    }
    }
}


bsoncxx::builder::basic::document BsonQtConverter::toDocument(const QVariantMap &m)
{
    bsoncxx::builder::basic::document result;
    for (auto it = m.cbegin(); it != m.cend(); ++it) {
        // TODO: in MongoDB, key cannot start with '$'
        auto key = it.key();
        auto value = it.value();
        switch (value.type()) {
//        case QVariant::List: {
//            result.append(
//                        kvp(it.key().toStdString(), toDocument(value.toList()))
//                        );
//            break;
//        }
        case QVariant::Map: {
            result.append(kvp(it.key().toStdString(), toDocument(value.toMap())));
            break;
        }
        default:
            append_qvar_to_document(result, key.toStdString(), value);
        }
    }
    return result;
}

bsoncxx::builder::basic::array BsonQtConverter::toDocument(const QVariantList &m)
{
    bsoncxx::builder::basic::array result;
    for (const auto &value : m) {
        switch (value.type()) {
        case QVariant::List: {
            result.append(toDocument(value.toList()));
            break;
        }
        case QVariant::Map: {
            result.append(toDocument(value.toMap()));
            break;
        }
        default:;
            // TODO
            // append_qvar_to_document(result, key.toStdString(), value);
        }
    }
    return result;
}

QVariantMap BsonQtConverter::fromDocument(const bsoncxx::document::view &doc)
{
    QVariantMap result;
    for (const auto &it : doc) {
        auto key = it.key().to_string();
        auto type = it.type();
        auto qKey = QString::fromStdString(key);
        if (type == bsoncxx::type::k_oid)
            continue; // ignore oid
        if (type == bsoncxx::type::k_document) {
            result.insert(qKey, fromDocument(it.get_document()));
        } else {
            result[qKey] = bsoncxx_to_qvariant(it.get_value());
        }
    }
    return result;
}

bsoncxx::document::value BsonQtConverter::from_json(const QJsonObject &obj)
{
    return  bsoncxx::from_json(
        QJsonDocument(obj).toJson(QJsonDocument::Indented).toStdString()
    );
}

QJsonObject BsonQtConverter::to_json(const bsoncxx::document::view &obj)
{
    std::string str = bsoncxx::to_json(obj);
    QByteArray byteArray(str.c_str(), str.length());
    QJsonDocument doc = QJsonDocument::fromJson(byteArray);
    return doc.object();
}
