#include "AddPnpDialog.h"
#include "ui_AddPnpDialog.h"

#include <QStringList>
#include <QHeaderView>
#include <QTableWidget>
#include <QComboBox>
#include <QPushButton>
#include "pnp-server/PNPMessage.h"
#include "AddPnpModel.h"

namespace {
enum COLUMNS{COL_PROG_TYPE, COL_HOST_NAME, COL_PROG_INDEX, COL_TOTAL};
constexpr auto k_all_types = "All types";
constexpr auto k_all_hosts = "All hosts";
constexpr auto k_all_indexes = "All indexes";
bool hasSelection(const QTableView *t)
{
    return t->selectionModel()->hasSelection();
}
void fixSpacerSizePolicy(QSpacerItem *s) {
    const auto &sh = s->sizeHint();
    s->changeSize(sh.width(), sh.height(), QSizePolicy::Minimum, QSizePolicy::Expanding);
}
}

//---public

AddPnpDialog::AddPnpDialog(PnpInfoCache progDescrs, bool multi_select, QWidget *parent)
    : QDialog(parent)
    , ui(new Ui::AddClientDialog)
    , pnpCache(progDescrs)
    , multi_select(multi_select)
    , filter_func(nullptr)
    , leftModel(new AddPnpModel(this))
    , rightModel(new AddPnpModel(this))
{
    ui->setupUi(this);
    init_table(ui->leftTableView, leftModel);

    if(multi_select) {
        fixSpacerSizePolicy(ui->verticalSpacerUp);
        fixSpacerSizePolicy(ui->verticalSpacerDown);
        init_table(ui->rightTableView, rightModel);
    } else {
        ui->leftTableView->setSelectionMode(QAbstractItemView::SingleSelection);
        ui->rightTableView->hide();
        ui->widgetAddRemove->hide();
    }

    applyProgFilter();
    check_ok_button_enable();
}

AddPnpDialog::~AddPnpDialog()
{
    delete ui;
}

QVector<ProgramDescription> AddPnpDialog::getProgDescr() const
{
    QVector<ProgramDescription> ret;
    if(multi_select) {
        for(auto pd : selected_progs)
            ret.push_back(pd);
    } else {
        for(auto i : ui->leftTableView->selectionModel()->selection().indexes()) {
            auto ptr = modelIndex_to_progdescr(i);
            if(ptr)
                ret.push_back(*ptr);
        }
    }
    return ret;
}

void AddPnpDialog::set_function_filter(std::function<bool(const ProgramDescription&)> f)
{
    filter_func = f;
    applyProgFilter();
}

void AddPnpDialog::set_selected_progs(PnpInfoCache progDescrs)
{
    selected_progs = progDescrs;
    applyProgFilter();
    auto it = selected_progs.constFind(detailed_program_uuid);
    if(it != selected_progs.cend()) {
        update_detail_label(&(*it));
    }
}

//--- public slots

void AddPnpDialog::pnpCacheUpdated(const PnpInfoCache &cache)
{
    pnpCache = cache;
    applyProgFilter();
}

//--- private slots

void AddPnpDialog::on_comboBoxType_currentIndexChanged(int i)
{
    auto v = ui->comboBoxType->itemData(i);
    std::optional<DaqModule> type;
    if(!v.isNull())
        type = v.value<DaqModule>();

    if(cbFilter.type != type) {
        cbFilter.type = type;
        filter_item_changed();
    }
}

void AddPnpDialog::on_comboBoxHost_currentIndexChanged(int i)
{
    auto v = ui->comboBoxHost->itemData(i);
    std::optional<QString> host;
    if(!v.isNull())
        host = v.toString();

    if(cbFilter.host != host) {
        cbFilter.host = host;
        filter_item_changed();
    }
}

void AddPnpDialog::on_comboBoxIndex_currentIndexChanged(int i)
{
    auto v = ui->comboBoxIndex->itemData(i);
    std::optional<ProgramIndex> index;
    if(!v.isNull())
        index = v.value<ProgramIndex>();

    if(cbFilter.index != index) {
        cbFilter.index = index;
        filter_item_changed();
    }
}

void AddPnpDialog::on_leftTableView_doubleClicked(const QModelIndex &index)
{
    if(multi_select) {
        add_prog(modelIndex_to_progdescr(index));
        applyProgFilter();

        populate_all_filter();
        check_ok_button_enable();
        check_add_remove_button_enable(true);
    } else {
        selected_progs.clear();
        auto pdPtr = modelIndex_to_progdescr(index);
        Q_ASSERT(pdPtr);
        if(pdPtr) {
            selected_progs[pdPtr->uuid] = *pdPtr;
            accept();
        }
    }
}

void AddPnpDialog::on_rightTableView_doubleClicked(const QModelIndex &index)
{
    remove_prog(modelIndex_to_progdescr(index));
    applyProgFilter();

    populate_all_filter();
    check_ok_button_enable();
    check_add_remove_button_enable(false);
}

void AddPnpDialog::on_pushButtonAdd_clicked()
{
    for(auto i : ui->leftTableView->selectionModel()->selection().indexes()) {
        add_prog(modelIndex_to_progdescr(i));
    }
    applyProgFilter();

    populate_all_filter();
    check_ok_button_enable();
    check_add_remove_button_enable(true);
}

void AddPnpDialog::on_pushButtonRemove_clicked()
{
    for(auto i : ui->rightTableView->selectionModel()->selection().indexes()) {
        remove_prog(modelIndex_to_progdescr(i));
    }
    applyProgFilter();

    populate_all_filter();
    check_ok_button_enable();

    populate_all_filter();
    check_ok_button_enable();
    check_add_remove_button_enable(false);
}

void AddPnpDialog::tableView_selectionChanged(const QItemSelection &selected, const QItemSelection &selected2)
{
    bool isLeft = ui->leftTableView->selectionModel() == sender();
    refreshTableSelection(isLeft);
}

void AddPnpDialog::model_reseted()
{
    bool isLeft = leftModel == sender();
    refreshTableSelection(isLeft);
}

// private

void AddPnpDialog::applyProgFilter()
{
    filtredProgs = pnpCache;

    if(filter_func) {
        for(auto i=filtredProgs.cbegin(); i!=filtredProgs.cend();) {
            if(filter_func(*i)) {
                ++i;
            } else {
                i = filtredProgs.erase(i);
            }
        }
    }
    for(auto i=selected_progs.cbegin();i!=selected_progs.cend(); ++i)
        filtredProgs.remove(i.key());

    leftModel->setCache(filtredProgs);
    if(multi_select)
        rightModel->setCache(selected_progs);
    populate_all_filter();
    auto detPdIt = filtredProgs.constFind(detailed_program_uuid);
    if(detPdIt != filtredProgs.cend())
        update_detail_label(&(*detPdIt));

    // todo: somehow mark missing progs at right model
    // set_background_row_color(row, Qt::gray);
}

void AddPnpDialog::populate_all_filter()
{
    populate_program_filter(ui->comboBoxType);
    populate_program_filter(ui->comboBoxHost);
    populate_program_filter(ui->comboBoxIndex);
}

QVariant AddPnpDialog::getComboBoxField(QComboBox *cb, const ProgramDescription &pd) const
{
    if(cb == ui->comboBoxType) return QVariant::fromValue(pd.getDaqModule());
    if(cb == ui->comboBoxIndex) return QVariant::fromValue(pd.getProgramIndex());
    if(cb == ui->comboBoxHost) return pd.getHostName();
    Q_UNREACHABLE();
}

void AddPnpDialog::populate_program_filter(QComboBox *box)
{
    QVariant curCbData = box->currentData();
    auto tmpFilter = cbFilter;
    QString default_item;
    if(box==ui->comboBoxType) {
        tmpFilter.type.reset();
        default_item = k_all_types;
    } else if(box==ui->comboBoxHost) {
        tmpFilter.host.reset();
        default_item = k_all_hosts;
    } else if(box==ui->comboBoxIndex) {
        tmpFilter.index.reset();
        default_item = k_all_indexes;
    }

    QVariantMap filterItemsMap;
    for(auto pd : filtredProgs) {
        if(is_prog_match_cb_filter(pd, tmpFilter)) {
            auto val = getComboBoxField(box, pd);
            QString key;
            if(box==ui->comboBoxIndex) {
                key = val.value<ProgramIndex>().toString();
            } else
                key = val.toString();
            filterItemsMap[key] = val;
        }
    }

    bool old_state = box->blockSignals(true);
    if(box->count() == 0)
        box->addItem(default_item);

    if(!curCbData.isNull() && filterItemsMap.key(curCbData).isNull()) {
        curCbData = {};
    }
    auto it = filterItemsMap.cbegin();
    int cbIndex = 1;
    while(cbIndex < box->count() && it != filterItemsMap.cend()) {
        if(box->itemData(cbIndex) != *it)
            box->setItemData(cbIndex, *it);
        if(box->itemText(cbIndex) != it.key())
            box->setItemText(cbIndex, it.key());
        ++cbIndex;
        ++it;
    }

    while(box->count() != cbIndex)
        box->removeItem(cbIndex);
    for(; it!=filterItemsMap.cend(); ++it) {
        box->addItem(it.key(), *it);
    }

    box->blockSignals(old_state);
    set_current_index(box, curCbData);
}

void AddPnpDialog::filter_item_changed()
{
    leftModel->setComboBoxFiler(cbFilter);
    populate_all_filter();
}

void AddPnpDialog::update_detail_label(const ProgramDescription *pd)
{
    ui->labelProgInfo->show();
    ui->labelProgInfo->clear();
    ui->labelProgInfo->setText(pd ? pd->toString() : "");
    ui->labelProgInfo->adjustSize();
    detailed_program_uuid = pd ? pd->uuid : QUuid();
}

void AddPnpDialog::set_current_index(QComboBox *box, const QVariant &item)
{
    int i = std::max(0, box->findData(item));
    box->setCurrentIndex(i);
}

void AddPnpDialog::init_table(QTableView *table, AddPnpModel *model)
{
    table->setModel(model);
    connect(table->selectionModel(), &QItemSelectionModel::selectionChanged,
            this, &AddPnpDialog::tableView_selectionChanged);
    connect(model, &AddPnpModel::modelReset,
            this, &AddPnpDialog::model_reseted);
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
}

void AddPnpDialog::refreshTableSelection(bool isLeft)
{
    auto sModel = (isLeft ? ui->leftTableView : ui->rightTableView)->selectionModel();

    const auto &indexList = sModel->selection().indexes();
    bool hasAny = !indexList.isEmpty();
    set_push_button_enable(isLeft, hasAny);

    const ProgramDescription *pd = nullptr;
    if(!indexList.isEmpty()) {
        if(isLeft) {
            ui->rightTableView->clearSelection();
        } else {
            ui->leftTableView->clearSelection();
        }
        pd = modelIndex_to_progdescr(indexList.last());
    }
    if(isLeft)
        check_ok_button_enable();
    update_detail_label(pd);
}

void AddPnpDialog::set_push_button_enable(bool isAdd, bool enable)
{
    auto button = isAdd ? ui->pushButtonAdd : ui->pushButtonRemove;
    if(enable) {
        button->setEnabled(true);
        auto col = isAdd ? QColor(159, 255, 73) : QColor(248, 86, 27);
        change_button_color(button, col);
    } else {
        button->setEnabled(false);
        change_button_color(button, QColor(255, 255, 255));
    }
}

void AddPnpDialog::change_button_color(QPushButton *push_button, QColor color)
{
    QPalette pal = push_button->palette();
    pal.setColor(QPalette::Button, color);
    push_button->setAutoFillBackground(true);
    push_button->setPalette(pal);
    push_button->update();
}

const ProgramDescription *AddPnpDialog::modelIndex_to_progdescr(const QModelIndex &i) const
{
    auto uuid = i.model()->data(i, Qt::UserRole).toUuid();
    const auto &cache = (i.model() == leftModel) ? filtredProgs : selected_progs;
    auto it = cache.constFind(uuid);
    if(it != cache.cend())
        return &(*it);
    return nullptr;
}

void AddPnpDialog::check_add_remove_button_enable(bool isAdd)
{
    bool en = hasSelection(isAdd ? ui->leftTableView : ui->rightTableView);
    set_push_button_enable(isAdd, en);
}

void AddPnpDialog::check_ok_button_enable()
{
    bool en = !selected_progs.isEmpty();
    if(!multi_select && !en) {
        en = ui->leftTableView->selectionModel()->hasSelection();
    }

    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(en);
}

bool AddPnpDialog::is_prog_match_cb_filter(const ProgramDescription &pd) const
{
    return is_prog_match_cb_filter(pd, cbFilter);
}

bool AddPnpDialog::is_prog_match_cb_filter(const ProgramDescription &pd, const AddPnpComboBoxFilter &filter) const
{
    if(filter.type && pd.getDaqModule() != filter.type.value())
        return false;
    if(filter.host && pd.getHostName() != filter.host.value())
        return false;
    if(filter.index && pd.getProgramIndex() != filter.index.value())
        return false;
    return true;
}

void AddPnpDialog::add_prog(const ProgramDescription *pdPtr)
{
    if(pdPtr)
        selected_progs[pdPtr->uuid] = *pdPtr;
}

void AddPnpDialog::remove_prog(const ProgramDescription *pdPtr)
{
    if(pdPtr)
        selected_progs.remove(pdPtr->uuid);
}
