#ifndef PNPMESSAGE_H
#define PNPMESSAGE_H

class QByteArray;
class QDomElement;
#include <QHostAddress>
#include <QHostInfo>
#include <QStringList>
#include <QMetaType>
#include <QUuid>
#include <QElapsedTimer>

#include "base/ClientIndex.h"

class DaqModule;
class ProgramIndex;

enum ProgramInterfaceType { ProgramInterfaceEmty,
                            ProgramInterfaceRemoteControl,
                            ProgramInterfaceDataFlow,
                            ProgramInterfaceMonitorData,
                            ProgramInterfaceConfigTransfer
                          };

struct ProgramInterfacePeer
{
    ProgramInterfacePeer() = default;
    ProgramInterfacePeer(const QHostAddress &h, quint32 p) : host(h), port(static_cast<quint16>(p))
    {}
    friend bool operator==(const ProgramInterfacePeer &, const ProgramInterfacePeer &) = default;

    friend uint qHash(const ProgramInterfacePeer &r, uint seed = 0)
    {
        return qHash(r.host) ^ qHash(r.port) ^ seed;
    }

    QString toString() const { return QString("%1:%2").arg(host.toString()).arg(port); }

    QHostAddress host;
    quint16 port = 0;
};
Q_DECLARE_METATYPE(ProgramInterfacePeer)

class ProgramInterface
{
public:
    ProgramInterface() : host(QHostAddress::LocalHost) {}
    ProgramInterface(const QHostAddress &h, const quint32 p, const QString &t, bool free = true, int i = 0) :
        host(h), port(static_cast<quint16>(p)), type(t), enabled(true), isFree(free), id(i)
    {}

    friend bool operator==(const ProgramInterface &, const ProgramInterface &) = default;

    static QString getTypeName(ProgramInterfaceType type);
    void setType(ProgramInterfaceType t);
    bool isEmpty() const { return type.isEmpty(); }

    void setPeer(const QHostAddress &h, quint32 p);
    void setPeer(const ProgramInterfacePeer &p);
    void addPeer(const QHostAddress &h, quint32 p);
    void addPeer(const ProgramInterfacePeer &p);

    QHostAddress host;
    quint16 port = 0;
    QString type;
    QSet<ProgramInterfacePeer> peers;
    bool enabled = false;
    bool isFree = true;
    int id = 0;
};
Q_DECLARE_METATYPE(ProgramInterface)

class ProgramDescription
{
public:
    ProgramDescription();
    QUuid uuid;
    quint32 seq = 0;
    ClientIndex clientIndex;
    ClientIndex parentClientIndex;
    QHostAddress host;
    QString hostName;
    QVector<ProgramInterface> interfaces;
    ProgramInterface dummyInterface;
    QMap<QString, QString> options;
    QString ver_hash;
    QString ver_date;

    ProgramInterface &getProgramInterface(ProgramInterfaceType type, bool *ok=nullptr, int id=-1);
    const ProgramInterface &getProgramInterface(ProgramInterfaceType type, bool *ok=nullptr) const;
    ProgramInterface &getProgramInterface(QString type, bool *ok=nullptr, int id=-1);
    const ProgramInterface &getProgramInterface(QString type, bool *ok=nullptr) const;
    int getProgramInterfaceNum(ProgramInterfaceType type);
    int getProgramInterfaceNum(QString type=QString());
    bool containsProgramInterface(ProgramInterfaceType type, int id);
    bool containsProgramInterface(QString type, int id);
    bool updateProgramInterface(ProgramInterfaceType type, int id, int port, bool isListening, bool isFree, const QHostAddress &peerHost, int peerPort);

    bool isOutOfDate(const QElapsedTimer &other) const;
    QString getAnnounceStr() const;
    QString getClosingStr() const;
    QString toString() const;
    void clear();
    quint32 getHostIp() const;
    QString getHostName() const;
    DaqModule getDaqModule() const { return clientIndex.getProgramType(); }
    ProgramIndex getProgramIndex() const { return clientIndex.getProgramIndex(); }
    ClientIndex getClientIndex() const { return clientIndex; }
    void setClientIndex(const ClientIndex &i) { clientIndex = i; }
    ClientIndex getParent() const { return parentClientIndex; }
    void setParent(const ClientIndex &i) { parentClientIndex = i; }
    QElapsedTimer getDiscoverTimer() const { return discoverTimer; }

    void setProgramIndex(const ProgramIndex &i) { clientIndex.program_index = i; }
    QString getName() const { return clientIndex.toString(); }
    bool isEmpty() const { return clientIndex.program_type.isEmpty()
                || clientIndex.program_index.isDefault(); }
    bool sameProgram(const ProgramDescription &descr) const {
        return clientIndex == descr.getClientIndex();
    }
    friend bool operator <(const ProgramDescription &pd1, const ProgramDescription &pd2){
        return pd1.getName()<pd2.getName() || (pd1.getName()==pd2.getName() && pd1.uuid<pd2.uuid);
    }
    friend bool operator==(const ProgramDescription &, const ProgramDescription &) = default;

private:
    QElapsedTimer discoverTimer;
};
Q_DECLARE_METATYPE(ProgramDescription)

class PnpInfoCache: public QHash<QUuid, ProgramDescription> {};
Q_DECLARE_METATYPE(PnpInfoCache);

enum PNPMessage_Type { PNPMessage_Type_Search, PNPMessage_Type_Discover, PNPMessage_Type_Close};

class PNPMessage
{
public:
    PNPMessage();
    QString toString() const;
    QString getXml() const;
    bool decode(const QByteArray &data, const QHostAddress &host, bool *ok=nullptr);

    void addInterface(const QString &type, const QHostAddress &host, const quint32 port);
    bool decodeOptions(const QDomElement &domEl, const QString &hostName);
    bool decodeInterfaces(const QDomElement &domEl, const QHostAddress &host);

    PNPMessage_Type type;
    DaqModuleSet searchTargets; // list program types

    ProgramDescription progDesc;
};

#endif // PNPMESSAGE_H
