/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RemoteControlTlvServer.h"

#include <QDebug>
#include <QHostInfo>
#include <QLoggingCategory>
#include <QSet>
#include <QTcpServer>
#include <QTcpSocket>
#include <QTimer>

#include <stdexcept>
#include <unistd.h>

#include "RemoteControlProtocol.h"
#include "RemoteControlServerStatus.h"
#include "RunControlCommands.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_rc_server)
Q_LOGGING_CATEGORY(lc_rc_server, "afi.rc_server")
}

namespace {
const qint32 MAX_BLOCK_SIZE = 100e6; // upper limit for incoming block
}

RemoteControlTlvServer::RemoteControlTlvServer(quint16 preferedTcpPort, QObject *parent) :
    QObject(parent),
    tcpSocket(nullptr),
    qTcpServer(new QTcpServer(this)),
    status(new RemoteControlServerStatus(preferedTcpPort))
{
    qRegisterMetaType<RemoteControlServerStatus> ();
    qRegisterMetaType<TypeLengthHeader>();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &RemoteControlTlvServer::updateConnections);
}

RemoteControlTlvServer::~RemoteControlTlvServer()
{
    blockSignals(true);
    startListen(false);
    blockSignals(false);
    delete qTcpServer;
    delete status;
}

int RemoteControlTlvServer::getTcpPort() const
{
    return status->tcpPort;
}

bool RemoteControlTlvServer::isFree() const
{
    return !status->portInUse;
}

bool RemoteControlTlvServer::isListening() const
{
    return status->listening;
}

void RemoteControlTlvServer::updateConnections()
{
    while (QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if (tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)) {
            tcpSocket->abort();
            Q_ASSERT(!tcpSocket);
            status->portInUse = false;
        }
        if (tcpSocket == nullptr) {
            tcpSocket = pendingConnection;
            status->portInUse = true;
            status->peerHost = tcpSocket->peerAddress();
            status->peerPort = tcpSocket->peerPort();

            connect(tcpSocket, &QTcpSocket::readyRead,
                    this, &RemoteControlTlvServer::readData);

            connect(tcpSocket, &QTcpSocket::disconnected,
                    this, &RemoteControlTlvServer::onDisconnected);

            QHostInfo info = QHostInfo::fromName(status->peerHost.toString());
            qInfo(lc_rc_server).noquote()
                << QString("accepted connection from %1:%2")
                       .arg(info.hostName())
                       .arg(tcpSocket->peerPort());
            updateStatus();
            emit connectionAccepted();
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if (tcpSocket == nullptr) {
        qDebug(lc_rc_server).noquote()
            << "There are no connections.";
    }
}

void RemoteControlTlvServer::updateStatus() const
{
    emit connectionStatusUpdated(*status);
}

void RemoteControlTlvServer::startListen(bool state)
{
    if (status->listening != state) {
        if (state) {
            listen();
        } else {
            closeServer();
        }
        status->listening = state;
    }
    updateStatus();
}

void RemoteControlTlvServer::listen()
{
    while (!qTcpServer->listen(QHostAddress::Any, status->preferedTcpPort)) {
        //        QAbstractSocket::SocketError errNum = qTcpServer->serverError();
        //        if(errNum==QAbstractSocket::AddressInUseError){
        //            if(!portInUse){
        //                QString mess = tr("Unable to start the remote control server on port %1: %2, restart in 5 s")
        //                        .arg(tcpPort).arg(qTcpServer->errorString());
        //                sendLog(mess, Qt::red);
        //                portInUse=true;
        //                qFatal(mess.toStdString().c_str());
        //            }
        //            exit(1);
        //            sleep(5);
        //            continue;
        //        }
        QString s = tr("RemoteControlServer unable to listen on TCP port %1: %2")
                        .arg(status->preferedTcpPort)
                        .arg(qTcpServer->errorString());
        qCritical(lc_rc_server) << s;
        throw std::runtime_error(s.toStdString());
    }
    status->tcpPort = qTcpServer->serverPort();

    qInfo(lc_rc_server).noquote()
        << QString("listening on %1:%2")
               .arg(qTcpServer->serverAddress().toString())
               .arg(status->tcpPort);
}

void RemoteControlTlvServer::closeServer()
{
    qTcpServer->close();
    closeSocket();
    qDebug(lc_rc_server).noquote()
        << QString("closed %1:%2")
               .arg(qTcpServer->serverAddress().toString())
               .arg(status->tcpPort);
}

void RemoteControlTlvServer::closeSocket()
{
    if (tcpSocket) {
        tcpSocket->abort();
        status->portInUse = false;
        updateStatus();
    }
    data.clear();
}

void RemoteControlTlvServer::onDisconnected()
{
    if(auto obj = sender()) {
        if(tcpSocket == obj)
            tcpSocket = nullptr;
        obj->deleteLater();
    }

    data.clear();
    status->portInUse = false;
    updateStatus();
    qInfo(lc_rc_server).noquote()
        << "disconnected";
}

bool RemoteControlTlvServer::isConnectionAlive() const
{
    return tcpSocket && tcpSocket->isValid()
            && (tcpSocket->state() == QTcpSocket::ConnectedState);
}

inline bool RemoteControlTlvServer::isDataSizeEnoughForHdr() const
{
    return static_cast<uint>(data.size()) >= sizeof(TypeLengthHeader);
}

TypeLengthHeader* RemoteControlTlvServer::findHeader()
{
    TypeLengthHeader *hdr=nullptr;
    bool type_warn=true, size_warn=true;


    while(isDataSizeEnoughForHdr()) {
        hdr= reinterpret_cast<TypeLengthHeader*>(data.data());
        if((hdr->type == RC_MSG_CONFIG_V4) || (hdr->type == RC_MSG_CMD_V4) || (hdr->type == RC_MSG_JSON_V5)) {
            if(hdr->len <= MAX_BLOCK_SIZE) {
                break;
            } else if(size_warn) {
                size_warn = false;
                qWarning(lc_rc_server).noquote()
                    << QString("message error: block size is %1. Permissible maximum is %2")
                                  .arg(hdr->len).arg(MAX_BLOCK_SIZE);
            }
        } else if(type_warn) {
            qWarning(lc_rc_server).noquote()
                << QString("message error: unknown message type 0x%1")
                       .arg(hdr->type, 8, 16, QChar('0'));
            type_warn = false;
        }
        if(!prevDataTrashed) {
            qWarning(lc_rc_server).noquote()
                    << "Start trashing some data from client";
            prevDataTrashed = status->trashedSize;
        }

        ++status->trashedSize;
        data.remove(0, 1);
        hdr = nullptr;
    }
    return hdr;
}

void RemoteControlTlvServer::send_byte_array(quint32 type, const QByteArray &payload) const
{
    send_rc_block(tcpSocket, type, payload);
}

void RemoteControlTlvServer::send_json(const QJsonObject &obj) const
{
    send_rc_json(tcpSocket, obj);
}

void RemoteControlTlvServer::readData()
{
    try{
        while(tcpSocket && tcpSocket->bytesAvailable()){
            data.append(tcpSocket->readAll());

            while(isDataSizeEnoughForHdr()) {
                TypeLengthHeader *hdr = findHeader();
                if(!hdr)
                    continue;

                if(data.size() < static_cast<int>(hdr->len))
                    break;

                if(prevDataTrashed) {
                    qWarning(lc_rc_server).noquote()
                            << QString("Got good TLV data; Trashed size=%1(%2)")
                               .arg(status->trashedSize).arg(status->trashedSize-*prevDataTrashed);
                    prevDataTrashed.reset();
                }

                switch (hdr->type) {
                case RC_MSG_JSON_V5 :
                case RC_MSG_CONFIG_V4 :
                case RC_MSG_CMD_V4 : {
                    const auto &payload = data.mid(sizeof(TypeLengthHeader),
                                                   static_cast<int>(hdr->len));
                    emit tlvDataReceived(*hdr, payload);
                    break;
                }
                default:
                {
                    static QSet<quint32> missingTypes;
                    if(!missingTypes.contains(hdr->type)) {
                        missingTypes << hdr->type;
                        qCritical(lc_rc_server) << __PRETTY_FUNCTION__
                                   << "missing handler for type:"
                                   << Qt::hex << hdr->type;
                    }
                }
                }
                quint32 tsd_size = hdr->len + sizeof(TypeLengthHeader);
                data.remove(0, static_cast<int>(tsd_size));
            }
        }
    }
    catch(...) {
        qWarning(lc_rc_server).noquote()
            << "Exception in" << __PRETTY_FUNCTION__;
    }
}
